/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#ifndef BIGDECIMAL_H_
#define BIGDECIMAL_H_

#include "CarobException.hpp"
#include "Common.hpp" //for java_byte, ResultSetDataType and controllerDecimalPoint

#include <stdint.h> // for uint8_t, int32_t, uint32_t, int64_t, uint64_t
// Alternative:
// #include <cstdint>

// Warning: this header and all the types it defines (including "long
// long") are part of C99 and are MISSING from C++98.

// See:
// - "Incompatibilities Between ISO C and ISO C++" - David R. Tribble
// - C++ Technical Report 1 (TR1)

#include <gmp.h>
#include <string>
#include <sstream>

namespace CarobNS {

class DriverSocket;
class std::locale;
/**
 * Java math.BigDecimal equivalent class for storing immutable, 
 * arbitrary-precision signed decimal numbers. A BigDecimal consists of an 
 * arbitrary precision integer <i>unscaled value</i> and a non-negative 32-bit 
 * integer <i>scale</i>, which represents the number of digits to the right of
 * the decimal point.  The number represented by the BigDecimal is
 * <tt>(unscaledValue/10<sup>scale</sup>)</tt>.
 * One notable difference with Java is the unscaled value, that is not
 * implemented with a BigInteger, but directly with an array of int (actually 
 * exactly as in BigInteger, minus utility functions the class provides)
 */
class BigDecimal
{
public:

  /**
   * Empty constructor.
   * Sets fields to default values. This big decimal becomes zero (0 scale 
   * and 0 unscaled_value)
   */
  BigDecimal();

  /**
   * Constructs a BigDecimal which value is parsed from the given wstring 
   * (converted to a string) using #createFromString() function
   * @param s wide string to parse into a BigDecimal
   * @throw ConversionException if the given string does not represent a valid
   *        BigDecimal
   * @see #createFromString()
   */
  BigDecimal(const std::wstring& s) throw (ConversionException, UnexpectedException);
  /**
   * Constructs a BigDecimal which value is parsed from the given string using
   * #createFromString() function.
   * @param s string to parse into a BigDecimal
   * @throw ConversionException if the given string does not represent a valid
   *        BigDecimal
   * @see #createFromString()
   */
  BigDecimal(const std::string& s) throw (ConversionException, UnexpectedException);
  /**
   * Constructs a BigDecimal which value is taken from the given argument
   */
  BigDecimal(const int32_t);
  /**
   * Constructs a BigDecimal which value is taken from the given argument
   */
  BigDecimal(const int64_t);
  //TODO:BigDecimal(const float);
  //TODO:BigDecimal(const double);

  /** Copy constructor */
  BigDecimal(const BigDecimal&);

  /**
   * Copies scale sign and unscaled_value
   */
  BigDecimal& operator = (const BigDecimal&);
  
  /**
   * Compares scale, sign and unscaled_value
   */
  bool operator == (const BigDecimal&) const;
  /**
   * BigDecimal display or serialization as a string with default C locale
   * decimal point. This operator is used to send the bigdecimal as a string to
   * the controller for parameter statements. The controller uses the default
   * decimal point separator for parsing, so we have to send it the right one,
   * which is the default C one.
   * @param os stream in which to write the string representation of this BD
   * @param bd value to be serialized
   */
  friend std::basic_ostream<wchar_t>& operator << (std::basic_ostream<wchar_t>& os, 
      const BigDecimal &bd)  { return os<<bd.toString(std::locale::classic()); }
  /**
   * Constructor to deserialize a BigDecimal from a stream.
   * @param input socket from which to deserializer Big Decimal
   * @throws SocketIOException if an error occurs on the stream
   */
  BigDecimal(const DriverSocket& input)
      throw (SocketIOException, UnexpectedException);
  /**
   * Frees intVal if applicable
   */
  virtual ~BigDecimal();

  /**
   * Conversion to string in the user defined locale (for decimal separator)
   * A leading minus sign is used to indicate sign, and the number of digits to
   * the right of the decimal point is used to indicate scale.
   */ 
  operator std::wstring() const;
  /**
   * Conversion to string with the given locale.<br>
   * Adds optionnal leading minus sign, decimal point and thousand separator.
   * Decimal and thousand separators are taken from the given locale.<br>
   * Note that the locale's grouping property is only used to determine if a
   * grouping must be done. If so, grouping is assumed to be "\003" whatever
   * the locale is (eg. 12345 will always be "12<sep>345").<br> 
   * <i>This function is mandatory due to the internal representation that we use
   * (full number + scale), which doesn't allow to use gmp's or stl's
   * functions.</i>
   * @param loc the locale to be used for string conversion
   */ 
  std::wstring  toString(const std::locale& loc) const;
  /**
   * Conversion to int32.
   * Any fractional part of this BigDecimal will be discarded, and if the
   * resulting "BigInteger" is too big to fit in an int32, throws an exception
   * @throw ConversionException if the value is to big to be converted
   */ 
  operator int32_t() const throw (ConversionException);
  /**
   * Conversion to int64_t
   * Any fractional part of this BigDecimal will be discarded, and if the
   * resulting "BigInteger" is too big to fit in a int64_t, throws an exception
   * @throw ConversionException if the value is to big to be converted
   */ 
  operator int64_t() const throw (ConversionException);
  /**
   * Conversion to uint64_t
   * Any fractional part of this BigDecimal will be discarded, and if the
   * resulting "BigInteger" is too big to fit in an uint64_t, throws an
   * exception
   * @throw ConversionException if the value is to big to be converted
   */ 
  operator uint64_t() const throw (ConversionException);
  /**
   * Conversion to float
   * TODO: exact behavior description
   * @throw ConversionException if the value is to big to be converted
   */ 
  operator float() const throw (ConversionException);
  /**
   * Conversion to double
   * TODO: exact behavior description
   * @throw ConversionException if the value is to big to be converted
   */ 
  operator double() const throw (ConversionException);
protected:
  /**
   * Initializes this big decimal to 0 (zero).
   * There will be 0 signum, zero unscaled_value, zero scale, zero byte array
   * length and byte array pointer set to NULL
   */
  void          initToZero();
  /**
   * Fills this BigDecimal fields (scale, signum, unscaled_value) by converting
   * the given string into a BigDecimal, assuming it is base 10.
   * Finds the scale of the given number by parsing the input string in order to
   * find the decimal point taken from the current locale. The scale is the
   * number of digits found after the decimal point, ignoring white spaces.
   * @param str string representing the number in base 10, that can contain
   * white spaces (ignored).
   * @throw ConversionException if the given string does not represent a valid
   * BigDecimal in base 10
   */
  void          createFromString(const std::string &str)
                    throw (ConversionException, UnexpectedException);
  /**
   * Converts this BigDecimal's absolute value to a GMP integer by scaling
   * its unscaled_value (thus discarding fractional part)
   * @param     result the scaled value. Note: result must be initialized !
   */
  void          toBigInteger(mpz_t result) const;
  /**
   * Converts the given gmp integer to an array of integers.
   * The integer array will be allocated by this function, and must be deleted
   * by the caller
   * @param     scaled_val gmp integer to convert (usually the return value of
   *            toBigInteger)
   * @param intArrayLength size of the returned array
   */
  int*          toIntArray(mpz_t scaled_val, int* intArrayLength) const;
private:
  /** Length of the unscaled value java_byte-array */
  size_t        byteArrayLength;
  /**
   * Unscaled value of this BigDecimal, as a java_byte array two's-complement 
   * <i>big-endian</i> order: the zeroth element of this array is the most-
   * significant byte of the scale.
   */
  java_byte*    byteArray;

  /** Scale of this BigDecimal. */
  int           scale;

  /** Sign (+ or - or 0) */
  int           signum;
  /** Unscaled value as a GMP integer */
  mpz_t         unscaled_value;
};

} //namespace CarobNS
#endif /*BIGDECIMAL_H_*/
