/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2007 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#ifndef _BUFFERED_SOCKET_H_
#define _BUFFERED_SOCKET_H_

#include "JavaSocket.hpp"

#include <string.h>

namespace CarobNS {

/** Size of the input buffer */
#define READ_BUFFER_SIZE    2048
#define WRITE_BUFFER_SIZE    512

/**
 * Holds a buffer for read operations
 */
class ReadCache
{
public:
  /** Creates and empty cache */
  ReadCache() : data_boundary(0), data_index(0) {} ;
  /** Returns the size of valid data left to be read*/
  int               getDataSize() { return data_boundary - data_index; }
  /** Returns a pointer to the first byte of data available for reading */
  java_byte*        getData() { return &(buffer[data_index]); }
  /** Tells that the given length of data has been read */
  void              discard(int length) { data_index += length; }
  /** Set the new valid data size */
  void              setDataSize(int newSize) { data_boundary = newSize; }
  /** Discards any data */
  void              reset() { data_boundary = 0; data_index = 0; }
private:
  /** Buffer in which data read from the network is stored */
  java_byte         buffer[READ_BUFFER_SIZE];
  /** Last available byte of read data + 1 */
  int               data_boundary;
  /** Index of the next byte to be read */
  int               data_index;
};

/**
 * Holds a buffer for writes operations
 */
class WriteCache
{
public:
  /** Creates and empty cache */
  WriteCache() : write_index(0) {};
  /** Returns the size of data written until now */
  int               getDataSize() { return write_index; }
  /** Returns a pointer to the first available byte */
  java_byte*        getData() { return buffer; }
  /** Add data to cache */
  void              append(const java_byte* src, int len)
  {
    memcpy(&buffer[write_index], src, len);
    write_index += len;
  }
  /** Discards all data */
  void              reset() { write_index = 0; }
  /** Returns the size left in the buffer */
  int               getRemainingSize() { return WRITE_BUFFER_SIZE - write_index; }
//private:
  /** Buffer in which data read from the network is stored */
  java_byte         buffer[WRITE_BUFFER_SIZE];
  /** Index of the next byte to be read */
  int               write_index;
};


/**
 * Extends the <code>JavaSocket</code> class to allow buffered read operations
 * @see JavaSocket
 */
class BufferedSocket : public JavaSocket
{
public:
  /**
   * Sends the buffered write data to the socket
   */
  void                flush()  const throw (SocketIOException, UnexpectedException);
protected:
  /**
   * Empty constructor
   */
  BufferedSocket() throw (CodecException) : JavaSocket() {} ;
  /**
   * Substitute for recv. Waits for incomming data by calling pollOnSingleFd
   * and loops until full length has been received or an error occured. If
   * shutdown() is called during the loop, throws a SocketIOException to inform
   * callers that the socket is not longer readable. If there are output data to
   * be sent in the write buffer, this function will flush these data. This
   * prevents from doing a flush() everywhere in the code, but still can be
   * overriden by calling manually #flush() (could be an optimization (?))
   * @param buf data to send
   * @param len full buffer length
   * @param flags send options, see recv man page
   * @return the total number of bytes send, -1 in case of failure
   * @throw SocketIOException if interrupted by shutdown() function
   */
  int32_t             recvFully(void *buf, const int len, const int flags)
                          const throw (SocketIOException, UnexpectedException);
  /**
   * Wrapper over send(...) function to buffer output data and handle errors.<br>
   * Data is only sent to the socket when the buffer is full or when a read
   * operation is requested.
   *
   * @param fctName name of the calling function (for logging purposes)
   * @param buf data to be send
   * @param len length of buf
   * @param flags send option, see recv man page
   * @throws SocketIOException
   * @see #recvFully(void *, const int, const int) const
   */
  void                sendToSocket(const wchar_t fctName[],
                          const void* buf, int len, int flags) const
                          throw (SocketIOException, UnexpectedException);
private:
  /**
   * Buffer in which data read from the network is stored.<br>
   * As read operations don't really modify the socket, this object doesn't
   * affects constness
   */
  mutable ReadCache   read_cache;
  /**
   * Buffer for data to be written<br>
   */
  mutable WriteCache  write_cache;
  /**
   * Buffered socket receive function.<br>
   * Reads available data in <code>read_buffer</code> and returns the number of
   * bytes read (ie. available)
   *
   * @param recvFlags flags to pass to recv function
   * @throw SocketIOException if an error occurs during the receive operation
   * @return the number of bytes received
   */
  int                 fillReadBuffer(int recvFlags) const throw (SocketIOException,
                          UnexpectedException);
};

} //namespace CarobNS
#endif /*_BUFFERED_SOCKET_H_*/
