/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#ifndef _CAROBEXCEPTION_H_
#define _CAROBEXCEPTION_H_

#include <ostream>

#include <string>
#include <vector>

#ifdef __GLIBC__ // probably not the right macro
#include <cxxabi.h>
#define CAROB_CXX_DEMANGLING 1
#endif

namespace {
  const int            DEFAULTVENDORCODE(-1);
  const std::wstring   DEFAULTSQLSTATE(L"HY000");
}

namespace CarobNS {

class StackTraceElement;
class DriverSocket;

/** Exceptions' backtrace consisting in a vector of trace strings */
// redefine the default ctor to get the current backtrace
class BackTrace : public std::vector<std::string>
{
public:
  BackTrace();
};

/**
 * Backtrace printer
 */
std::wostream& operator<< (std::wostream& outstream, const BackTrace& bt);

typedef std::vector<StackTraceElement> StackTrace;

/**
 * Stacktrace printer
 */
std::wostream& operator<< (std::wostream& outstream, const StackTrace& st);

/**
 * Mother class of all exceptions send by the driver defining common exception 
 * interface.
 */
class CarobException
{
public:
  /**
   * Copy constructor. Copying, NOT chaining!
   * @param ce source exception to copy from
   */
  CarobException(const CarobException& ce);
  /**
   * Constructs/reads a new Exception chain from the stream
   * @param sock socket from which to deserialize exception
   * @throw SocketIOException. Not specified in method signature because of
   * forward declarations/loop issue (SocketIOException inherits from
   * CarobException
   */
  CarobException(const DriverSocket& sock); // throw (SocketIOException)
  /**
   * Constructs a new exception with the specified detail message and
   * cause.  <p>Note that the detail message associated with
   * <code>cause</code> is <i>not</i> automatically incorporated in
   * this exception's detail message.
   *
   * @param  messagePrm the detailed user-friendly message
   * @param  SQLStatePrm the SQL2 standard error code.
   * @param  causePrm the exception that generated this one, if any.
   * @param  vendorPrm the (deprecated) SQL1 error code
   */
  CarobException(const std::wstring& messagePrm,
                 const std::wstring& SQLStatePrm = DEFAULTSQLSTATE,
                 CarobException* causePrm = 0,
                 int vendorPrm = DEFAULTVENDORCODE) :
    message(messagePrm),
     // deep copy constructor (and... slicing)
    causePtr(causePrm == 0 ? 0 : new CarobException(*causePrm)),
    SQLState(SQLStatePrm),
    errorCode(vendorPrm)
  { }
  /**
   * Destructor to clean up stack trace and cause exception
   */
  virtual ~CarobException();

  /** Gets the exception error message */
  virtual const std::wstring&  description() const { return message; }

  /** Returns the remote java stack trace */
  const StackTrace& getStackTrace() const { return stackTrace; }

  /** Returns the local stack trace */
  const BackTrace& getBackTrace() const { return backTrace; }

  /** Returns a reference to the next exception in the chain, NULL if tail */
  const CarobException* getNext() const { return causePtr; }

  /** get the SQLState if any */
  const std::wstring & getSQLState() const { return SQLState; }
  /** Retrieves the vendor-specific exception code */
  int getErrorCode() const { return errorCode; }

  /** wstring cast operator - makes use of << operators */
  operator std::wstring() const;
   
protected:
  /** Error message */
  std::wstring                message;
  /** Stack trace as a vector of traces */
  StackTrace                  stackTrace;
  // TODO: try to use a real field + a special "end of chain" exception.
  /** pointer to the exception that generated this one */
  CarobException*             causePtr;

  /** SQL2 standard error code */
  std::wstring                SQLState;
  // According to appendix A of the CLI standard
  // http://www.opengroup.org/bookstore/catalog/c451.htm
  // See also appendix B of the SQL standard available here:
  // http://www.opengroup.org/bookstore/catalog/c449.htm
  // The latter is a superset of the former, minus HY___ and a couple
  // of other serialization or transactional CLI error codes.
  //
  // Quite extensive lists of possible SQLStates can also be found on-line here:
  // - http://www.easysoft.com/developer/interfaces/odbc/sqlstate_status_return_codes.html 
  // - http://msdn.microsoft.com/library/default.asp?url=/library/en-us/odbc/htm/odbcodbc_error_codes.asp

  /** Deprecated SQL1 "standard" error code. Negative for errors, positive for warnings. */
  int                         errorCode;
  /** Local stack trace when available */
  const BackTrace             backTrace;
private:
  CarobException & operator =( CarobException & forbidden);
};

/**
 * CarobException printer
 */
std::wostream& operator<<(std::wostream& outstream, const CarobException& ce);


/**
 * The goal UnexpectedException of is to prevent from having a std::unexpected() 
 * call when a function throws an exception not declared in its prototype. Thus,
 * all functions declared as throwing an exception MUST also declare <code>throw
 * (UnexpectedException)</code> so that even if an unknown exception is thrown
 * at runtime, we still will catch it with <code>catch (UnexpectedException e)
 * </code>.
 */
class UnexpectedException : public CarobException
{
public:
  /**
   * Constructs an UnexpectedException with the given message
   *
   * @param  messagePrm the detailed user-friendly message
   * @param  SQLStatePrm the SQL2 standard error code.
   * @param  causePrm the exception that generated this one, if any.
   * @param  vendorPrm the (deprecated) SQL1 error code
   * @see CarobException
   */
  UnexpectedException(const std::wstring& messagePrm,
                  const std::wstring& SQLStatePrm = DEFAULTSQLSTATE,
                  CarobException* causePrm = 0,
                  int vendorPrm = DEFAULTVENDORCODE) :
    CarobException(messagePrm, SQLStatePrm, causePrm, vendorPrm)
  { }
  /**
   * std::unexpected handler. Throws an Unexpected exception instead of
   * terminating the program
   */
  static void convertUnexpected();
};

/**
 * Exception inside the Driver
 */
class DriverException : public CarobException
{
public:
  /**
   * Constructs a DriverException with the given message
   *
   * @param  messagePrm the detailed user-friendly message
   * @param  SQLStatePrm the SQL2 standard error code.
   * @param  causePrm the exception that generated this one, if any.
   * @param  vendorPrm the (deprecated) SQL1 error code
   */
  DriverException(const std::wstring& messagePrm,
                  const std::wstring& SQLStatePrm = DEFAULTSQLSTATE,
                  CarobException* causePrm = 0,
                  int vendorPrm = DEFAULTVENDORCODE) :
    CarobException(messagePrm, SQLStatePrm, causePrm, vendorPrm)
  { }
};

/**
 * Exception on the socket
 */
class JavaSocketException : public DriverException
{
public:
  /**
   * Constructs a JavaSocketException with the given message
   * @param s error message of the exception
   */
	JavaSocketException(std::wstring s) : DriverException(s) {};
};

/**
 * Exception on connection. Send when connecting to a controller
 */
class ConnectionException : public DriverException
{
public:
  /**
   * Constructs a ConnectionException
   *
   * @param  messagePrm the detailed user-friendly message
   * @param  SQLStatePrm the SQL2 standard error code.
   * @param  causePrm the exception that generated this one, if any.
   * @param  vendorPrm the (deprecated) SQL1 error code
   * @see DriverException
   */
  ConnectionException(const std::wstring& messagePrm,
                  const std::wstring& SQLStatePrm = DEFAULTSQLSTATE,
                  CarobException* causePrm = 0,
                  int vendorPrm = DEFAULTVENDORCODE) :
    DriverException(messagePrm, SQLStatePrm, causePrm, vendorPrm)
  { }
};

/**
 * Exception while sending/receiving data
 */
class SocketIOException : public JavaSocketException
{
public:
  /**
   * Constructs a SocketIOException with the given message
   * @param s error message of the exception
   */
   SocketIOException(std::wstring s) : JavaSocketException(s) {};
};


/**
 * Exception thrown when an attempt to convert a value (typically in a result
 * set) to another type fails
 */
class ConversionException : public DriverException
{
public:
  /**
   * Constructs a ConversionException with the given message
   * @param s error message of the exception
   * @param state corresponding SQL state, HYC00 if none
   */
  ConversionException(std::wstring s, std::wstring state)
    : DriverException(s, state)
  { };
};

/** Errors */
const std::wstring CONV_NUM_RANGE(L"22003");
const std::wstring CONV_DATE_INVALID(L"22007");
const std::wstring CONV_DATE_RANGE(L"22008");
const std::wstring CONV_INTERVAL(L"22015"); // we don't support intervals
const std::wstring CONV_CHAR(L"22018");

const std::wstring CONV_INVALID(L"07006");

/** Warnings */
const std::wstring CONV_STRING_TRUNC(L"01004");
const std::wstring CONV_NUM_TRUNC(L"01S07"); // missing from standard

/**
 * Generic exception for types or feature not implemented yet.
 */
class NotImplementedException : public DriverException
{
public:
  /**
   * Constructs a NotImplementedException with the given message
   * @param s error message of the exception
   */
  NotImplementedException(std::wstring s) : DriverException(s, L"HYC00") {};
};

/**
 * Exception on Protocol.
 */
class ProtocolException : public CarobException
{
public:
  /**
   * Constructs a ProtocolException with the given message
   * @param s error message of the exception
   */
  ProtocolException(std::wstring s) : CarobException(s) {};
};

/**
 * Exception thrown by a backend and serialized by the controller
 * over the socket
 */
class BackendException : public CarobException
{
public:
  /**
   * Constructs a BackendException with the given message
   * @param s error message of the exception
   */
  BackendException(std::wstring s) : CarobException(s) {};
  /**
   * De-serializes a BackendException from the given stream
   * @param sock driver socket from which to deserialize exception
   */
  BackendException(const DriverSocket& sock) throw (SocketIOException,
      UnexpectedException) : CarobException(sock) {};
};

/**
 * Exception thrown by the controller and received via the socket
 */
class ControllerException : public CarobException
{
public:
  /**
   * Constructs a ControllerException with the given message
   * @param s error message of the exception
   */
  ControllerException(std::wstring s) : CarobException(s) {};
  /**
   * De-serializes a ControllerException from the given stream
   * @param sock driver socket from which to deserialize exception
   */
  ControllerException(const DriverSocket& sock) : CarobException(sock) {};
};

/**
 * Not a real exception but subclasses, as in java, an Exception
 */
class SQLWarning : public CarobException
{
public:
  /**
   * Constructs a SQLWarning with the given message
   * @param messagePrm description of the warning
   * @param SQLStatePrm XOPEN code identifying the warning
   * @param causePrm exception that caused this one. Can be null
   * @param vendorPrm database vendor-specific warning code
   */
  SQLWarning(const std::wstring& messagePrm,
      const std::wstring& SQLStatePrm = DEFAULTSQLSTATE,
      CarobException* causePrm = 0,
      int vendorPrm = DEFAULTVENDORCODE) :
      CarobException(messagePrm, SQLStatePrm, causePrm, vendorPrm) {};
};

/**
 * Map of the Java stack trace element contained in an exception
 * (actually a CarobException)
 */
struct StackTraceElement
{
public:
  /**
   * Create a stacktrace element by deserialization from the given socket
   * @param ds  the socket from which to deserialize element
   * @throws SocketIOException in case of problem on the socket
   */
  StackTraceElement(const DriverSocket& ds) throw (SocketIOException,
      UnexpectedException);
  /**
   * Provides a string representation of the element.
   * @param oss stream to write to
   * @param ste stack trace element to be printed
   * @return given stream printed with all members formated in a single string
   */
  friend std::wostream& operator<< (std::wostream& oss, const StackTraceElement& ste);

private:
  // we would like to have const members (and initializer lists)
  // but we would then rely on declaration order; plus there is
  // a syntax issue with operator>>
  std::wstring  declaringClass;
  std::wstring  methodName;
  std::wstring  fileName;
  int           lineNumber;
};
} //namespace CarobNS
#endif //_CAROBEXCEPTION_H_


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
