/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */
#ifndef _COMMON_H_
#define _COMMON_H_

// This file is included by almost everyone: think before adding anything here
// (and don't think to much before _removing_ something from here).

#include <string>

#define   DEBUG_PRE  L"DEBUG:\t"
#define   INFO_PRE  L"INFO:\t"
#define   WARN_PRE  L"WARN:\t"
#define   ERROR_PRE  L"ERROR:\t"
#define   FATAL_PRE  L"FATAL:\t"

#define   WIDENLITERAL(literalstring) L ## literalstring
#define   WIDENMACRO(macro) WIDENLITERAL(macro)
#define   __WFILE__ WIDENMACRO(__FILE__)

#ifdef __MINGW32__
  #define _UNICODE
  #define _WIN32_WINNT 0x0501 // workaround to have getaddrinfo function

  typedef uint16_t in_port_t;
  typedef uint32_t in_addr_t;
#endif

#if 0 // CAROB_LOG4CXX_NAME
// This is disabled to avoid including log4cxx headers EVERYWHERE,
// and because most people don't need to do any advanced log4cxx
// tricks.  You can safely copy/paste this in your cpp carob client
// to declare this method if you need to.
// (a forward declaration of LoggerPtr does not seem possible?)
#include <log4cxx/logger.h>
namespace CarobNS
{
  log4cxx::LoggerPtr& getLogger();
}
#endif


namespace CarobNS {

/** Java byte (signed) */
typedef   int8_t    java_byte;

/**
 * Available log levels, sorted by priority (last = higher)
 */
enum LogLevel
{
  LOG_LEVEL_DEBUG,
  LOG_LEVEL_INFO,
  LOG_LEVEL_WARN,
  LOG_LEVEL_ERROR,
  LOG_LEVEL_FATAL,
  LOG_LEVEL_OFF
};

/**
 * Controller ping constants
 */
const java_byte CONTROLLER_PING_VERSION = 0x01;
/**
 * Utility class to convert a plain text log level into the internaly defined
 * log level code
 */
class StringLogLevel : public std::string
{
public:
  /**
   * Creation from a given string.
   * Accepted strings are:<ul>
   * <li><code>"DEBUG"</code>
   * <li><code>"INFO"</code>
   * <li><code>"WARN"</code>
   * <li><code>"ERROR"</code>
   * <li><code>"FATAL"</code>
   * <li><code>"OFF"</code></ul>
   * @param s string representing the desired log level
   */
  StringLogLevel(const char * s)
    : std::string(s)
  { } ; // we have nothing more than the derived std::string

  /**
   * Converter from std::string to the log level enum
   */
  operator LogLevel() const
    throw (std::string) // error message
  {
    if (0 == compare("DEBUG")) return LOG_LEVEL_DEBUG;
    if (0 == compare("INFO")) return LOG_LEVEL_INFO;
    if (0 == compare("WARN")) return LOG_LEVEL_WARN;
    if (0 == compare("ERROR")) return LOG_LEVEL_ERROR;
    if (0 == compare("FATAL")) return LOG_LEVEL_FATAL;
    if (0 == compare("OFF")) return LOG_LEVEL_OFF;

    throw (*this + " is an unknown LogLevel");
  }
};


// Line separator (EOL)
//TODO: get this at runtime => ???
const wchar_t LINE_SEPARATOR[] = 
#ifdef WIN32
  L"\r\n"
#else
  L"\n"
#endif
  ;

// Decimal point (decimal separator)
extern wchar_t userDecimalPoint;
/**
 * The controller uses the default decimal point separator for parsing numbers,
 * so we have to send him the right one, ie. the default C one.
 */
extern wchar_t controllerDecimalPoint; 


/** Sets the log level to l */
void setLogLevel(const LogLevel l);
/**
 * Writes a log message with 'Debug' level.
 * Uses isDebugEnabled() to determine if the message should be printed.
 * Cannot be used in static initializers.
 */
void logDebug(const std::wstring& fctName, const std::wstring& message);
/**
 * Writes a log message with 'Info' level.
 * Uses isInfoEnabled() to determine if the message should be printed
 * Cannot be used in static initializers.
 */
void logInfo(const std::wstring&, const std::wstring&);
/**
 * Writes a log message with 'Warn' level.
 * Uses isWarnEnabled() to determine if the message should be printed.
 * Cannot be used in static initializers.
 */
void logWarn(const std::wstring&, const std::wstring&);
/**
 * Writes a log message with 'Error' level.
 * Uses isErrorEnabled() to determine if the message should be printed.
 * Cannot be used in static initializers.
 */
void logError(const std::wstring&, const std::wstring&);
/**
 * Writes a log message with 'Fatal' level.
 * Uses isFatalEnabled() to determine if the message should be printed.
 * Cannot be used in static initializers.
 */
void logFatal(const std::wstring&, const std::wstring&);

/** Returns true if the debug level or higher is enabled */
bool isDebugEnabled();
/** Returns true if the info level or higher is enabled */
bool isInfoEnabled();
/** Returns true if the warn level or higher is enabled */
bool isWarnEnabled();
/** Returns true if the error level or higher is enabled */
bool isErrorEnabled();
/** Returns true if the fatal level or higher is enabled */
bool isFatalEnabled();

/**
 * Removes given whitespaces from the begining and end of the given string
 * @param source the string to be trimed
 * @param delims string representing the whitespaces to remove
 * @return copy of the input string without whitespaces and begining nor end
 */
std::wstring trim(const std::wstring& source, 
                         const wchar_t* delims = L" \t\r\n");


/**
 * Converts any type to wstring using C locale (locale independant). To be used
 * for internal conversions
 * Explicitly instantiated:
 *  int16_t
 *  uint16_t
 *  int32_t
 *  uint32_t
 *  int64_t
 *  uint64_t
 *  float
 *  double
 * @return a wstring representation of the given argument
 */
template <class T> std::wstring toWString(const T& t);

/**
 * Converts any type to wstring using user current locale (ie. std::locale()).
 * To be used for user-oriented info (logs, exception)
 * Explicitly instantiated: see above
 * @see toWString(const T&)
 * @return a wstring representation of the given argument
 */
template <class T> std::wstring toUserString(const T& t);

/**
 * Converts the given wstring to any given type.
 * Uses wistringstream >> operator.
 * Ignores whitespaces and eventual characters following the number.
 * eg. 123abc will be transformed to 123
 * Function found at http://www.codeguru.com/forum/showthread.php?t=231054
 * Explicitly instantiated: see above
 * @return true if the conversion succeeded, false otherwise
 */
template <class T> bool wstringTo(T& t, const std::wstring& s);

/**
 * Replace all occurence of string <code>from</code> to string <code>to</code>
 * in string <code>src</code>
 * @param src source string
 * @param from string to searched
 * @param to replacement string
 * @return the replaced string
 */
std::wstring replaceAll(std::wstring src, const std::wstring &from, const std::wstring &to);

/**
 * Utility to get the current time as number of ms since Epoch
 */
int64_t currentTimeMillis();
/**
 * Same as above, in seconds
 */
int currentTimeSeconds();
} //namespace CarobNS
#endif //_COMMON_H_


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */

