/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */
#ifndef CONTROLLER_INFO_HPP_
#define CONTROLLER_INFO_HPP_

#include "SocketAddress.hpp"
#include "CarobException.hpp"

#ifdef __MINGW32__
// NEEDED ?  #include <winsock2.h>
  #include <ws2tcpip.h>
  #include "Common.hpp" //for in_port_t
#else
// NEEDED ?  #include <sys/socket.h>
  #include <netinet/in.h> // sockaddr & in_port_t
  #include <netdb.h> // for addrinfo
#endif

#include <string>

namespace CarobNS {

class DriverSocket;
class AbstractControllerPool;

/**
 * Identifies a controller.<br>
 * This class holds the socket address to connect to and send/receive pings 
 * to/from, plus the original hostname (for debug)
 */
class ControllerInfo
{
public:
  /** Empty contructor does nothing */
  ControllerInfo() {};
  /**
   * Constructs a Controller info with the given name and port.<br>
   * This constructor is the one used at init time, by the user. It will
   * determine the socket address to connect to and to send/receive pings
   * to/from, which will be the first one returned by getaddrinfo()
   * @param host hostname as a string. Can be either qualified name or numeric
   * @param port host port
   * @throw ConnectionException if the given hostname cannot be resolved
   * */
  ControllerInfo(const std::wstring& host, in_port_t port) throw (ConnectionException, UnexpectedException);
  /**
   * Copy contructor. Calls = operator
   * @param ci original controller to copy from
   * @throw ConnectionException in case of memory allocation failure 
   */
  ControllerInfo(const ControllerInfo& ci) throw (ConnectionException, UnexpectedException);
  /**
   * Frees addrinfo
   */
  ~ControllerInfo();
  /**
   * Tries to connect to this controller and returns the connected socket, or
   * throw ConnectionException in case of failure. <b>Warning:</b> Property of
   * the created socket is transfered to caller, it is its job to delete it!
   * @param pool the controller pool that manages us, in order to register the
   * socket
   */
  DriverSocket*           connect(AbstractControllerPool& pool)
                              throw (ConnectionException, UnexpectedException);
  /**
   * Copy operator
   * @param ci controller to copy informations from
   * @throw ConnectionException in case of memory allocation failure 
   */
  ControllerInfo&         operator =(const ControllerInfo& ci);
  /** Less than operator used by ControllerPoolManager */
  bool                    operator <(const ControllerInfo& ci) const;
  /** Comparison operator, uses < operator */
  bool                    operator ==(const ControllerInfo& ci) const;
  /** Returns the socket address associated to this controller */
  const SocketAddress&    getSocketAddress() const { return socket_address; }
  /**
   * Cast to wstring operator to ease printing (for debug/traces).
   * @return sql query as a string
   */
  operator                std::wstring () const;
  /**
   * Pretty printer for controller info.<br>
   * Prints: <given/original hostname> (<resolved name/port>)
   * @param ostr the wide stream to print to
   * @param ci controller info to be pretty printed to ostr
   */ 
  friend std::wostream& CarobNS::operator<< (std::wostream& ostr, const ControllerInfo& ci);

private:
  /**
   * Retrieves all socket addresses for the controller identified by the given
   * name/port.<br>
   * Calls getaddrinfo() with a filter SOCK_STREAM filter (TCP)
   * @param name name of the controller
   * @param port port on which the controller listens to
   */
  static addrinfo*        getAddrs(const std::wstring& name, in_port_t port)
                              throw (ConnectionException, UnexpectedException);

  /** Host name as given by the user */
  std::wstring            original_hostname;
  /** The socket address to which pings and connection will be done */
  SocketAddress           socket_address;
};

std::wostream& operator<< (std::wostream& ostr, const std::vector<ControllerInfo>& clist);

} //namespace CarobNS
#endif /*CONTROLLER_INFO_HPP_*/
