/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#ifndef CONTROLLER_PING_SENDER_H_
#define CONTROLLER_PING_SENDER_H_

#include <sys/types.h> // for pthread_t

// Thread function
extern "C" void *PingSenderThread(void *);

namespace CarobNS {

class WatchedControllers;
/**
 * This thread sends a ping to all controllers in a given list at a given
 * frequency
 * 
 * @author <a href="mailto:gilles.rayrat@continuent.com">Gilles Rayrat</a>
 * @version 1.0
 */
class ControllerPingSender
{
public:
  /**
   * Creates a new <code>ControllerPingSender</code> on the given controller
   * list with the given frequency
   * 
   * @param controllerList controllers to ping. Should never be null
   * @param socket open socket to use for sending pings
   * @param pingDelayInMs time to wait between two successive pings
   */
  ControllerPingSender(WatchedControllers& controllerList,
      int socket, int pingDelayInMs);

  /**
   * Starts pinging controllers.<br>
   * Controllers will be ping with an interval of at least pingDelayInMs. But as
   * this function is synchronized on the controller list it can be slowed down
   * if controllers are added/removed, raising up the ping delay
   */
  void                run();

  /**
   * Stops the thread
   */
  void                stop() { is_stopped = true; }
private:
  // No empty or copy constructor nor = operator
  // private + no implem will forbid their use
  ControllerPingSender();
  ControllerPingSender(const ControllerPingSender&);
  ControllerPingSender& operator= (const ControllerPingSender&);
  
  /** List of controllers to send pings to */
  WatchedControllers& controllers;
  /** Socket for outgoing pings */
  int                 send_socket;
  /** Time to wait between 2 successive pings to a controller */
  int                 ping_delay_in_ms;
  /** To check for big time shifts */
  int                 max_sleep_shift;
  /** to stop the thread */
  bool                is_stopped;
};
} //namespace CarobNS
#endif /* CONTROLLER_PING_SENDER_H_*/
