/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */
#ifndef CONTROLLERPOOL_H_
#define CONTROLLERPOOL_H_

#include "ControllerInfo.hpp"
#include "CarobException.hpp"
#include "CriticalSection.hpp"
#include "Common.hpp" // for currentTimeInSeconds()
#include <vector>
#include <map>

namespace CarobNS
{

enum ConnectPolicy
{
  ROUND_ROBIN
};


/**
 * Exception on connection send when no more controller is reachable in the
 * controller list
 */
class NoMoreControllerException : public DriverException
{
public:
  /**
   * Constructs a NoMoreControllerException with the given message
   * @param s error message of the exception
   */
  NoMoreControllerException(std::wstring s) : DriverException(s) {};
};

class ControllerWatcher;
class SocketKillerCallback;
class JavaSocket;
class DriverSocket;

/** Holds a vdb member and its last failure */
class ControllerAndVdbState
{
public:
  /**
   * Constructs a <code>ControllerAndVdbState</code> object with the given
   * controller, considering last vdb failure is now minus 60 seconds (ie. no
   * failure)
   *
   * @param ctrl controller to refer to
   */
  ControllerAndVdbState(const ControllerInfo& ctrl) : controller(ctrl),
      lastVdbFailure(currentTimeSeconds() - 60) {};
  /**
   * Gives the state of the vdb. Returns true if the last vdb failure is older
   * than 5 seconds (refering to given time), false if the vdb failed during
   * the last 5 seconds
   * 
   * @param now the time to compare last vdb failure to
   * @return tru is the vdb can be considered as up, false otherwise
   */
  bool            isVdbUp(int now) { return ((now - 5) > lastVdbFailure); }

  /**
   * Gives a string representation of this object: prints controller + state
   */
  operator        std::wstring() { return static_cast<std::wstring>(controller) + L" ("
                      + (isVdbUp(currentTimeSeconds()) ? L"VDB up)" : L"VDB down)"); }
  /** Controller */
  ControllerInfo  controller;
  /** Last time the vdb has been seen failing */
  int             lastVdbFailure;

};

/**
 * Abstract class over each policy implementation, which is used by the driver
 * to choose a controller to connect to.
 */
class AbstractControllerPool
{
  // to provide access to add/removeRef()
  friend class ControllerPoolManager;

public:
  /**
   * Unique constructor
   * 
   * @param controllerList the controller list on which the policy applies
   * @param pingDelayInMs Interval in milliseconds between two pings of a
   *          controller
   * @param controllerTimeoutInMs timeout in milliseconds after which a
   *          controller is considered as dead if it did not respond to pings
   */
  AbstractControllerPool(const std::vector<ControllerInfo>& controllerList,
      int pingDelayInMs, int controllerTimeoutInMs) throw (DriverException, UnexpectedException);
  /**
   * Clears list of controllers
   */
  virtual ~AbstractControllerPool();
  /**
   * Gets a controller using the implementation specific policy
   * @return <code>ControllerInfo</code> of the selected controller
   * @throws NoMoreControllerException if no controller in the controller list
   *         is reachable
   */
  virtual ControllerInfo      getController() throw (NoMoreControllerException,
      UnexpectedException) = 0;

  /**
   * Gets the number of controllers still known to be alive
   * @return the number of alive controllers
   */
  int                         numberOfAliveControllers();

  /**
   * Registers the given socket to the callback so it can kill this socket when
   * the controller is detected as failed
   * 
   * @param controller the controller to which the socket is connected
   * @return the registered socket
   */
  DriverSocket*               newRegisteredSocket(const ControllerInfo& controller)
    throw (ConnectionException);

  /**
   * Unregisters the given socket from the callback so it can no more be killed
   * or deleted. To be used when socket creation or connection failed
   * 
   * @param socket the socket to unregister
   */
  void                        deleteRegisteredSocket(DriverSocket* socket);

  /**
   * Tell the watcher that a failure has been detected on the given controller.<br>
   * This function should be called when a connection error occurs on the given
   * controller
   * 
   * @param controller the controller suspected of failure
   */
  void                        forceControllerDown(const ControllerInfo& controller);

  /**
   * Removes this controller from the list of alive ones. <b>Warning:</b> This
   * function should never be called directly, only the callback should make us
   * of this.
   * 
   * @param controller the suspect controller
   */
  void                        controllerDown(const ControllerInfo& controller);

  /**
   * Adds the specified controller to the list of alive ones. <b>Warning:</b>
   * This function should never be called directly, only the callback should
   * make us of this.
   * 
   * @param controller the controller that came back
   */
  void                        controllerUp(const ControllerInfo& controller);

  /**
   * Informs that the given controller's vdb is no more available
   * @param controller the controller set vdb to down
   */
  void                        setVdbDownOnController(const ControllerInfo& controller);

  /**
   * Tells if the given controller VDB is available.<br>
   * 
   * @param controller the controller to check for vdb state
   * @return false if the lastVdbFailure associated to the given controller is
   *         not older than ??? milliseconds, true otherwise
   */
  bool                        isVdbUpOnController(const ControllerInfo& controller);

  /**
   * Gets a reference to the lock object used in this class
   * @see void Connection::initConnection() 
   */
  CriticalSection&            getLock() { return pool_CS; }

  /**
   * Display useful info in a string
   */
  operator                    std::wstring();
protected:
  /**
   * Tells the pool that a new connection is using it
   */
  void                        addRef() { ref_counter++; }
  /**
   * Tells the pool that one of its 'user' (ie. connection) doesn't use it
   * anymore, and return the new number of references
   */
  int                         removeRef();
  /** Retrieves the number of references left on this instance */
  int                         getRefCount() { return ref_counter; }
  /** Returns the last time the number of references was decreased to zero */
  int64_t                     getLastTimeZeroRef() { return last_time_zero_ref; }
  /** Up-to-date list of controllers that respond to pings */
  std::vector<ControllerAndVdbState> alive_controllers;
  /** Mutex on alive controllers */
  CriticalSection             pool_CS;
private:
  // Forbid copies and empty construction
  AbstractControllerPool();
  AbstractControllerPool(AbstractControllerPool&);
  AbstractControllerPool& operator=(AbstractControllerPool&);
  /** Controller watcher thread */
  ControllerWatcher*          watcher_ptr;
  /** Associated pthread */
  pthread_t                   watcher_thread;
  /** Callback for controller failures and come backs */
  SocketKillerCallback*       callback_ptr;
  /** References counter managed by controller pool manager */
  int                         ref_counter;
  /** Last time this instance saw its ref_counter decreased to zero */
  int64_t                     last_time_zero_ref;
};

/**
 * Implements a controller selection policy that iterates through the controller
 * list, one after the other, starting with the first one in the list
 */
class RoundRobinControllerPool : public AbstractControllerPool
{
public:
  /**
   * Creates a new <code>RoundRobinControllerPool</code> object
   * 
   * @param controllerList the controller list on which the policy applies
   * @param pingDelayInMs Interval in milliseconds between two pings of a
   *          controller
   * @param controllerTimeoutInMs timeout in milliseconds after which a
   *          controller is considered as dead if it did not respond to pings
   */
  RoundRobinControllerPool(const std::vector<ControllerInfo>& controllerList,
    int pingDelayInMs, int controllerTimeoutInMs) throw (DriverException, UnexpectedException);
  /**
   * @see AbstractControllerPool#getController()
   */
  ControllerInfo              getController() throw (NoMoreControllerException,
                                  UnexpectedException);
private:
  /** index of last controller given by getController */ 
  int                         index;
};

/**
 * Utility struct for controller list / controller pool association (see
 * controller pool manager
 */
typedef struct
{
  /** List of controllers (warning: order matters !) */
  std::vector<ControllerInfo> controllers;
  /** Associated Connection policy */
  ConnectPolicy               policy;
  /** Policy ping delay parameter */
  int                         ping_delay_in_ms;
  /** Policy controller timeout parameter */
  int                         controller_timeout_in_ms;
} PoolIndex;

/**
 * PoolIndex "Less Than" functor.<br>
 * First compares policies (just by comparing their enum values)<br>
 * Then their parameters (ping delay and controller timeout)<br>
 * Then the number of controllers<br>
 * Finally the controllers themselves using ControllerInfo '<' operator<br>
 * This functor is used by ControllerPoolManager for sorting values in the pool
 * map
 */
struct ltPoolIndex
{
  /**
   * Less than operator
   */
  bool operator()(const PoolIndex& pi1, const PoolIndex& pi2) const;
};

/** Number of milliseconds to wait before deleting a pool that is not used */
#define POOL_TIMEOUT_IN_MS 10000 /* 10s */
/**
 * Holds a list of controller pools, that are indexed by their policy and list 
 * of controllers. Permits to have only one instance of ControllerPool for the
 * same controllers and policy, thus to share the connection info (ie. on which
 * controller the last connection occured).<br>
 */
class ControllerPoolManager
{
public:
  /**
   * Returns a pool for the given pair controllers/policy.<br>
   * If the controller pool list already contains a pool suitable for given
   * parameters, returns it after incrementing the reference count on it. If not
   * found, creates a new controller pool, adds it to the list, increments its
   * reference counter and returns it.
   * @param ctrls (ordered) controller list to connect to
   * @param cp policy to use
   * @param pingDelayInMs Interval in milliseconds between two pings of a
   *          controller
   * @param controllerTimeoutInMs timeout in milliseconds after which a
   *          controller is considered as dead if it did not respond to pings
   * @return a pool following the given policy for the given controllers
   * distribution
   * @throws DriverException if the given policy is invalid 
   */
  static AbstractControllerPool&  getPool(std::vector<ControllerInfo>& ctrls,
                                     ConnectPolicy cp, int pingDelayInMs,
                                     int controllerTimeoutInMs)
                                     throw (DriverException, UnexpectedException);
  /**
   * Decrements reference counter on the given pool. If no more references are
   * made on this pool, removes it from the pool list and deletes it
   * @param pool the pool to release
   */
  static void                     freePool(AbstractControllerPool* pool);
protected:
  /**
   * Walks through the pool map to look for pools that can be deleted, ie.
   * a pool that is not used anymore and that been freed at least POOL_TIMEOUTms
   * ago
   */
  static void                     tryTocleanUpPools();
private:
  /** To manage pool map concurrent accesses */
  static CriticalSection          pool_map_CS;
  /** The map that associates controller lists to pools */
  static std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>  pool_map;
};

} //namespace CarobNS

#endif /*CONTROLLERPOOL_H_*/
