/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#ifndef CONTROLLERS_STATE_CHANGED_CALLBACK_H_
#define CONTROLLERS_STATE_CHANGED_CALLBACK_H_

#include "CriticalSection.hpp"
#include "ControllerInfo.hpp"

#include <map>
#include <vector>

namespace CarobNS {

class AbstractControllerPool;
class JavaSocket;

/**
 * Provides the two methods that will be called when a controller is detected as
 * failed (does not respond to pings anymore) or as back alive (responds to
 * pings again after silence). Note that the <b>processing time</b> of these
 * two methods implementation <b>is critical</b>: if needed, thread them!
 */
class ControllerStateChangedCallback
{
public:
  /**
   * What to do when a controller stops responding to pings?<br>
   * Warning: execution time is critical here, if the method's implementation is
   * too slow, thread it!
   */
  virtual void onControllerDown(const ControllerInfo& ctrl) = 0;

  /**
   * What to do when a controller responds again to pings after having been
   * detected as failed<br>
   * Warning: execution time is critical here, if the method's implementation is
   * too slow, thread it!
   */
  virtual void onControllerUp(const ControllerInfo& ctrl) = 0;

  /**
   * Mandatory destructor
   */
  virtual ~ControllerStateChangedCallback() {};
};

/**
 * Implementation of ControllerStateChangedCallback targeted for driver pings to
 * controllers.<br>
 * Holds lists of socket file descriptors associated to a controller. Upon
 * controller failure detection, this list of sockets will be killed to allow
 * driver failover
 */
class SocketKillerCallback : public ControllerStateChangedCallback
{
public:
  /** Policy is needed to inform about controller status changes */
  SocketKillerCallback(AbstractControllerPool& p) : policy(p) {}
  /** Clears constrollers and socket lists */
  ~SocketKillerCallback();
  /** Kills sockets associated to the given controller */
  void onControllerDown(const ControllerInfo& ctrl);
  /** Just informs policy that a controller came back */
  void onControllerUp(const ControllerInfo& ctrl);
  /** Adds the given socket to the list of monitored sockets for given controller */
  void registerSocket(const ControllerInfo& ctrl, JavaSocket* socketFd);
  /**
   * Removes the given socket from the list of monitored sockets for a given
   * controller. This function should be called when socket creation or
   * connection failed
   * @param ctrl controller to which given socket was connected 
   * @param socket the socket to un register
   */
  void unRegisterSocket(const ControllerInfo& ctrl, JavaSocket* socket);
private:
  /** The list of controllers with their set of sockets */
  std::map<ControllerInfo, std::vector<JavaSocket*> >  controllers_and_sockets;
  /** Policy that created us, to tell about controller state changes */
  AbstractControllerPool&  policy;
  /** controllers&sockets synchronization object */
  static CriticalSection   controllers_and_sockets_CS;
};

} //namespace CarobNS
#endif /* CONTROLLERS_STATE_CHANGED_CALLBACK_H_ */
