/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */
#ifndef CONTROLLER_WATCHER_H_
#define CONTROLLER_WATCHER_H_

#include "CarobException.hpp"

#include <sys/types.h> // for pthread_t
#include <vector>

/**
 * Function to be passed to pthread_create. Launches run on given
 * ControllerWatcher instance
 */ 
extern "C" void *ControllerWatcherThread(void *);

namespace CarobNS {

class ControllerStateChangedCallback;
class ControllerInfo;
class WatchedControllers;
class ControllerPingSender;

/**
 * Main class for controller ping monitoring.<br>
 * Launches a controller pinger thread and creates a list of controllers with
 * their states. Reads controllers responses (pongs) and updates state
 * accordingly
 */
class ControllerWatcher
{
  // This function destroy some of our members
  friend void* ::ControllerWatcherThread(void *);
public:
  /**
   * Creates a controller watcher with the given controllers to watch.<br>
   * Pings will be sent to all given controllers
   * 
   * @param controllerList controllers to ping
   * @param callbackPtr Callback implementation to call when a controller state
   *          changes
   * @param pingDelayInMs time to wait between two successive pings
   * @param controllerTimeout delay after which a controller will be considered
   *          as failing if it did not respond to pings
   */
  ControllerWatcher(const std::vector<ControllerInfo>& controllerList,
      ControllerStateChangedCallback* callbackPtr, int pingDelayInMs,
      int controllerTimeout) throw (ConnectionException, UnexpectedException);

  ~ControllerWatcher();

  /**
   * Starts watching controllers.<br>
   * Reads ping responses from controllers and updates corresponding controller
   * data in the <code>ControllerList</code>.
   */
  void                  run();
  /**
   * Stops the execution
   */
  void                  stop() { is_stopped = true; }
  /**
   * Forces a given controller to be considered as not responding
   * @param c controller that is suspected of failure 
   * @see WatchedControllers#setControllerDown(ControllerInfo&)
   */
  void                  forceControllerDown(const ControllerInfo& c);

protected:
  // These accessors are only meant to be used by the thread C function
  /** Gets the watched controllers to delete */
  WatchedControllers*   getControllersPtr() { return controllers_ptr; }
  /** Gets the pinger thread to stop him */
  pthread_t             getPingerThread() { return pinger_thread; }
  /** Gets the ControllerPingSender to delete */
  ControllerPingSender* getControllerPingSenderPtr() { return pinger_ptr; }
  /** Gets the socket file descriptor to close */
  int                   getSocketFd() { return socket_fd; }
private:
  /**
   * Reads data from socket and updates lastTimeSeen values accordingly
   */
  void                  processAnswers();
  // Forbid any copy of this object
  ControllerWatcher();
  ControllerWatcher(ControllerWatcher&);
  ControllerWatcher& operator= (ControllerWatcher&);

  /** List of controllers to send pings to */
  WatchedControllers*   controllers_ptr;
  /** Thread object that sends pings */
  ControllerPingSender* pinger_ptr;
  /** Associated pthread */
  pthread_t             pinger_thread;
  /**
   * Time to wait between 2 successive pings to a controller, also used for
   * select timeout
   */
  int                   ping_delay_in_ms;
  /**
   * Delay after which a controller will be considered as failing if it did not
   * respond to pings. Used here as a timeout for select()
   */
  int                   controller_timeout;
  /**
   * UDP socket for sending and receiving
   */
  int                   socket_fd;
  /** To stop the thread */
  bool                  is_stopped;
};

} //namespace CarobNS

#endif /*CONTROLLER_WATCHER_H_*/
