/*
 * //  Written by Phillip Sitbon
 * //  Copyright 2003
 * 
 * This code is based on Phillip Sitbon code available here:
 * http://www.codeproject.com/threads/thread_class.asp
 * 
 * It has been refactored for Carob/Sequoia needs, keeping only mutex code.
 */

#ifndef _CRITICALSECTION_H_
#define _CRITICALSECTION_H_

#include <errno.h>

#ifdef WIN32
  #ifndef _WINDOWS_
    // WIN32 Excludes
    #ifdef WIN32_LEAN_AND_MEAN
      #define VC_EXTRALEAN
      #define WIN32_LEAN_AND_MEAN
      #define _PRSHT_H_
      #define  NOGDICAPMASKS     // CC_*, LC_*, PC_*, CP_*, TC_*, RC_
      #define  NOVIRTUALKEYCODES // VK_*
      #define  NOWINMESSAGES     // WM_*, EM_*, LB_*, CB_*
      #define  NOWINSTYLES       // WS_*, CS_*, ES_*, LBS_*, SBS_*, CBS_*
      #define  NOSYSMETRICS      // SM_*
      #define  NOMENUS           // MF_*
      #define  NOICONS           // IDI_*
      #define  NOKEYSTATES       // MK_*
      #define  NOSYSCOMMANDS     // SC_*
      #define  NORASTEROPS       // Binary and Tertiary raster ops
      #define  NOSHOWWINDOW      // SW_*
      #define  OEMRESOURCE       // OEM Resource values
      #define  NOATOM            // Atom Manager routines
      #define  NOCLIPBOARD       // Clipboard routines
      #define  NOCOLOR           // Screen colors
      #define  NOCTLMGR          // Control and Dialog routines
      #define  NODRAWTEXT        // DrawText() and DT_*
      #define  NOGDI             // All GDI defines and routines
      #define  NOKERNEL          // All KERNEL defines and routines
      #define  NOUSER            // All USER defines and routines
      #define  NONLS             // All NLS defines and routines
      #define  NOMB              // MB_* and MessageBox()
      #define  NOMEMMGR          // GMEM_*, LMEM_*, GHND, LHND, associated routines
      #define  NOMETAFILE        // typedef METAFILEPICT
      #define  NOMINMAX          // Macros min(a,b) and max(a,b)
      #define  NOMSG             // typedef MSG and associated routines
      #define  NOOPENFILE        // OpenFile(), OemToAnsi, AnsiToOem, and OF_*
      #define  NOSCROLL          // SB_* and scrolling routines
      #define  NOSERVICE         // All Service Controller routines, SERVICE_ equates, etc.
      #define  NOSOUND           // Sound driver routines
      #define  NOTEXTMETRIC      // typedef TEXTMETRIC and associated routines
      #define  NOWH              // SetWindowsHook and WH_*
      #define  NOWINOFFSETS      // GWL_*, GCL_*, associated routines
      #define  NOCOMM            // COMM driver routines
      #define  NOKANJI           // Kanji support stuff.
      #define  NOHELP            // Help engine interface.
      #define  NOPROFILER        // Profiler interface.
      #define  NODEFERWINDOWPOS  // DeferWindowPos routines
      #define  NOMCX             // Modem Configuration Extensions
    #endif //WIN32_LEAN_AND_MEAN
    #include <windows.h>
  #endif //ifndef _WINDOWS_
#else //UNIX
  #include <pthread.h>
#endif //WIN32

namespace CarobNS {

/**
 * Implements a critical section for unix systems (Posix compliant) with mutex.
 * Just declare the critical section object and use Enter() and Leave()
 * functions.
 */
class CriticalSection
{
public:
  /**
   * Default constructor to be used.
   * Creates and initializes the critical section mutex
   */
  CriticalSection()
  {
#ifdef WIN32
    InitializeCriticalSection(&mutex);
#else
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&mutex,&attr);
    pthread_mutexattr_destroy(&attr);
#endif
  }
  /**
   * Unlock and destroy the mutex
   */
  virtual   ~CriticalSection()
  {
#ifdef WIN32
    DeleteCriticalSection(&mutex);
#else
    pthread_mutex_unlock(&mutex);
    pthread_mutex_destroy(&mutex);
#endif
  }
  /**
   * Locks the mutex
   */
  void       Enter() const
  {
#ifdef WIN32
    EnterCriticalSection(&mutex);
#else
    pthread_mutex_lock(&mutex);
#endif
  }
  /**
   * Unlocks the mutex
   */
  void       Leave() const
  {
#ifdef WIN32
    LeaveCriticalSection(&mutex);
#else
    pthread_mutex_unlock(&mutex);
#endif
  }

private:
#ifdef WIN32
  mutable   CRITICAL_SECTION  mutex;
#else
  /** Locking mutex */
  mutable   pthread_mutex_t   mutex;
#endif
  /** Private copy constructor to avoid copies */
  CriticalSection(const CriticalSection &cs ) {}
  /** Private operator = to avoid copies */
  void      operator=(CriticalSection &cs) {}
};

/**
 * Utility class to lock a scope by entering the given critical section at
 * creation time, and leaving it when destroyed
 * Example usage:
 * <code>
 * void fct()
 * {
 *    LockScope ls(myCriticalSectionHandle);
 *    //do what is critical
 * } //no need to do anything else, the scope locker destructor is called
 *   //automatically, releasing the lock
 * </code>
 */
class LockScope
{
public:
  /** Locks the current scope using the given CriticalSection object */
  LockScope(CriticalSection* CSPtrPrm)
  {
    CSPtr = CSPtrPrm;
    CSPtr->Enter();
  }
  ~LockScope()
  {
    CSPtr->Leave();
  }
private:
  CriticalSection *CSPtr;
};

} //namespace CarobNS
#endif //_CRITICALSECTION_H_
