/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#ifndef DRIVERRESULTSET_H_
#define DRIVERRESULTSET_H_

#include "SQLDataSerialization.hpp" // brings C99 stdint.h (intNN_t, long long etc.)
#include "CarobException.hpp"

#include <vector>
#include <locale>

namespace CarobNS {

/**
 * The constant indicating the type for a <code>ResultSet</code> object
 * whose cursor may move only forward.
 */
#define TYPE_FORWARD_ONLY 1003

/**
 * The constant indicating the type for a <code>ResultSet</code> object
 * that is scrollable but generally not sensitive to changes made by others.
 */
#define TYPE_SCROLL_INSENSITIVE 1004

/**
 * The constant indicating the type for a <code>ResultSet</code> object
 * that is scrollable and generally sensitive to changes made by others.
 */
#define TYPE_SCROLL_SENSITIVE 1005

/**
 * The constant indicating the concurrency mode for a
 * <code>ResultSet</code> object that may NOT be updated.
 */
#define CONCUR_READ_ONLY 1007

/**
 * The constant indicating the concurrency mode for a
 * <code>ResultSet</code> object that may be updated.
 */
#define CONCUR_UPDATABLE 1008

class Statement;
class Connection;
class Field;
class TypeTag;
class BigDecimal;
class SQLWarning;

/**
 * Exception thrown when client tries to retrieve a value that is NULL,
 * typically via GetString() GetInt(), etc.
 */
class NullValueException : public DriverException
{
public:
  /**
   * Constructs a NullValueException with the given message
   * @param s error message of the exception
   */
  NullValueException(std::wstring s) : DriverException(s) {};
};

/**
 * A <code>ResultSet</code> provides access to a table of data generated by
 * executing a Statement. The table rows are retrieved in sequence. Within a row
 * its column values can be accessed in any order.
 * <p>
 * A <code>ResultSet</code> maintains a cursor pointing to its current row of
 * data. Initially the cursor is positioned before the first row. The 'next'
 * method moves the cursor to the next row.
 * <p>
 * The <code>getXXX</code> methods retrieve column values for the current row.
 * You can retrieve values either using the index number of the column, or by
 * using the name of the column. In general using the column index will be more
 * efficient. Columns are numbered from 1.
 * <p>
 * For maximum portability, <code>ResultSet</code> columns within each row
 * should be read in left-to-right order and each column should be read only
 * once.
 * <p>
 * Column names used as input to <code>getXXX</code> methods are case
 * insensitive. When performing a <code>getXXX</code> using a column name, if
 * several columns have the same name, then the value of the first matching
 * column will be returned. The column name option is designed to be used when
 * column names are used in the SQL Query. For columns that are NOT explicitly
 * named in the query, it is best to use column numbers. If column names were
 * used there is no way for the programmer to guarentee that they actually refer
 * to the intended columns.
 * <P>
 * A <code>ResultSet</code> is automatically closed by the
 * <code>Statement</code> that generated it when that <code>Statement</code>
 * is closed, re-executed, or is used to retrieve the next result from a
 * sequence of multiple results.
 * <P>
 * The number, types and properties of a ResultSet's columns are provided by the
 * <code>ResultSetMetaData</code> object returned by the
 * <code>getMetaData</code> method.
 */
class DriverResultSet
{
friend class Statement;
friend class ResultSetMetaData;
public:
  /**
   * Creates a DriverResultSet by deserializing infos from the given connection
   * @param conn connection from which to deserialize information
   * @throw ProtocolException if unexpected data is read
   * @throw SocketIOException if an error occurs on the stream
   */
	DriverResultSet(Connection* conn) throw (ProtocolException, SocketIOException,
    NotImplementedException, UnexpectedException);
  /** Frees dynamically allocated fields */
	virtual ~DriverResultSet();
  /**
   * Cast to wstring. Handy for displaying the result set infos
   */  
  operator                    std::wstring();
  /**
   * A ResultSet is initially positioned before its first row, the first call to
   * next makes the first row the current row; the second call makes the second
   * row the current row, etc.
   * <p>
   * If an input stream from the previous row is open, it is implicitly closed.
   * The ResultSet's warning chain is cleared when a new row is read
   * 
   * @return <code>true</code> if the new current is valid; <code>false</code>
   *         if there are no more rows
   */
  bool                        next() throw (SocketIOException,
                                  ControllerException, ProtocolException,
                                  BackendException, UnexpectedException);

  /**
   * Gets the string value of a column in the current row only if the given
   * column is of string type. Throws an exception otherwise. To get the value
   * as a string anyway, use #getAsString(int, const std::locale&)
   * 
   * @param columnIndex the first column is 1, the second is 2...
   * @return the column value, null for SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is
   *            out of bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  std::wstring                getString(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as a wstring with the
   * given locale. If the value at the given row/col is not of type string,
   * tries to convert it using STL std::wostringstream conversion functions
   * 
   * @param columnIndex the first column is 1, the second is 2...
   * @param loc (optional) the locale to be used for string conversions. Ignored
   *            if the column is of type STRING
   * @return the column value, null for SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is
   *            out of bounds
   * @throw NullValueException if the retrieved value is NULL
   */
  std::wstring                getAsString(int columnIndex, const std::locale& loc = std::locale())
                                  throw (DriverException, NullValueException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Gets the 32bits integer value of a column in the current row only if the
   * given column is of int type, throws an exception otherwise. To get the
   * value as an int anyway, use #getAsInt(int)
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  int32_t                     getInt32(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as an int. If the
   * value at the given row/col is not of type int, tries to convert it using
   * wstringToA conversion function.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw NullValueException if the retrieved value is NULL
   * @throw ConversionException if the retrieved value cannot be converted to
   *        a float
   */
  int                         getAsInt(int columnIndex) throw (DriverException,
                                  NullValueException, ConversionException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Gets the 64bits integer value of a column in the current row only if the
   * given column is of int type, throws an exception otherwise. To get the
   * value as an int64 anyway, use #getAsInt64(int)
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  int64_t                     getInt64(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as an int64. If the
   * value at the given row/col is not of type long, tries to convert it using
   * wstringToA conversion function.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw NullValueException if the retrieved value is NULL
   * @throw ConversionException if the retrieved value cannot be converted to
   *        an int64
   */
  int64_t                     getAsInt64(int columnIndex) throw (DriverException,
                                  NullValueException, ConversionException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as an uint64. If the
   * value at the given row/col is not of type long, tries to convert it using
   * wstringToA conversion function.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw NullValueException if the retrieved value is NULL
   * @throw ConversionException if the retrieved value cannot be converted to
   *        an unsigned int64
   */
  uint64_t                     getAsUInt64(int columnIndex) throw (DriverException,
                                  NullValueException, ConversionException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Gets the float value of a column in the current row only if the
   * given column is of float type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  float                       getFloat(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as a
   * float. If the value at the given row/col is not of type float or
   * double, miserably fails TODO.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw NullValueException if the retrieved value is NULL
   * @throw ConversionException if the retrieved value cannot be converted to
   *        a float
   */
  float                       getAsFloat(int columnIndex) throw (DriverException,
                                  NullValueException, ConversionException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Gets the double value of a column in the current row only if the
   * given column is of double type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  double                      getDouble(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
  /**
   * Tries to get the value of a column in the current row as a double.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value converted into a double
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw NullValueException if the retrieved value is NULL
   * @throw ConversionException if the retrieved value cannot be converted to
   *        a double
   */
  double                      getAsDouble(int columnIndex) throw (DriverException,
                                  NullValueException, ConversionException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Gets the timestamp value of a column in the current row only if the
   * given column is of timestamp type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  SQLTimeStamp                getTimeStamp(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);

  /**
   * Gets the blob data value of a column in the current row only if the
   * given column is of blob type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  LargeData&                  getBlob(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException)
                                      { return getLargeObject(columnIndex); }

  /**
   * Gets the clob data value of a column in the current row only if the
   * given column is of blob type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  LargeData&                  getClob(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException)
                                      { return getLargeObject(columnIndex); }

  /**
   * Gets the byte array value of a column in the current row only if the
   * given column is of byte array type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  LargeData&                  getByteArray(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException)
                                      { return getLargeObject(columnIndex); }

  /**
   * Gets the bigdecimal value of a column in the current row only if the
   * given column is of bigdecimal type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  const BigDecimal&           getBigDecimal(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);

  /**
   * Closes the remote ResultSet if the ResultSet was streamed else just closes
   * the ResultSet locally.
   */
  void                        close() throw (SocketIOException,
                                  BackendException, ControllerException,
                                  ProtocolException, UnexpectedException);
  /**
   * Gets the number of column in this resultSet
   * @return number of columns value
   */
  int                         getNumberOfColumns() { return nbOfColumns; }
  /**
   * Gets the number of rows that have been fetched yet.
   * Note: this is not the number of rows of the ResultSet, but the current
   * number of rows in memory !
   * @return number of rows fetched until now
   */
  int                         getNumberOfRowsInMemory() { return nbOfRows; }
  /**
   * A column may have the value of SQL NULL; wasNull() reports whether the
   * last column read had this special value.
   * Note that you must first call getXXX on a column to try to read its value
   * and then call wasNull() to find if the value was SQL NULL
   * If you need to know if a value is NULL beforehand, use isNull(int)
   * 
   * @return <code>true</code> if the last column read was SQL NULL
   * @see #isNull(int) const
   */
  bool                        wasNull() const { return wasNullFlag; }
  /**
   * A column may have the value of SQL NULL; isNull(int) reports whether
   * the value at the given column and current row is NULL.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return <code>true</code> if the current row's value for the given column
   *         is SQL NULL
   * @exception DriverException if the ResultSet is closed or the cursor is
   *            out of bounds
   */
  bool                        isNull(int columnIndex) const
                                  throw (DriverException, UnexpectedException);
  /**
   * Gets the SQL warnings associated to this ResultSet
   * @return a SQL warning chain
   * @throw DriverException if the ResultSet is closed
   */
  SQLWarning*                 getWarnings() throw (DriverException, UnexpectedException);
  /**
   * Sets the owning statement warnings. Only fills statementWarnings field, 
   * does *not* set the owningStatement->warnings
   * @param stWarns the statement warnings to set
   */
  void                        setStatementWarnings(SQLWarning* stWarns)
  { statement_warnings_ptr = stWarns; }
  /**
   * Retrieves the warnings linked to the owning statement
   * @return warnings on owning statement
   */
  SQLWarning*                 getStatementWarnings() { return statement_warnings_ptr; }
protected:
  /**
   * De-serialize only data rows, not any metadata.
   * 
   * @param socket opened socket to controller
   * @throw SocketIOException stream error
   * @throw ProtocolException protocol corruption
   */
  void                        receiveRows(const DriverSocket& socket)
                                  throw (SocketIOException, ProtocolException,
                                  UnexpectedException);
  /**
   * Sets the statement.
   * 
   * @param stmtPtr pointer to the statement to set
   */
  void                        setStatement(Statement* stmtPtr);
private:
  /** Cursor to current row */
  int                         currentRow;
   /** Number of rows */
  int                         nbOfRows;
   /** Number of columns */
  int                         nbOfColumns;
  /** The results */
  std::vector< std::vector<ResultSetDataType>* >  data;
  /** Indicates that no data has been read yet */
  bool                        dataIsNull;
  /**
   * Matrix for null results.
   * Should be optimized with a bitfield array.
   * bitfield class would be used as follows:
   * if (bitfield[n])...
   * and implemented like
   * operator[] => switch (n%8) { case 0: return bit_struct.b0;
   *                           ...case x: return bit_struct.bx;
   * bit_struct would be: typedef struct { bool b0:1; bool b1:1; ...}
   */
  std::vector< std::vector<bool>* > nulls;
  /** Type of columns as defined by the controller */
  std::vector<TypeTag>        columnTypeTags;
  /** True if there is more data to fetch from the controller */
  bool                        hasMoreData;
  /** The fetch size */
  int                         fetchSize;
  /** ResultSet cursor name */
  std::wstring                cursorName;
  /** The fields */
  std::vector<Field*>         fields;
  /** Pointers to column-specific deserializer functions */
  std::vector<deserializerPtr> deserializers;
  /** for wasNull() */
  bool                        wasNullFlag;
//TODO:
  /** column name -> index in ResultSet data array */
//  hash_map<wstring, int>  columnNameToIndex;
  /** Type of ResultSet */
  int                         resultSetType;
  /** Concurrency for this ResultSet */
  int                         resultSetConcurrency;

  /** Warning chain */
  SQLWarning*                 warningsPtr;
  /** Warnings on owning statement */
  SQLWarning*                 statement_warnings_ptr;
  /** Statement corresponding to this ResultSet, if any (not for metadata) */
  Statement*                  owningStatementPtr;
  /** The driver connection we were received from. Useful for streaming */
  Connection*                 connectionPtr;

  bool                        isClosed;

  /**
   * Checks that the cursor is on row in the ResultSet
   * @throw DriverException if the cursor is out of bounds
   */
  void                        checkRowPos() const throw (DriverException,
                                  UnexpectedException);
  /**
   * Check if the ResultSet if closed and throws a DriverException if so.
   * @throw DriverException if the ResultSet is closed
   */
  void                        checkIfClosed() const throw (DriverException,
                                  UnexpectedException);
  /**
   * Sanity checks for result parsing
   * 
   * @param columnIndex the column to check
   * @exception DriverException if the ResultSet is closed or the cursor is
   *            out of bounds
   */
  void                        checkRowAndColPosAndSetNullFlag(int columnIndex)
                                  throw (DriverException, UnexpectedException);
  /** Clear/delete data, nulls and typeTags vectors */
  void                        clearAllData();

  /**
   * Gets the large data value of a column in the current row only if the
   * given column is of blob, clob or byte array type, throws an exception otherwise.
   * 
   * @param columnIndex the first column is 1, the second is 2,...
   * @return the column value; NULL if SQL NULL
   * @throw DriverException if the ResultSet is closed or the cursor is out of
   *        bounds
   * @throw DriverException if the column is not of type string
   * @throw NullValueException if the retrieved value is NULL
   */
  LargeData&                  getLargeObject(int columnIndex) throw (DriverException,
                                  NullValueException, NotImplementedException,
                                  UnexpectedException);
};

/**
 * Structure for storing the result of an execute() command.
 * This result is a list of either ResultSet or updateCount. So the result
 * representation will be a list of \<this structure\> elements
 */
typedef struct
{
  /** Whether the value is a ResultSet (true) or an updateCount (false) */
  bool isResultSet;
  /** Value itself. Can be either a DriverResultSet or an updateCount */
  union
  {
    /** Number of column concerned by the update */
    int                       updateCount;
    /** Pointer (to avoid copy of large objects) to a DriverResultSet object */
    DriverResultSet*          resultSetPtr;
  };
} ResultSetOrUpdateCount;

} //namespace CarobNS
#endif /*DRIVERRESULTSET_H_*/
