/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#ifndef _DRIVERSOCKET_H_
#define _DRIVERSOCKET_H_

#include "ControllerInfo.hpp"

#include "JavaSocket.hpp"
#include "BufferedSocket.hpp"

namespace CarobNS {

class ControllerInfo;

const int STRING_CHUNK_SIZE = 64000/3;

/**
 * Higher level class for socket.
 * Defines handy methods to read/write to/from a controller following
 * the controller/driver communication protocol.
 */
class DriverSocket
#ifdef CAROB_UNBUFFERED_SOCKET
  : public JavaSocket
#else
  : public BufferedSocket
#endif
{

  // to avoid missing (de-)registrations, only the pool is allowed to
  // create/delete us. See CAROB-133.
  friend class AbstractControllerPool;
private:
  /**
   * Empty constructor
   * @throws ConnectionException if the connection fails
   */
  DriverSocket(const ControllerInfo& ctrl) throw (CodecException, UnexpectedException)
#ifdef CAROB_UNBUFFERED_SOCKET
    : JavaSocket()
#else
    : BufferedSocket()
#endif
    , controller(ctrl)
  {}

  ~DriverSocket() { };

public:
  /** The controller we are (trying to) be connected to. */
  const ControllerInfo controller;
  // Hold our own (and cheap) copy to avoid thinking about lifecycles
  /**
   * Writes a string to the socket according to the controller protocol
   * FIXME: we don't support string longer than 64K once UTF8 encoded.
   * @throws SocketIOException
   */
  const DriverSocket & operator<<(const std::wstring &) const 
      throw (SocketIOException, UnexpectedException);

  /**
   * Convenience method to send C-style strings. Creates a wstring and sends it.
   * Reminder: every C-style string can be a wstring; the opposite is not true.
   */
  const DriverSocket& operator <<(const wchar_t *) const
    throw (SocketIOException, UnexpectedException);
    
  /**
   * Reads a string from the socket according to the controller protocol
   * FIXME: we don't support string longer than 64K once UTF8 encoded.
   * @throws SocketIOException
   */
	const DriverSocket & operator>>(std::wstring &) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Writes a 32-bits int to the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator<<(const int32_t&) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Reads a 32-bits int from the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator>>(int32_t& i) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Writes a 64-bits long to the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator<<(const int64_t&) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Reads a 64-bits long from the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator>>(int64_t& i) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Writes a boolean to the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator<<(const bool &) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Reads a boolean from the socket
   * @throws SocketIOException
   */
  const DriverSocket & operator>>(bool& i) const 
      throw (SocketIOException, UnexpectedException);
  /**
   * Reads given number of bytes from the socket into the given array
   * @throws SocketIOException
   */
  void readBytes(int32_t length, java_byte* data) const
      throw (SocketIOException, UnexpectedException);
  /**
   * Writes given number of bytes from the socket
   * @throws SocketIOException
   */
  void writeBytes(int32_t length, java_byte* data) const
      throw (SocketIOException, UnexpectedException);
};

} //namespace CarobNS
#endif //_DRIVERSOCKET_H_
