/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Csaba Simon
 */

#ifndef FIELD_H_
#define FIELD_H_

#include "CarobException.hpp"

#include <string>

namespace CarobNS {

class DriverSocket;

/**
 * Field is our private implementation of <code>ResultSetMetaData</code>,
 * holding the information for one column.
 */
class Field
{
public:
  /**
   * Creates a new <code>Field</code> object, deserialized from an input
   * socket.
   * 
   * @param sock the socket
   * @throws SocketIOException if a stream error occurs
   */
	Field(const DriverSocket& sock) throw (SocketIOException,
      UnexpectedException);
	virtual ~Field() {};

  /**
   * Returns the fieldLabel value.
   * @return the fieldLabel
   */
  std::wstring getFieldLabel() const { return field_label; }
  /**
   * Gets the field name.
   * @return a <code>wstring</code> containing the field name
   */
  std::wstring getFieldName() const { return field_name; }
  /**
   * Gets the full name: "tableName.fieldName"
   * @return a <code>String</code> containing the full field name
   */
  std::wstring getFullName() { return table_name + L"." + field_name; }
  /**
   * Get the number of decimal digits.
   * @return precision
   */
  int getPrecision() { return precision; }
  /**
   * Gets the number of digits to right of the decimal point.
   * @return scale
   */
  int getScale() { return scale; }
  /**
   * Returns the JDBC type code.
   * 
   * @return int Type according to java.sql.Types, reflected in
   * TypeConstants.hpp
   * @see TypeConstants.hpp
   */
  int getSqlType() { return sql_type; }
  /**
   * Gets the table name.
   * @return table name
   */
  std::wstring getTableName() { return table_name; }
  /**
   * Retrieves the database-specific type name.
   * @return SQL type name
   */
  std::wstring getTypeName() { return type_name; }
  /**
   * Returns the Java class used by the mapping.
   * Not very meanfull for carob, but keeps java driver code reflexion
   */
  std::wstring getColumnClassName() { return column_class_name; }
  /**
   * Indicates the normal maximum width in characters.
   * @return column display size
   */
  int getColumnDisplaySize() { return column_display_size; }
  /**
   * Indicates whether the column (field) is automatically numbered, thus
   * read-only.
   */
  bool isAutoIncrement() { return is_auto_increment; }
  /**
   * Indicates whether a column's case matters.
   * @return true if case sensitive
   */
  bool isCaseSensitive() { return is_case_sensitive; }
  /**
   * Indicates whether the column is a cash value.
   * @return true if the column is a currency
   */
  bool isCurrency() { return is_currency; }
  /**
   * Indicates whether a write will definitely succeed.
   * @return definitely writable
   */
  bool isDefinitelyWritable() { return is_definitely_writable; }
  /**
   * Indicates the nullability of values
   * @return the nullability status of the given column; one of 
   * <code>columnNoNulls</code>, <code>columnNullable</code> or 
   * <code>columnNullableUnknown</code>
   * @see ResultSetMetadata.hpp
   */
  int isNullable() { return is_nullable; }
  /**
   * Indicates whether the column is definitely not writable.
   * @return is readonly
   */
  bool isReadOnly() { return is_read_only; }
  /**
   * Indicates whether the column can be used in a where clause.
   * @return is searchable
   */
  bool isSearchable() { return is_searchable; }
  /**
   * Indicates whether values in the column are signed numbers.
   * @return is signed
   */
  bool isSigned() { return is_signed; }
  /**
   * Indicates whether it is possible for a write on the column to succeed.
   * @return is writable
   */
  bool isWritable() { return is_writable; }

  /**
   * Returns the full name of the column
   * @return full name
   * @see #getFullName()
   */
  std::wstring toString() { return getFullName(); }

  /**
   * Retrieves the encoding of the field.
   * @return encoding
   */
  std::wstring getEncoding() { return encoding; }
  
private:
  std::wstring        table_name;
  bool                table_name_not_null;
  std::wstring        field_name;
  std::wstring        field_label;
  int                 column_display_size;
  int                 sql_type;
  std::wstring        type_name;
  std::wstring        column_class_name;
  bool                is_auto_increment;
  bool                is_case_sensitive;
  bool                is_currency;
  int                 is_nullable;
  bool                is_read_only;
  bool                is_writable;
  bool                is_definitely_writable;
  bool                is_searchable;
  bool                is_signed;
  int                 precision;
  int                 scale;
  std::wstring        encoding;
};
typedef Field  *FieldPtr;

} //namespace CarobNS
#endif /*FIELD_H_*/
