/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon, Marc Herbert
 */

#ifndef _JAVA_SOCKET_H_
#define _JAVA_SOCKET_H_

#include "SocketAddress.hpp"
#include "StringCodecs.hpp"
#include "CarobException.hpp"

#include "Common.hpp" // for java_byte: cannot forward declare typedefs

#include <string>

#ifdef __MINGW32__
  #include <winsock2.h>
#else
  #include <netinet/in.h> //for in_port_t
#endif

namespace CarobNS {

/**
 * Generic socket wrapper class to exchange data with Java world.
 * This class can read and write basic Java types over a socket. It 
 * does *not* follow any protocol but java pure serialization.
 * @see DriverSocket class for sequoia protocol.
 */
class JavaSocket
{
public:
  /**
   * Gives socket validity status.
   * @return true if the socket is valid (has been created) so it can be
   * connected
   */
  inline bool   isValid() const { return socket_fd != -1; }
  /**
   * Gives connection status
   * @return true if the socket is connected to a host
   */
  inline bool   isConnected() const { return connected; }

  /**
   * Creates the socket.
   * @param domain communication domain (supported = PF_INET or PF_INET6)
   * @throws ConnectionException if the socket could not be create or its
   * options could not be set
   */
  void          create(int domain)
                    throw (ConnectionException, UnexpectedException);
  /**
   * Connects to the given socket address
   * @param addr socket address of the host to connect to
   * @return true upon successfull connection, false otherwise
   * @throws ConnectionException
   */
  bool          connectTo(const SocketAddress& addr)
                    throw (ConnectionException, UnexpectedException);
  /** 
   * Closes TCP socket. Why public?
   * @return true if the socket was successfully closed
   * @throws SocketIOException
   */
  virtual bool  closeSystemSocket() throw (SocketIOException, UnexpectedException);
  
  /**
   * Writes a UTF-8 encoded wstring to the socket. This is NOT Java's
   * modified UTF-8, see SEQUOIA-133.
   * @param str string to write
   * @return the number of bytes send
   * @throws SocketIOException
   */
  size_t        writeJavaUTF(const std::wstring& str) const
    throw (SocketIOException, CodecException, UnexpectedException);
  /**
   * Reads UTF-8 encoded string from socket. This is NOT Java's
   * modified UTF-8, see SEQUOIA-133.
   * @param str string to read
   * @return the length read in bytes
   * @throws SocketIOException
   */
  size_t        readJavaUTF(std::wstring& str) const
    throw (SocketIOException, CodecException, UnexpectedException);
  /**
   * Writes 32bits-integer to socket
   * @param i integer to send
   * @throws SocketIOException
   */
  void          writeJavaInt(int32_t i) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads 32bits-integer from socket
   * @param i resulting integer read from stream
   * @throws SocketIOException
   */
  void          readJavaInt(int32_t& i) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Writes 64bits-long to socket
   * @param l long to send
   * @throws SocketIOException
   */
  void          writeJavaLong(int64_t l) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads 64bits-long from socket
   * @param l resulting long read from stream
   * @throws SocketIOException
   */
  void          readJavaLong(int64_t& l) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Writes a boolean (as a byte) to the socket
   * @param b the boolean to send
   * @throws SocketIOException
   */
  void          writeJavaBool(bool b) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads a boolean (byte) from the socket
   * @param b resulting boolean to read from stream
   * @throws SocketIOException
   */
  void          readJavaBool(bool& b) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Reads an array of bytes from the socket. The array must be allocated !
   * @param length number of bytes to read
   * @param data allocated array of (at least) length bytes
   * @throws SocketIOException
   */
  void          readJavaBytes(int32_t length, java_byte* data) const
                    throw (SocketIOException, UnexpectedException);
  /**
   * Writes an array of bytes to the socket.
   * @param length number of bytes to write
   * @param data array of (at least) length bytes
   * @throws SocketIOException
   */
  void          writeJavaBytes(int32_t length, java_byte* data) const
                    throw (SocketIOException, UnexpectedException);
  /**
   * Interrupts read operations on this socket, mark it as disconnected and
   * makes read operations throw a SocketIOException
   * @see #recvFully(void *buf, const int len, const int flags) const
   */
  void          shutdown() { canceled = true; }
  /**
   * Execute poll (or select depending on compilation flag CAROB_USE_SELECT) on
   * the given file descriptor and return after either data is ready to be read
   * or given timeout is elapsed
   * @param socketFd socket file descriptor on which to wait for incomming data
   * @param pollTimeoutInMs timeout after which to return if no data is ready to
   *                        be read on socket
   * @param pollOnWrites true makes poll on write (false by default = on read)
   * @return >0 if data is ready to be read, 0 if timeout reached, -1 if an
   *         error occured
   */
  static int    pollOnSingleFd(int socketFd, int pollTimeoutInMs, bool pollOnWrites = false);
protected:
  /** Socket file descriptor */
  int                         socket_fd;
  /** true if the socket is connected to a host */
  bool                        connected;
  /** true cancels all input and connections attempts */
  bool                        canceled;
  /**
   * Default constructor - No connection here, just creates and empty socket
   */
  JavaSocket() throw (CodecException);
  /**
   * Destructor for disconnection - closes the socket.
   * @throws SocketIOException
   */
  virtual ~JavaSocket();
  /**
   * Substitute for recv. Waits for incomming data by calling pollOnSingleFd
   * and loops until full length has been received or an erro occured. If
   * shutdown() is called during the loop, throws a SocketIOException to inform
   * callers that the socket is not longer readable
   * @param buf data to send
   * @param len full buffer length
   * @param flags send options, see recv man page
   * @return the total number of bytes send, -1 in case of failure
   * @throw SocketIOException if interrupted by shutdown() function
   */
  virtual int   recvFully(void *buf, const int len, const int flags)
                    const throw (SocketIOException, UnexpectedException);
  /**
   * Wrapper over send(...) function to handle errors and throw exceptions
   *
   * @param fctName name of the calling function (for logging purposes)
   * @param buf data to be send
   * @param len length of buf
   * @param flags send option, see recv man page
   * @throws SocketIOException
   */
  virtual void  sendToSocket(const wchar_t fctName[],
                    const void* buf, int len,
                    int flags) const throw (SocketIOException,
                    UnexpectedException);
  /**
   * Wrapper around recv(...) function to handle errors and throw exceptions
   *
   * @param fctName name of the calling function (for logging purposes)
   * @param buf buffer in which to put the received data in
   * @param len length of buf
   * @param flags send option, see send man page
   * @throws SocketIOException in case of error
   */
  void          receiveFromSocket(const wchar_t fctName[],
                    void *buf, int len, int flags)
                    const throw (SocketIOException, UnexpectedException);

private:
  const CarobNS::StringCodec& utf8_codec;
  /**
   * Sets the socket to desired blocking mode
   * 
   * @param fctName calling function name to display debug info
   * @param blocking true to set to blocking mode, false to non-blocking
   * @throw ConnectionException if the operation fails
   */
  void          setBlockingMode(const wchar_t fctName[], bool blocking)
                    throw (ConnectionException, UnexpectedException);
};

} //namespace CarobNS
#endif /*_JAVA_SOCKET_H_*/
