/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon
 * Contributor(s): Gilles Rayrat, Marc Herbert
 */
#ifndef PARAMETERSTATEMENT_HPP_
#define PARAMETERSTATEMENT_HPP_

#include "Statement.hpp"

#include "CarobException.hpp"

#include <string>
#include <vector>

namespace CarobNS {

class Connection;
class BigDecimal;

/**
 * A SQL Statement template is stored in a
 * <code>ParameterStatement</code> object. This object can then be used to
 * efficiently execute this statement multiple times with different parameters.
 * <p>
 * <b>Note: </b> The setXXX methods for setting IN parameter values must specify
 * types that are compatible with the defined SQL type of the input parameter.
 * For instance, if the IN parameter has SQL type Integer, then setInt should be
 * used.
 * open one exists.
 */
class ParameterStatement: public Statement
{
friend class Connection;

public:
  //No constructor nor destructor, statements must be created and destroyed by
  //connections

  /**
   * Release objects and call Statement.close().
   */
  virtual void            close();

  /**
   * Call Statement.close(). Will not release the objects.
   * Call this member function only if you know what are you doing.
   */
  virtual void            closeStatement();

  /**
   * Execute a SQL statement that may return multiple results.
   * @return true if the result is a ResultSet or false if it is an integer
   * @throw DriverException if the statement is closed
   */
  bool                    execute() throw (DriverException, SocketIOException, 
                              BackendException, ControllerException, ProtocolException,
                              NotImplementedException, UnexpectedException)
                          { return Statement::execute(sql_request, compileQuery()); }

  /**
   * Executes a SQL statement that returns a single ResultSet
   * @return a ResulSet pointer that contains the data produced by the query
   * @throw DriverException if the statement is closed
   */
  DriverResultSet *       executeQuery() throw (DriverException, SocketIOException,
                              BackendException, ControllerException,
                              ProtocolException, UnexpectedException)
                          { return Statement::executeQuery(sql_request, compileQuery()); }
  /**
   * Execute a SQL INSERT, UPDATE or DELETE statement. In addition SQL
   * statements that return nothing such as SQL DDL statements can be executed
   * 
   * @return either a row count, or 0 for SQL commands
   */
  int                     executeUpdate() throw (DriverException, SocketIOException,
                              BackendException, ControllerException,
                              ProtocolException, UnexpectedException)
                          { return Statement::executeUpdate(sql_request, compileQuery()); }

  /**
   * Sets a parameter to SQL NULL.
   * @param parameterIndex the first parameter is 1, etc...
   * @param sqlType the SQL type code defined in TypeConstants. <b>Note:</b>
   *          For <b>some</b> databases this sqlType can be a dummy one (eg. you
   *          can always give type "SQLT_SMALLINT")
   */
  void                    setNull(const int parameterIndex, const int sqlType)
                              throw (DriverException, UnexpectedException);

  /**
   * This JDBC call will be forwarded to the JDBC driver of the
   * backend(s) of a sequoia controller. Some type conversions of the
   * value may happen on the way (at different places) when called
   * with a argument of a type different from the corresponding java
   * type.
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setBoolean(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setByte(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setShort(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setInt(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setLong(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setFloat(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setDouble(int parameterIndex, const CParamType& x)
                              throw (DriverException, UnexpectedException);

  /**
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setBigDecimal(int parameterIndex, const CParamType &x)
                              throw (DriverException, UnexpectedException);

  /**
   *
   * Warning:
   *   setString                    (i, L"foo");
   *
   * does not compile with gcc 4.0.2 because it searches the following instantiation:
   *   setString <      wchar_t *>  (int parameterIndex, const wchar_t * const& x);
   *
   * which cannot exist. So you have to invoke the pointed-to-const instance:
   *   setString <const wchar_t *>  (int parameterIndex, const wchar_t * const& x);
   *
   * using a explicit instantiation like this:
   *   setString <const wchar_t *>  (i, L"foo");
   *
   * Or you can use this std::wstring alternative:
   *   setString                    (i, std::wstring(L"foo");
   *
   * @see #setBoolean()
   *
   * @param parameterIndex the first parameter is 1...
   * @param x the parameter value
   */
  template <typename CParamType>
  void                    setString(int parameterIndex, const CParamType &x)
                              throw (DriverException, UnexpectedException);

  /**
   *
   * This JDBC call will be forwarded to the JDBC driver of the
   * backend(s) of a sequoia controller. It will encode the buffer
   * to Base64 after that will set the parameter as a string.
   *
   * @param parameterIndex the first parameter is 1...
   * @param buffer the parameter value
   * @param len length of the buffer
   */
  void                    setByteArray(int parameterIndex, const void *buffer, size_t len)
                              throw (DriverException, UnexpectedException);

  /**
   *
   * This JDBC call will be forwarded to the JDBC driver of the
   * backend(s) of a sequoia controller. It will encode the buffer
   * to Base64 after that will set the parameter as a string.
   *
   * @param parameterIndex the first parameter is 1...
   * @param buffer the parameter value
   * @param len length of the buffer
   */
  void                    setBlob(int parameterIndex, const void *buffer, size_t len)
                              throw (DriverException, UnexpectedException);

  /**
   * In general, parameter values remain in force for repeated used of a
   * <code>Statement</code>. Setting a parameter value automatically clears
   * its previous value. However, in coms cases, it is useful to immediately
   * release the resources used by the current parameter values; this can be
   * done by calling <code>clearParameters()</code>.
   */
  void                    clearParameters();

  /**
   * Returns the SQL statement with the current template values substituted.
   * <p>
   * <b>Note: </b>: This is identical to <code>compileQuery()</code> except
   * instead of throwing SQLException if a parameter is <code>null</code>, it
   * places ? instead.
   * @return the SQL statement
   */
  operator                std::wstring()
                              { return compileQuery(false); }

  /**
   * Returns the parameter count.
   * @return the number of <code>?</code> in the statement
   */
  int                     getParameterCount() { return inStrings.size(); }

protected:
  /**
   * Forbid construction to the rest of the world.
   * ParameterStatements must be created by connection.
   * 
   * @param c the instantiating connection
   * @param sqlStatement the SQL statement with ? for IN markers
   */
  ParameterStatement(Connection *c, const std::wstring &sqlStatement);

  /**
   * Actually stores the IN parameter into parameters String array. Called by
   * most setXXX() methods.
   * @param paramIndex the index into the inString
   * @param typeTag the JDBC call that will be used by the controller
   * @param value the parameter
   * @exception DriverException if something goes wrong
   */
  template <class T> void     set(const int paramIndex, const std::wstring &typeTag, 
                                  const T& value) 
                              throw (DriverException, UnexpectedException);

  /**
   * Forbid destruction to the rest of the world.
   * ParameterStatements must be deleted by connection.
   */
  virtual ~ParameterStatement();
private:
  /** IN parameters, ready to be sent with the request */
  std::vector<std::wstring>   inStrings;

  /**
   * Helper - this compiles the SQL query, serializing the parameters
   * plus appropriate escape tags into the string returned. It throws
   * an exception if a parameter was not set.
   * 
   * @return the compiled query
   * @throws DriverException if a parameter was not set
   */
  std::wstring            compileQuery() throw (DriverException)
                              { return compileQuery(true); }

  /**
   * Helper - this serializes the parameters into a unique wstring.
   *
   * If a parameter was not set, throws a NullValueException if throwException is true,
   * otherwise replace the missing parameter with ?
   * @param throwException to throw a NullValueException if a parameter is missing
   * @return the compiled query
   * @throws DriverException if a parameter was not set
   */
  std::wstring                compileQuery(bool throwException) throw (DriverException);

  /**
   * Helper - this encode a byte of characters in base64.
   * @param throwException to throw a NullValueException if a parameter is missing
   * @return the encoded string
   */
  std::wstring                base64Encode(const void *buffer, size_t len);
};

} //namespace CarobNS
#endif /*PARAMETERSTATEMENT_HPP_*/


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
