/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */

#ifndef REQUEST_HPP_
#define REQUEST_HPP_

#include "CarobException.hpp"

#include <string>

namespace CarobNS {

class DriverSocket;

/**
 * Defines a Request object, a private implementation of [Parameter]Statement
 * class that holds and handles sql requests.<br>
 * It basically carries the SQL statement and the SQL template if this is a
 * ParameterStatement.
 */
class Request
{
// We are private implementation of these two classes. So it is ok to make them
// friends
friend class Statement;
friend class ParameterStatement;
// Contructor and Destructor are protected to avoid access by anyone else than
// our friends
public:
  /**
   * Serializes the request on the output stream by sending only the needed
   * parameters to reconstruct it on the controller.
   * 
   * @param out destination socket to controller
   * @throws SocketIOException an error occurs on the stream
   */
  void              sendToStream(const DriverSocket& out) const
                        throw (SocketIOException, UnexpectedException);
  /**
   * Returns the escapeProcessing value.
   * @return the escapeProcessing.
   */
  bool              isEscapeProcessing() const { return escape_processing; }
  /**
   * Sets the escapeProcessing value.
   * @param escapeProcessingPrm escapeProcessing to set.
   */
  Request&          setEscapeProcessing (bool escapeProcessingPrm)
                         { escape_processing = escapeProcessingPrm; return *this;}
  /**
   * Returns the sqlQuery value.
   * @return Returns the sqlQuery.
   */
  std::wstring      getSqlQueryOrTemplate() const { return sqlQueryOrTemplate; }
  /**
   * Sets the sqlTemplate.
   */
  Request&          setParameters(std::wstring preparedStatementParametersPrm)
  	{ preparedStatementParameters = preparedStatementParametersPrm; return *this;}
  /**
   * Sets the sqlQueryOrTemplate value.
   */
  void              setSqlQueryOrTemplate(std::wstring sqlQueryOrTemplatePrm)
                        { sqlQueryOrTemplate = sqlQueryOrTemplatePrm; }
  /**
   * Cast to wstring operator to ease printing (for debug/traces).
   * @return sql query as a string
   */
  operator          std::wstring& () { return sqlQueryOrTemplate; }
  /**
   * Returns the preparedStatementParameters value.
   * @return Returns the preparedStatementParameters.
   */
  std::wstring      getPreparedStatementParameters() const { return preparedStatementParameters; }
  /**
   * Returns the isReadOnly value.
   * @return true if the connection has been set to read only
   */
  bool              isReadOnly() { return is_read_only; }
  /**
   * Sets the isReadOnly value.
   * @param isReadOnlyPrm value to set.
   */
  void              setIsReadOnly(bool isReadOnlyPrm)
                        { is_read_only = isReadOnlyPrm; }
  /**
   * Returns the id value.
   * @return the request identifier
   */
  int64_t           getId() const { return id; }
  /**
   * Sets the id value.
   * @param idPrm to set
   */
  void              setId(int64_t idPrm) { id = idPrm; }
  /**
   * Returns the isAutoCommit value.
   * @return the isAutoCommit
   */
  bool              isAutoCommit() const { return is_auto_commit; }
  /**
   * Sets the isAutoCommit value.
   * @param isAutoCommitPrm The isAutoCommit to set.
   */
  void              setIsAutoCommit(bool isAutoCommitPrm)
                        { is_auto_commit = isAutoCommitPrm; }
  /**
   * Returns the timeoutInSeconds value.
   * @return Returns the timeoutInSeconds.
   */
  int               getTimeoutInSeconds() const { return timeout_in_seconds; }
  /**
   * Sets the timeoutInSeconds value.
   * @param timeoutInSecondsPrm The timeoutInSeconds to set.
   */
  Request&          setTimeoutInSeconds(int timeoutInSecondsPrm)
                        { timeout_in_seconds = timeoutInSecondsPrm; return *this;}

protected:
  /**
   * Creates a new <code>Request</code> object. Protected to ensure no one but
   * friend classes can use us
   * @param sql the SQL statement
   */
  Request(std::wstring sql) :
    sqlQueryOrTemplate(sql),
    is_read_only(false),
    is_auto_commit(true),
    timeout_in_seconds(0),
    escape_processing(true) // default = true according to the JDBC spec
  {
  }

  virtual ~Request();

private:
  // forbid copies (java style)
  Request (const Request&);
  /** SQL query (should be set in constructor). TODO: should probably be const */
  std::wstring      sqlQueryOrTemplate;
  /** Parameters of the PreparedStatement */
  std::wstring      preparedStatementParameters;
  /** True if the connection has been set to read-only */
  bool              is_read_only;
  /**
   * Whether this request has been sent in <code>autocommit</code> mode or
   * not.
   */
  bool              is_auto_commit;
  /**
   * Timeout for this request in seconds, value 0 means no timeout (should be
   * set in constructor). This timeout is in seconds, reflecting the jdbc-spec,
   * and is passed as-is to the backends jdbc-driver. Internally converted to ms
   * via getTimeoutMs().
   */
  int               timeout_in_seconds;

  /**
   * Should the backend driver do escape processing before sending to the
   * database? Simply forwarded to backend driver. No setter for this member,
   * should be set in constructor.
   */
  bool              escape_processing;
  /**
   * Request identifier only used for failover purposes. This id is sent back by
   * the controller.
   */
  int64_t           id;

  
};

} //namespace CarobNS
#endif /*REQUEST_HPP_*/
