/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#ifndef REQUESTWITHRESULTSETPARAMETERS_HPP_
#define REQUESTWITHRESULTSETPARAMETERS_HPP_

#include "Request.hpp"

#include "CarobException.hpp"

#include <string>

namespace CarobNS {

/**
 * This class defines a Request with additional parameters indicating how the
 * ResulSet should be fetched. This is only useful for calls to
 * Statement.executeQuery() or Statement.execute() when ResultSets are returned.
 */
class RequestWithResultSetParameters : public Request
{
// We are private implementation of these two classes. So it is ok to make them
// friends
friend class Statement;
friend class ParameterStatement;
// Contructor and Destructor are protected to avoid access by anyone else than
// our friends
public:
  /**
   * Also serialize ResultSet parameters to the stream. Optionally used by
   * serializers of those derived requests that expect a ResultSet.
   * @param out destination socket to controller
   * @throws SocketIOException an error occurs on the stream
   * @see Request#sendToStream(const DriverSocket&) const
   */
  void              sendToStream(const DriverSocket& out) const
                        throw (SocketIOException, UnexpectedException);
  /**
   * Returns the cursorNameIsNull value
   * @return true if the cursorName is NULL (not valid)
   */
  bool              isCursorNameNull() const { return cursor_name_is_null; }
  /**
   * Returns the cursorName value.
   * @return Returns the cursorName.
   */
  std::wstring      getCursorName() const { return cursor_name; }
  /**
   * Sets the cursorName value.
   * @param cursorNamePrm cursorName to set.
   */
  void              setCursorName(std::wstring cursorNamePrm)
                        { cursor_name = cursorNamePrm; }
  /**
   * Returns the fetchSize value.
   * @return Returns the fetchSize.
   */
  int               getFetchSize () const { return fetch_size; }
  /**
   * Sets the fetchSize value.
   * @param fetchSizePrm fetchSize to set.
   */
  void              setFetchSize(int fetchSizePrm) { fetch_size = fetchSizePrm; }
  /**
   * Returns the maxRows value.
   * @return maximum number of rows.
   */
  int               getMaxRows() const { return max_rows; }
  /**
   * Sets the maxRows value.
   * @param maxRowsPrm maxRows to set.
   */
  void              setMaxRows(int maxRowsPrm) { max_rows = maxRowsPrm; }
protected:
  /**
   * Creates a new <code>RequestWithResultSetParameters</code> object.
   * Protected to ensure no one but friend classes can use us
   * @param sql the SQL statement
   */
  RequestWithResultSetParameters(std::wstring sql);

  virtual ~RequestWithResultSetParameters();
private:
  /** Maximum number of rows for the ResultSet*/
  int               max_rows;
  /** Size to be fetched for the ResultSet */
  int               fetch_size;
  /** Name of the cursor of the ResultSet */
  std::wstring      cursor_name;
  /** Whether the cursor name is valid */
  bool              cursor_name_is_null;

};

} //namespace CarobNS
#endif /*REQUESTWITHRESULTSETPARAMETERS_HPP_*/

