/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s):
 */

#ifndef RESULTANDWARNINGS_H_
#define RESULTANDWARNINGS_H_

#include "DriverResultSet.hpp"
#include "CarobException.hpp"

#include <list>

namespace CarobNS {

/**
 * Container that holds XXXexecute() and XXXexecuteUpdate result (update count
 * or result list) and the eventual associated warnings. Note that
 * XXXexecuteQuery result and warnings are held by DriverResultSet class
 */
class ResultAndWarnings
{
public:
  /**
   * Empty constructor
   */
  ResultAndWarnings() : warnings_ptr(NULL) {};
  /**
   * Constructs a <code>ResultAndWarning</code> that will hold an updateCount
   * and the given SQLWarnings.<br>
   * This constructor will typically be called by XXXexecuteUpdate() functions
   * 
   * @param uc the update count
   * @param warnsPtr pointer to the warnings to set, can be null
   */
  ResultAndWarnings(int uc, SQLWarning* warnsPtr)
  {
    warnings_ptr = warnsPtr;
    update_count = uc;
    is_update_count = true;
  }

  /**
   * Constructs a <code>ResultAndWarning</code> that will hold a resultList
   * and the given SQLWarnings.<br>
   * This constructor will typically be called by XXXexecute() functions
   * 
   * @param reslist list of results
   * @param warnsPtr pointer to the warnings to set, can be null
   */
  ResultAndWarnings(std::list<ResultSetOrUpdateCount>  reslist, SQLWarning* warnsPtr)
  {
    warnings_ptr = warnsPtr;
    result_list = reslist;
    is_update_count = false;
  }

  /**
   * Copy constructor
   */
  ResultAndWarnings(const ResultAndWarnings& cp) : warnings_ptr(NULL)
  {
    *this = cp;
  }

  /**
   * Assignement operator. Deletes the previous warnings and replaces it with 
   * the one from the right value
   * @param cp the right value
   */
  ResultAndWarnings & operator =(const ResultAndWarnings& cp)
  {
    if (warnings_ptr != NULL)
      delete warnings_ptr;
    warnings_ptr = cp.getWarnings();
    result_list = cp.getResultList();
    update_count = cp.getUpdateCount();
    is_update_count = cp.isUpdateCount();
    return *this;
  }
  /**
   * Tells whether this object holds an updateCount
   * 
   * @return true if the result is an updateCount
   */
  bool isUpdateCount() const { return is_update_count; }

  /**
   * Tells whether this object holds a resultList
   * 
   * @return true if the result is an resultList
   */
  bool isResultList() const { return !is_update_count; }

  /**
   * Gets the warning chain pointer associated to the result
   * 
   * @return a <code>SQLWarning</code> pointer or null if no warnings
   */
  SQLWarning* getWarnings() const { return warnings_ptr; }

  /**
   * Gets the updateCount. <i>Note:</i> if the held value is a resultList (ie.
   * if {@link #isUpdateCount()} returns false), then the returned value is not
   * specified
   * 
   * @return the updateCount value
   */
  int getUpdateCount() const { return update_count; }

  /**
   * Gets the ResultList. <i>Note:</i> if the held value is an udpateCount
   * (ie. if {@link #isResultList()} returns false), then the returned value is
   * not specified
   * 
   * @return the updateCount value
   */
  std::list<ResultSetOrUpdateCount>   getResultList() const { return result_list; }

private:
  /** Warning chain, null means no warning */
  SQLWarning*                         warnings_ptr;
  /** True if this object holds an update count, false if it holds a result List */
  bool                                is_update_count;
  /** Result of XXXexecuteUpdate */
  int                                 update_count;
  /** Result of XXXexecute */
  std::list<ResultSetOrUpdateCount>   result_list;
};

} //namespace CarobNS
#endif /*RESULTANDWARNINGS_H_*/
