/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#ifndef RESULTSETMETADATA_HPP_
#define RESULTSETMETADATA_HPP_

#include "CarobException.hpp"

namespace CarobNS {

//Nullability constants
/**
 * The constant indicating that a
 * column does not allow <code>NULL</code> values.
 */
#define columnNoNulls             0
/**
 * The constant indicating that a
 * column allows <code>NULL</code> values.
 */
#define columnNullable            1
/**
 * The constant indicating that the
 * nullability of a column's values is unknown.
 */
#define columnNullableUnknown     2

class DriverResultSet;

/**
 * ResultSet metadata provided for pretty printing of the ResultSet by a
 * console.
 */
class ResultSetMetaData
{
public:
  /**
   * Constructs a ResultSetMetaData from a DriverResultSet.
   * @param rs DriverResultSet to be associated to this MetaData
   * @exception DriverException if the given result set is NULL
   */
	ResultSetMetaData(DriverResultSet* rs) throw (DriverException, 
      UnexpectedException);
      
	virtual ~ResultSetMetaData() {};
  /**
   * Returns the number of columns in this <code>ResultSet</code> object.
   * @return the number of columns
   */
  int                   getColumnCount();
  /**
   * Indicates whether the designated column is automatically numbered, thus
   * read-only.
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isAutoIncrement(int column) throw (DriverException, 
                            UnexpectedException);
  /**
   * Indicates whether a column's case matters.
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isCaseSensitive(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Indicates whether the designated column can be used in a where clause.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isSearchable(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Indicates whether the designated column is a cash value.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isCurrency(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Indicates the nullability of values in the designated column.    
   *
   * @param column the first column is 1, the second is 2, ...
   * @return the nullability status of the given column; one of
   *         <code>columnNoNulls</code>, <code>columnNullable</code> or 
   *         <code>columnNullableUnknown</code>
   * @exception DriverException if a the given column number is out of range
   */
  int                   isNullable(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Indicates whether values in the designated column are signed numbers.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isSigned(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Indicates the designated column's normal maximum width in characters.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return the normal maximum number of characters allowed as the width
   *          of the designated column
   * @exception DriverException if a the given column number is out of range
   */
  int                   getColumnDisplaySize(int column) throw (DriverException,
                            UnexpectedException);

  /**
   * Gets the designated column's suggested title for use in printouts and
   * displays.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return the suggested column title
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getColumnLabel(int column) throw (DriverException, 
                            UnexpectedException);  

  /**
   * Get the designated column's name.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return column name
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getColumnName(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Get the designated column's table's schema.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return schema name or "" if not applicable
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getSchemaName(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Get the designated column's number of decimal digits.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return precision
   * @exception DriverException if a the given column number is out of range
   */
  int                   getPrecision(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Gets the designated column's number of digits to right of the decimal point.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return scale
   * @exception DriverException if a the given column number is out of range
   */
  int                   getScale(int column) throw (DriverException, 
                            UnexpectedException); 

  /**
   * Gets the designated column's table name. 
   *
   * @param column the first column is 1, the second is 2, ...
   * @return table name or "" if not applicable
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getTableName(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Gets the designated column's table's catalog name.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return the name of the catalog for the table in which the given column
   *          appears or "" if not applicable
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getCatalogName(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Retrieves the designated column's SQL type.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return SQL type
   * @exception DriverException if a the given column number is out of range
   * @see TypeConstants.hpp
   */
  int                   getColumnType(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Retrieves the designated column's database-specific type name.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return type name used by the database. If the column type is
   * a user-defined type, then a fully-qualified type name is returned.
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getColumnTypeName(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Indicates whether the designated column is definitely not writable.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isReadOnly(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Indicates whether it is possible for a write on the designated column to succeed.
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isWritable(int column) throw (DriverException, 
                            UnexpectedException);

  /**
   * Indicates whether a write on the designated column will definitely succeed.  
   *
   * @param column the first column is 1, the second is 2, ...
   * @return <code>true</code> if so; <code>false</code> otherwise
   * @exception DriverException if a the given column number is out of range
   */
  bool                  isDefinitelyWritable(int column) throw (DriverException, 
                            UnexpectedException);

   /**
   * Returns the fully-qualified name of the Java class.
   * Not very meanfull for carob, but keeps java driver code reflexion
   *
   * @param column the first column is 1, the second is 2, ...
   * @return the fully-qualified name of the class in the Java programming
   *         language.
   * @exception DriverException if a the given column number is out of range
   */
  std::wstring          getColumnClassName(int column) throw (DriverException, 
                            UnexpectedException);

private:
  /**
   * Pointer to the corresponding result set. Must be set at construction and 
   * should never be NULL
   */
  DriverResultSet*      resultSetPtr;
  /**
   * Utility function: checks that the given column number is between 1 and 
   * ResultSet's <code>nbOfColumns</code> and throws and exception if not.
   * @exception DriverException if a the given column number is out of range
   */
  void                  checkColumnIndex(int column) throw (DriverException,
    UnexpectedException);
};

} //namespace CarobNS
#endif /*RESULTSETMETADATA_HPP_*/
