/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#ifndef SQLDATASERIALIZATION_H_
#define SQLDATASERIALIZATION_H_


#include "CarobException.hpp"
#include "Common.hpp" // for java_byte: cannot forward declare typedefs


#include <stdint.h> // for uint8_t, int32_t, uint32_t, int64_t, uint64_t
// Alternative:
// #include <cstdint>

// Warning: this header and all the types it defines (including "long
// long") are part of C99 and are MISSING from C++98.

// See:
// - "Incompatibilities Between ISO C and ISO C++" - David R. Tribble
// - C++ Technical Report 1 (TR1)

// See also related file "carob/include/C99Math.h"

namespace CarobNS {

class TypeTag;
class DriverSocket;
union ResultSetDataType; // defined below

/** To ease and simplify deserialization function pointers declaration */
typedef ResultSetDataType (*deserializerPtr)(const DriverSocket&);

/**
 * This class defines Serializers for SQL Data: one function
 * of deserialization per data type
 */
class SQLDataSerialization
{
public:
  /**
   * To get the deserializer, give the object type and you'll get the function.
   * @param ttPrm type of object to be deserialized
   * @throws NotImplementedException if the deserialization for the given type
   * is not implemented yet
   */
  static deserializerPtr getDeserializer(TypeTag ttPrm)
      throw (NotImplementedException, UnexpectedException);

};

/** Hold a SQL <code>TIMESTAMP</code> value */
typedef struct {
  /** Integral seconds (no milliseconds) */
  int64_t time;
  /** Fractional seconds (milliseconds) */
  int32_t nanos;
} SQLTimeStamp;

/** Large Data: byte array, blob, clob */
typedef struct
{
  /** Number of bytes held */
  int32_t     length;
  /** Actual data, as an array of bytes */
  java_byte*  data;
} LargeData;

/**
 * Defines data type that can be found in ResultSets
 */
union ResultSetDataType
{
  /** for booleans and bits */
  bool    as_bool;
  /** for bytes */
  uint8_t as_byte;
  /** for integers */
  int32_t as_int;
  /** for longs */
  int64_t as_long;
  /** for floats */
  float   as_float;
  /** for doubles */
  double  as_double;
  /** for all other types */
  void*   as_other;
};

} //namespace CarobNS
#endif /*SQLDATASERIALIZATION_H_*/
