/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): Marc Herbert
 */
#ifndef SOCKETADDRESS_HPP_
#define SOCKETADDRESS_HPP_

#include "CarobException.hpp"

#ifdef __MINGW32__
  #include <winsock2.h>
  // Needed ? #include <ws2tcpip.h>
  #include "Common.hpp" //for in_port_t
#else
  #include <sys/socket.h> // sockaddr & socklen_t
  #include <netdb.h> // addrinfo
#endif

namespace CarobNS {

std::wostream& operator<< (std::wostream& ostr, const sockaddr& toprint);

/**
 * Wraps a sockaddr, its size and provides a less than operator so it can be
 * the keys of stl maps
 */
class SocketAddress
{
public:
  /** Empty constructor */
  SocketAddress();
  /**
   * Copy constructor, calls = operator
   * @param sa the socket address to copy infos from
   */
  SocketAddress(const SocketAddress& sa);
  /** frees address field */
  ~SocketAddress();
  /**
   * Creates a <code>SocketAddress</code> with the given address and length.
   * Calls #setAddress(const sockaddr* sa, socklen_t l)
   * @param sa the socket address
   * @param l length of sa
   */
  SocketAddress(sockaddr* sa, socklen_t l);
  /**
   * Set the given address and length
   * @param sa the socket address to set
   * @param l length of sa
   */
  void            setAddress(const sockaddr* sa, socklen_t l);
  /** Assignement operator, calls #setAddress(const sockaddr* sa, socklen_t l) */
  SocketAddress&  operator = (const SocketAddress& sa);
  /** Less than operator, so SocketAddress can be keys of stl maps */
  bool            operator < (const SocketAddress& sa) const;
  /** Returns a pointer to the socket address */
  // This C-style cast is mandatory: reinterpret_cast brakes constness...
  const sockaddr* getAddressPtr() const { return (sockaddr*)&address; }
  /** Returns the socket address length */
  socklen_t       getAddressLength() const { return address_length; }
  /** Gets the address family of this socket */
  sa_family_t     getFamily() const { return address.ss_family; }
private:
  /** Pointer to the sockaddr struct */
  sockaddr_storage address;
  /** Size of the sockaddr struct */
  socklen_t       address_length;
};

} //namespace CarobNS
#endif /*SOCKETADDRESS_HPP_*/
