/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */
#ifndef STATEMENT_HPP_
#define STATEMENT_HPP_

#include "DriverResultSet.hpp"

#include "Common.hpp"

#include <string>
#include <list>

namespace CarobNS {

/** 
 * The constant indicating that the current <code>ResultSet</code> object 
 * should be closed when calling <code>getMoreResults</code>.
 */
#define CLOSE_CURRENT_RESULT          1

/**
 * The constant indicating that the current <code>ResultSet</code> object
 * should not be closed when calling <code>getMoreResults</code>.
 */
#define KEEP_CURRENT_RESULT           2

/**
 * The constant indicating that all <code>ResultSet</code> objects that
 * have previously been kept open should be closed when calling
 * <code>getMoreResults</code>.
 */
#define CLOSE_ALL_RESULTS             3

const std::wstring emptyString = L"";

class Connection;
class ResultSetMetaData;
class RequestWithResultSetParameters;
class SQLWarning;

/**
 * A <code>Statement</code> object is used for executing a static SQL
 * statement and obtaining the results produced by it.
 * <p>
 * Only one <code>ResultSet</code> per <code>Statement</code> can be open at
 * any point in time. Therefore, if the reading of one <code>ResultSet</code>
 * is interleaved with the reading of another, each must have been generated by
 * different <code>Statements</code>. All <code>Statements</code> execute
 * methods implicitly close a statement's current <code>ResultSet</code> if an
 * open one exists.
 */
class Statement
{
/**
 * Connection class must have access to at least constructor and destructor
 * because only Connection can create statements
 */
friend class Connection;
/**
 * DriverResultSet must inform us when it is destroyed. Let's allow it to
 * reset the <code>resultPtr</code>
 */
friend class DriverResultSet;

public:
  //No constructor nor destructor, statements must be created and destroyed by
  //connections
  /**
   * Execute a SQL statement that may return multiple results.
   * @param sql SQL statement to set before executing
   * @return true if the result is a ResultSet or false if it is an integer
   * @throw DriverException if the statement is closed
   */
  bool                    execute(const std::wstring &sql)
                              throw (DriverException,
                                  SocketIOException, BackendException,
                                  ControllerException, ProtocolException,
                                  NotImplementedException, UnexpectedException)
                              { return execute(trim(sql), emptyString); }

  /**
   * Executes a SQL statement that returns a single ResultSet
   * @param sql SQL statement to set before executing. Typically a
   * static SQL <code>SELECT</code> statement
   * @return a ResulSet pointer that contains the data produced by the query
   * @throw DriverException if the statement is closed
   */
  DriverResultSet*        executeQuery(const std::wstring &sql)
                              throw (DriverException, SocketIOException,
                                  BackendException, ControllerException,
                                  ProtocolException, NotImplementedException,
                                  UnexpectedException)
                              { return executeQuery(trim(sql), emptyString); }

  /**
   * Execute a SQL INSERT, UPDATE or DELETE statement. In addition SQL
   * statements that return nothing such as SQL DDL statements can be executed
   * @param sql SQL statement to set before executing
   * @return either a row count, or 0 for SQL commands
   */
  int                     executeUpdate(std::wstring sql)
                              throw (DriverException, SocketIOException,
                                  BackendException, ControllerException,
                                  ProtocolException, UnexpectedException)
                              { return executeUpdate(sql, emptyString); }

  /**
   * Returns true if there are results available for retrieval with
   * getMoreResults(), but does NOT move to the next result in the list.
   * 
   * @return true if there are more results pending for retrieval.
   */
  bool                    moreResults() const;

  /**
   * Return the true if next available result is a ResultSet or false it this is
   * an update count. If the result is false and getUpdateCount() returns -1
   * then there is no more result.
   * <p>
   * Move to the next result. Any open ResultSet is implicitly closed.
   * 
   * @return true for a ResultSet, false for an updateCount or no more results
   */
  bool                    getMoreResults() throw (DriverException,
                              SocketIOException, BackendException,
                              ControllerException, ProtocolException,
                              UnexpectedException);
  /**
   * Moves to this <code>Statement</code> object's next result, deals with any
   * current <code>ResultSet</code> object(s) according to the instructions
   * specified by the given flag, and returns <code>true</code> if the next
   * result is a <code>ResultSet</code> object.
   * <p>
   * There are no more results when the following is <code>true</code>:
   * 
   * <pre>(!getMoreResults() &amp;&amp; (getUpdateCount() == -1)</pre>
   * 
   * @param current one of the following <code>Statement</code> constants
   *          indicating what should happen to current <code>ResultSet</code>
   *          objects obtained using the method
   *          <code>getResultSet</code>: <code>CLOSE_CURRENT_RESULT</code>,
   *          <code>KEEP_CURRENT_RESULT</code>, or <code>CLOSE_ALL_RESULTS</code>
   * @return <code>true</code> if the next result is a <code>ResultSet</code>
   *         object; <code>false</code> if it is an update count or there are
   *         no more results
   * @see #execute(const std::wstring &)
   */
  bool                    getMoreResults(int current) throw (DriverException,
                              SocketIOException, BackendException,
                              ControllerException, ProtocolException,
                              UnexpectedException);
  // GETTERS AND SETTERS
  /**
   * Returns the current result as an update count, if the result is a
   * <code>ResultSet</code> or there are no more results, -1 is returned.
   * 
   * @return the current result as an update count.
   */
  int                     getUpdateCount() const { return lastUpdateCount; }
  /**
   * The queryTimeout limit is the number of seconds the driver will wait for a
   * Statement to execute.
   * 
   * @return the current query timeout limit in seconds; 0 = unlimited
   */
  int                     getQueryTimeout() const { return timeout; }
  /**
   * Sets the number of seconds the driver will wait for a
   * <code>Statement</code> object to execute.
   * 
   * @param seconds the new query timeout limit in seconds; 0 means no timeout
   * @throw DriverException if the condition seconds >= 0 is not satisfied
   */
  void                    setQueryTimeout(int seconds)
                              throw (DriverException, UnexpectedException);
  /** Gets the fetch size */
  int                     getFetchSize() const { return fetchSize; }
  /**
   * Sets the default fetch size for the produced ResultSet.
   * 
   * @param rows number of rows that should be fetched from the database
   * @throw DriverException if the condition 0 <= size <= getMaxRows() is not
   *        satisfied
   */
  void                    setFetchSize(int rows)
                              throw (DriverException, UnexpectedException);
  /**
   * Defines the SQL cursor name that will be used by subsequent execute
   * methods. This name can then be used in SQL positioned update/delete
   * statements to identify the current row in the ResultSet generated by this
   * statement. If a database doesn't support positioned update/delete, this
   * method is a no-op.
   * 
   * @param name the new cursor name
   */
  void                    setCursorName(std::wstring name)
                              { cursorName = name; }
  /**
   * Retrieves the type of the generated <code>ResultSet</code>.
   * 
   * @return one of <code>TYPE_FORWARD_ONLY</code> or
   *         <code>TYPE_SCROLL_INSENSITIVE</code>
   */
  int                     getResultSetType() const { return resultSetType; }
  /**
   * @param value an <code>int</code> value
   */
  void                    setResultSetType(int value)
                              throw (DriverException, UnexpectedException);
  /**
   * Returns the current result as a <code>DriverResultSet</code> pointer.
   * 
   * @return pointer to the current result set; null if it is an
   * update count or there are no more.
   */
  DriverResultSet*        getResultSet() const { return lastResultPtr; }
  /**
   * Retrieves the concurrency mode for the <code>ResultSet</code>.
   * @return <code>CONCUR_READ_ONLY</code> or <code>CONCUR_UPDATABLE</code>
   */
  int                     getResultSetConcurrency() const
                              { return resultSetConcurrency; }
  /**
   * Sets the concurrency mode for the <code>ResultSet</code>
   * @param value an <code>int</code> value <code>CONCUR_READ_ONLY</code> or
   * <code>CONCUR_UPDATABLE</code>
   * @throw DriverException if the input value is not valid
   */
  void                    setResultSetConcurrency(int value)
                              throw (DriverException, UnexpectedException);
  /**
   * The maxRows limit is set to limit the number of rows that any
   * <code>ResultSet</code> can contain. If the limit is exceeded, the excess
   * rows are silently dropped.
   * 
   * @return the current maximum row limit; zero means unlimited
   */
  int                     getMaxRows() const { return maxRows; }
  /**
   * Sets the maximum number of rows that any <code>ResultSet</code> can
   * contain.
   * 
   * @param max the new max rows limit; 0 means unlimited
   * @exception DriverException if the condition max >= 0 is not satisfied
   */
  void                    setMaxRows(int max) throw (DriverException,
                              UnexpectedException);
  /**
   * If escape scanning is on (the default), the driver will do escape
   * substitution before sending the SQL to the database.
   * 
   * @param enable true to enable; false to disable
   */
  void                    setEscapeProcessing(bool enable)
                              { escapeProcessing = enable; }
  /**
   * Remotely closes the current ResultSet if applicable and frees result
   * memory. In many cases, it is desirable to immediately release a Statement's
   * database and JDBC resources instead of waiting for this to happen when it
   * is automatically closed. The close method provides this immediate release.
   * <p>
   */
  virtual void            close();
  /**
   * Returns the metadata of the current or future ResultSet. Note
   * that if there is no ResultSet available yet (before execution),
   * then this method will generate some network communication to the
   * controller. No caching is performed here: before execution this
   * network communication will happen at every call. Also consider
   * bug SEQUOIA-262 (empty returned instead of null).
   *
   * @return a pointer to the metadata of this statement or of its resultset
   */
   ResultSetMetaData*     getMetaData() throw (SocketIOException,
                              BackendException, ProtocolException, 
                              NotImplementedException, UnexpectedException);
  /**
   * Sets the sql request string. This is in total opposition with the
   * JDBC API where request strings are not mutable, see CAROB-81.
   *
   * @param sql string containing the sql statement
   */
  void                    setRequest(const std::wstring& sql)
                              { sql_request = sql; }
  /**
   * Gets the SQL warnings associated to this statement
   * @return a SQL warning chain
   */
  SQLWarning*             getWarnings()  throw (DriverException,
                              UnexpectedException);
protected:
  /** Original, untouched request (only trimmed) */
  std::wstring            sql_request;
  /**
   * Constructs a Statement with the given connection. Protected access so that
   * only Connection class can create a statement
   * @param c pointer to connection that created us
   */
  Statement(Connection* c);
  /**
   * Sets the maxRows, fetchSize and cursorName values of this Statement in the
   * given request
   * @param request on which to set parameters (unrelated to user parameters)
   */
  void                    setReadRequestParameters(
                              RequestWithResultSetParameters& request);
  /**
   * Nulls the given resultSet in the resultList. This function is called by
   * the driverResultSet destructor to inform that the give resultSet is beeing
   * destroyed, so that any reference to it is not valid anymore
   * @param drsPtr pointer to the ResultSet that must be invalidated
   */
  void                    removeResultSetFromList(DriverResultSet* drsPtr);
  /** List of updateCount/ResultSet */
  std::list<ResultSetOrUpdateCount>  resultList;
  /** Iterator on the list of results */
  std::list<ResultSetOrUpdateCount>::const_iterator resultListIterator;
  /**
   * Returns the pointer to the connection that created us.
   */
  Connection*             getConnectionPtr() {return connectionPtr;}

  /**
   * Convenience method for derived class ParameterStatement
   * @param skeleton trimmed SQL skeleton statement
   * @param parameters SQL parameters serialized into a string
   * @return true if the result is a ResultSet or false if it is an integer
   */
  bool                    execute(const std::wstring &skeleton, const std::wstring &parameters)
                              throw (DriverException,
                                  SocketIOException, BackendException,
                                  ControllerException, ProtocolException,
                                  NotImplementedException, UnexpectedException);
  /**
   * Convenience method for derived class ParameterStatement
   * @param skeleton trimmed SQL skeleton statement
   * @param parameters SQL parameters serialized into a string
   * @return a ResulSet pointer that contains the data produced by the query
   */
  DriverResultSet*        executeQuery(const std::wstring &skeleton, const std::wstring &parameters)
                              throw (DriverException, SocketIOException,
                                  BackendException, ControllerException,
                                  ProtocolException, NotImplementedException,
                                  UnexpectedException);

  /**
   * Convenience method for derived class ParameterStatement
   * @param skeleton trimmed SQL skeleton statement
   * @param serializedParameters SQL parameters serialized into a string
   * @return either a row count, or 0 for SQL commands
   */
  int                     executeUpdate(const std::wstring &skeleton,
					const std::wstring &serializedParameters)
                              throw (DriverException, SocketIOException,
                                  BackendException, ControllerException,
                                  ProtocolException, UnexpectedException);
  /**
   * Forbid destruction to the rest of the world.
   * Statements must be deleted by connection.
   * @see Connection::deleteStatement(Statement* stPtr)
   */
  virtual ~Statement();
private:
  /** Connection that created us */
  Connection*             connectionPtr;
  /** Current result for a read request */
  DriverResultSet*        lastResultPtr;
  /**
   * Empty ResultSet just to hold the metadata got from
   * Connection#preparedStatementGetMetaData(Request)
   */
  DriverResultSet*        metadata_holder_ptr;
  /** metadata got from #getMetaData() */
  ResultSetMetaData*      rs_metadata_ptr;
  /** Update count for a write request */
  int                     lastUpdateCount;
  /** Warning chain */
  SQLWarning*             warningsPtr;
  /** Query timeout in seconds (0 means no timeout) */
  int                     timeout;
  /** Default ResultSet fetch size */
  int                     fetchSize;
  /** Cursor name used jointly with fetch size */
  std::wstring            cursorName;
  /** Type of the ResultSet defaults to TYPE_FORWARD_ONLY */
  int                     resultSetType;
  /** ResultSet Concurrency defaults to CONCUR_READ_ONLY */
  int                     resultSetConcurrency;
  /** Maximum number of rows */
  int                     maxRows;
  /** Should the driver to escape processing before sending to the DB? */
  bool                    escapeProcessing;
  /** End of result list marker, declared and initialized once for all */
  static const ResultSetOrUpdateCount endOfResultList;
  /**
   * Tests if this statement is closed.
   * @return <code>true</code> if this statement is closed
   */
  bool                    isClosed() const { return (connectionPtr == NULL); }
  /**
   * Iterates through the result list to close and delete result sets, then
   * clears the list
   */
  void                    cleanUpResults();
};

} //namespace CarobNS
#endif /*STATEMENT_HPP_*/
