
/*
 * Sequoia: Database clustering technology Copyright 2002-2005
 * Continuent, Inc.

 * Licensed under the Apache License, Version 2.0 (the "License"); you
 * may not use this file except in compliance with the License. You
 * may *obtain a copy of the License at

 * http://www.apache.org/licenses/LICENSE-2.0 

 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * permissions and limitations under the License.

 * Initial developer(s): Marc Herbert
 * Contributor(s): Csaba Simon
 */

#ifndef CAROB_STRING_CODEC_HPP_
#define CAROB_STRING_CODEC_HPP_

#include "CarobException.hpp"

#include <locale>


namespace CarobNS {

/**
 * Exception thrown when an encoding error occurs.
 */
class CodecException : public DriverException
{
public:
  /**
   * Constructs a CodecException with the given message
   * @param s error message of the exception
   */
  CodecException(std::wstring s) : DriverException(s, L"22018") {};
  // In theory 22018 is only for SQL data conversions, and there is
  // a slight chance that we were doing something else. Even in this
  // case it is still more informative than nothing.
};


typedef std::codecvt<wchar_t, char, mbstate_t> mbs_codecvt;
typedef std::codecvt_byname<wchar_t, char, mbstate_t> mbs_codecvt_byname;


// For a direct access to iconv read this:
// file://localhost/usr/share/doc/gcc-4.0-base/libstdc++/html/22_locale/codecvt.html

// This code probably does not work with UTF-16 wide chars, only UCS-2
// and UCS-4

std::locale trylocale(const char * const name) throw (CodecException);

/** Base class for en/decoders */
class StringCodec
{
public:
    /** Encodes the given string using the encoding set in the
     * constructor. Thread safe */
    virtual std::string encode(const std::wstring&) const throw (CodecException) = 0;
    /** Decodes the given bytes using the encoding set in the
     * constructor. Thread safe */
    virtual std::wstring decode(const std::string&) const throw (CodecException) = 0;
    virtual ~StringCodec() { }; // not much to delete...

  /** Optional optimization. Maybe faster method, suitable ONLY for
   * non-concurrent use. Unlike the safe encode() method above, this
   * method may or may not be thread-safe, depending on the actual
   * codec implementation
   * @param s decoded wide string
   * @return encoded string
   */
    virtual std::string nomutex_encode(const std::wstring& s) const throw (CodecException)
    { return encode(s); }
    /** See nomutex_encode. Calling nomutex_encode() and
     * nomutex_decode() each only once at the same time is safe
     * @param s encoded string
     * @return decoded wide string
     */
    virtual std::wstring nomutex_decode(const std::string& s) const throw (CodecException)
    { return decode(s); }
};


/**
 * Multi-Byte String encoder and decoders according to a given or default locale
 */
class MBSCodec : public StringCodec
{
  // Here order matters (see init below)
  // Hold our own (and cheap) copy so contained codecvt facet is not deleted
  const std::locale loc;
  const mbs_codecvt& codecvt;

public:
  /** Encodes the given string using the locale set in the constructor */
  std::string encode(const std::wstring&) const throw (CodecException);
  /** Decodes the given string using the locale set in the constructor */
  std::wstring decode(const std::string&) const throw (CodecException);

  /**
   * Default constructor for the default converter, using
   * user-defined locale (typically LANG, LC_etc).  If you want the
   * C locale instead of this LANG-one use for instance this CTOR:
   *   MBSCodec(locale::classic())
   * or if you want the current (global) locale:
   *   MBSCodec(locale())
   */
  MBSCodec() :
      loc(std::locale("")),
      codecvt (std::use_facet<mbs_codecvt>(loc))
  { }

  /**
   * Use codecvt of locale locarg for conversions
   * @param locarg locale to be used
   */
  MBSCodec(const std::locale& locarg) :
      loc(locarg),
      codecvt (std::use_facet<mbs_codecvt>(loc))
  { }

  /**
   * Use codecvt of locale named locname for conversions.
   *
   * May log errors: this constructor is NOT allowed to construct a
   * static!
   *
   * @param locname name of the locale to be used
   */
  MBSCodec(const char* locname) throw (CodecException) :
      loc(trylocale(locname)),
      codecvt (std::use_facet<mbs_codecvt>(loc))
  { }
  
};


#ifdef CAROB_USE_ICONV

/*
 *  WARNING: carob with and carob without iconv are currently binary
 *  compatible. That is: clients with this macro OFF can link fine to
 *  carob with this macro ON. Be VERY CAREFUL with what you put inside
 *  here to avoid breaking that. See CAROB-79
 */

} // CarobNS namespace
#include <iconv.h>
#include <CriticalSection.hpp>
namespace CarobNS {


/**
 * Multi-Byte String encoder and decoders based on iconv
 */
class IconvCodec : public StringCodec
{
public:
  /**
   * Constructor.
   * @param code The character code from/to the conversion will be.
   */
  IconvCodec(const std::string &encodingName) throw (CodecException);

  /**
   * Destructor.
   */
  ~IconvCodec();

  /**
   * Encode from wstring to string.
   * @param w the wstring.
   * @return a string.
   */
  std::string encode(const std::wstring&) const throw (CodecException);

  std::string nomutex_encode(const std::wstring&) const throw (CodecException);

  /**
   * Decode from string to wstring.
   * @param s the string.
   * @return a wstring.
   */
  std::wstring decode(const std::string&) const throw (CodecException);

  std::wstring nomutex_decode(const std::string&) const throw (CodecException);


private:
  /**
   * the conversion descriptors
   */
  iconv_t cd_from, cd_to;
  /** iconv descriptors are stateful/not thread-safe */
  mutable CriticalSection iconvDecodeCS, iconvEncodeCS;

  std::wstring encodingName;
};
#endif // CAROB_USE_ICONV

/** 
 * The only purpose of this class is to prevent direct access to the
 * private codecs fields, while still be able to inline the calls.
 * Since the methods below rely on some static initialization, they
 * should NOT be used at initialization time!
 */
class StaticCodecs
{
public:
  /**
   * Converts ASCII "encoded" string to a wide string.<br>
   * To be used only when the input string is definitely ASCII
   * @param in ASCII string to convert
   * @return converted wide string
   * @see #fromString(const std::string&)
   */
  static std::wstring fromASCII(const std::string& in) throw (CodecException)
  { return ascii_codec.decode(in); }
  /**
   * Converts user's locale encoded string to a wide string<br>
   * To be used for conversions of locale dependent strings, eg. system messages
   * @param in string to convert
   * @return converted wide string
   */
  static std::wstring fromString(const std::string& in) throw (CodecException)
  { return user_codec.decode(in); }
  /** Converts a wide string to a user's locale encoded string */
  static std::string toString(const std::wstring& in) throw (CodecException)
  { return user_codec.encode(in); }
private:
  static const CarobNS::MBSCodec ascii_codec;
  /** User-defined locale (typically set using LANG, LC_etc) codec */
  static const CarobNS::MBSCodec user_codec;
};

/**
 * Default (locale) String to Wide-String conversion. Since this method relies
 * on some static initialization, it should NOT be used at initialization
 * time!
 * @param in original string
 * @return converted wstring
 */
inline std::wstring fromString(const std::string& in) throw (CodecException)
{
  return StaticCodecs::fromString(in);
}
/**
 * Wide-String to default (locale) String conversion. Since this method relies
 * on some static initialization, it should NOT be used at initialization
 * time!
 * FIXME: throw exception on error
 * @param in original wide string
 * @return converted string
 */
inline std::string toString(const std::wstring& in) throw (CodecException)
{
  return StaticCodecs::toString(in);
}

} // namespace CarobNS

#endif // include only once

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
