/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#ifndef WATCHED_CONTROLLERS_H_
#define WATCHED_CONTROLLERS_H_

#include "SocketAddress.hpp"
#include "ControllerInfo.hpp"
#include "CriticalSection.hpp"

#include <map>
#include <vector>

namespace CarobNS {

class ControllerStateChangedCallback;

/**
 * Provides a list of controllers and maintains their state (up or down),
 * launching appropriatly methods from callback given callback instance.<br>
 * To each controller is associated a lastTimeSeen value corresponding to the
 * last time the controller responded to a ping. Their state is updated each
 * time the lastTimeSeen value is accessed (read or written). At init time, all
 * controllers are considered as up (responding to pings). Controllers can also
 * be forced as 'no more responding to pings' (ie. down)
 */
class WatchedControllers
{
  /**
   * Holds a controller, its lastTimeSeen variable and its state (up or down)
   */
  class ControllerAndState
  {
  public:
    ControllerAndState();
    ControllerAndState(const ControllerInfo& ctrl);
    ControllerAndState(const ControllerInfo& ctrl, int64_t initTime);
    ControllerAndState(const ControllerAndState& src);

    ControllerAndState&   operator = (const ControllerAndState& src);

    void                  updateTime(int64_t newTime) { last_time_seen = newTime; }

    bool                  isUp() const { return responds_to_pings; }

    int64_t               getLastTimeSeen() const { return last_time_seen; }

    void                  setUp() { responds_to_pings = true; }

    void                  setDown() { responds_to_pings = false; }

    const ControllerInfo& getController() const { return controller; }
  private:
    ControllerInfo        controller;
    bool                  responds_to_pings;
    int64_t               last_time_seen;
  };

public:
  /**
   * Creates a new list of watched controllers with the given set of controllers
   * and an initial lastTimeSeen value.<br>
   * 
   * @param controllersPrm set of controllers identified by their
   *          <code>ControllerInfo</code>
   * @param initTime initial lastTimeSeen value, typically <code>clock()</code>
   * @param controllerTimeout delay after which a controller will be considered
   *          as failing if it did not respond to pings
   * @param cbPtr Callback implementation to call when a controller state
   *          changes
   */
  WatchedControllers(const std::vector<ControllerInfo>& controllersPrm, int64_t initTime,
      int controllerTimeout, ControllerStateChangedCallback* cbPtr);

  /**
   * Deletes the callback
   */
  ~WatchedControllers();

  /**
   * Adds a controller to the list and associates the given lastTimeSeenValue to
   * it. Note that this operation is thread-safe, thus can slow-down other
   * concurrent operations like {@link #removeController(const ControllerInfo&)},
   * {@link #setControllerResponsed(const SocketAddress&, int64_t)}
   * @param ctrl controller to be added
   * @param lastTimeSeen initial time to associate to this controller. Typically
   * the result of gettimeofday()
   */
  void addController(const ControllerInfo& ctrl, int64_t lastTimeSeen);

  /** Removes the selected controller and its associated lastTimeSeen */
  void removeController(const ControllerInfo& ctrl);

  /**
   * Associates given lastTimeSeen value to specified controller and update its
   * state: if the controller was considered as down, calls the
   * {@link ControllerStateChangedCallback#onControllerUp(const ControllerInfo&)}
   * @param ctrlAddr controller socket address to be updated
   * @param newTime time to associate to this controller. Typically the result
   * of gettimeofday()
   * 
   * @see ControllerStateChangedCallback
   */
  void setControllerResponsed(const SocketAddress& ctrlAddr, int64_t newTime);

  /**
   * Updates all controllers state according to the given time.<br>
   * Iterates through the controller list and (if the controller was not already
   * considered as down) if the lastTimeSeen value is older than the given time
   * minus controller_timeout, the controller will the be considered as
   * failing, and
   * {@link ControllerStateChangedCallback#onControllerDown(const ControllerInfo&)}
   * will be called.
   * @param currentTime reference time to compare to last time seen values.
   * Typically the result of gettimeofday()
   */
  void lookForDeadControllers(int64_t currentTime);

  /**
   * Forces the given controller to be considered as down.<br>
   * If the given controller was already down, does nothing. Otherwise, marks it
   * as dead and calls
   * {@link ControllerStateChangedCallback#onControllerDown(const ControllerInfo&)}
   * @param ctrlAddr controller socket address to consider as failing
   */
  void setControllerDown(const SocketAddress& ctrlAddr);

  /**
   * Returns a 'safe' read-only iterator on controllers' ControllerInfo.<br>
   * Creates a copy of the controller hashmap keys and returns a iterator on it.
   * This way, the iterator will not be affected by hashmap operations
   * @return a vector of all monitored controllers
   */
  std::vector<SocketAddress> getControllerAddresses() const;

private:
  /**
   * UDP addresses identifying controllers and associater controller and
   * states
   */
  std::map<SocketAddress, ControllerAndState> controllers;

  /** To manage controllers concurrent accesses */
  static CriticalSection                      controllers_CS;

  /**
   * Delay after which a controller will be considered as failing if it did not
   * respond to pings
   */
  int                                         controller_timeout;
  /** Actions to take when a controller goes down or back-alive */
  ControllerStateChangedCallback*             callback_ptr;
};
} //namespace CarobNS
#endif /* WATCHED_CONTROLLERS_H_ */
