/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#include "BigDecimal.hpp"
#include "DriverSocket.hpp"

#include "StringCodecs.hpp"
#include "CarobException.hpp"
#include "Common.hpp"

#include <limits>
#include <locale>
#include <sstream> //for wostringstream

using std::wstring;

using namespace CarobNS;

void BigDecimal::initToZero()
{
  byteArrayLength = 0;
  byteArray = NULL;
  scale = 0;
  signum = 0;
}

BigDecimal::BigDecimal()
{
  // mpz_init must be called in every constructor and only there. If mpz_init is
  // called multiple times, a memory leak will appear !
  mpz_init(unscaled_value);
  initToZero();
}

BigDecimal::BigDecimal(const std::wstring& wstr) throw (ConversionException, UnexpectedException)
{
  mpz_init(unscaled_value);
  createFromString(StaticCodecs::toString(wstr));
}

BigDecimal::BigDecimal(const std::string& str) throw (ConversionException, UnexpectedException)
{
  mpz_init(unscaled_value);
  createFromString(str);
}

void BigDecimal::createFromString(const std::string& str)
    throw (ConversionException, UnexpectedException)
{
  initToZero();

  std::locale currentLocale = std::locale();

  std::string unscaledStr = str;
  //TODO: multi-char decimal points
  int strSize = str.length();
  char currentChar = 0;
  bool decimalPointFound = false;
  for (int i=0; i<strSize; i++)
  {
    currentChar = unscaledStr[i];
    if (!isspace(currentChar, currentLocale))
    {
      if (currentChar == userDecimalPoint)
      {
        if (decimalPointFound)
        {
          //double decimal point => exception
          throw ConversionException(L"Detected more than one decimal point [" +
           std::wstring(1, userDecimalPoint) + L"] in BigDecimal string ["+ fromString(str) + L"]", CONV_NUM_RANGE);
        }
        // remove the decimal point (actually replace it by a space)
        unscaledStr.replace(i, 1, " ");
        decimalPointFound = true;
      }
      else if (decimalPointFound)
        scale++;
    }
  }
  if (mpz_set_str(unscaled_value, unscaledStr.c_str(), 10) != 0)
  {
    throw ConversionException(fromString(str) + L"does not represent a valid BigDecimal value", L"FIXME");
  }
  int uv_sign = mpz_sgn(unscaled_value);
  if (uv_sign < 0)
  {
    mpz_neg(unscaled_value, unscaled_value);
    signum = -1;
  }
  else if (uv_sign == 0)
    signum = 0;
  else
    signum = 1;
}

// FIXME: this operator is possibly bugged for 64bits machines
BigDecimal::BigDecimal(const int32_t arg)
{
  mpz_init(unscaled_value);
  initToZero();
  if (arg > 0)
    signum = 1;
  else if (arg < 0)
    signum = -1;
  else
    return;
  mpz_set_si(unscaled_value, arg);
  if (signum<0) //take the abs()
    mpz_neg(unscaled_value, unscaled_value);
}

BigDecimal::BigDecimal(const int64_t arg)
{
  mpz_init(unscaled_value);
  initToZero();
  if (arg > 0)
    signum = 1;
  else if (arg < 0)
    signum = -1;
  else
    return;
  // no setter for long long => import as an array of 1 int64...
  int64_t import = arg;
  bool handleMinLL = false;
  if (signum < 0) //handle sign bit & min int64 correcly
  {
    import = -arg;
    if (arg == std::numeric_limits<int64_t>::min())
    {
      // special value, its abs() cannot be represented in an int64
      // => set to max int64 and put flag so we will add 1 after the import
      import = std::numeric_limits<int64_t>::max();
      handleMinLL = true;
    }
  }
  mpz_import (unscaled_value,
              1,
              0, // only one "64bits word" to import => order doesn't matter
              sizeof(int64_t),
              0, // native endianness
              0, // sign bit has been removed by the previous mask
              &import);
  if (handleMinLL) // special case, see above min int64 handling
    mpz_add_ui(unscaled_value, unscaled_value, 1);
}

#if 0
BigDecimal::BigDecimal(const float arg)
{
  mpz_init(unscaled_value);
  //TODO
}

BigDecimal::BigDecimal(const double)
{
  mpz_init(unscaled_value);
  //TODO
}
#endif

BigDecimal::BigDecimal(const BigDecimal& ref)
{
  mpz_init(unscaled_value);
  *this = ref; //make use of = operator
}

BigDecimal::BigDecimal(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  mpz_init(unscaled_value);
  initToZero();
  //1. Read intVal:
  //1.1 Read intValLength
  int32_t lengthRead;
  input>>lengthRead;
  //We receive the length as an int32 but internally, we have an array of bytes,
  //which length actually is of size_t type
  byteArrayLength = (size_t)lengthRead;
  if (byteArrayLength != 0)
  {
    byteArray = new java_byte[byteArrayLength];
    //1.2 Compute padding
    unsigned int idx = 0;
    int32_t wordRead = 0;
    unsigned int padding = byteArrayLength%4;
    //If there is a padding, we must read the first 'padding' bytes
    //so we are aligned for the rest of the bytes
    if (padding > 0)
    {
      input>>wordRead;
      for (idx=0; idx<padding; idx++)
      {
        byteArray[idx] = static_cast<java_byte>((wordRead>>(8*(padding-idx-1)))&0xFF);
      }
    }
    //1.3 Read the byte array from integers
    //we start from the first aligned byte
    for (; idx<byteArrayLength; idx+=4)
    {
      input>>wordRead;
      byteArray[idx]   = static_cast<java_byte>((wordRead>>24)&0xFF);
      byteArray[idx+1] = static_cast<java_byte>((wordRead>>16)&0xFF);
      byteArray[idx+2] = static_cast<java_byte>((wordRead>> 8)&0xFF);
      byteArray[idx+3] = static_cast<java_byte> (wordRead     &0xFF);
    }
  }

  // 1.4 read sign
  input>>this->signum;
  
  //2. Read scale
  // TODO: what if this scale is negative ? from java it should never be, but
  // should we check ? because afterwards, we do implicit conversions to uint
  // see toBigInteger() or operator double() 
  input>>this->scale;

  //3. Compute gmp value
  if (signum != 0)// 0 signum means 0 value => no useless computations !
  {
    // convert byte array into and unscaled int
    mpz_import(unscaled_value, byteArrayLength, 1 /*MSB 1st*/,
          sizeof(java_byte), 1 /*MSB 1st*/, 0, byteArray);
  }
  
  // TODO: if ODBC/odbsequoia doesn't need the byte array, we should
  // delete and throw away the byte array now.
}

BigDecimal::~BigDecimal()
{
  mpz_clear(unscaled_value);
  if (byteArrayLength > 0 && byteArray != NULL)
  {
    delete[] byteArray;
  }
}

BigDecimal& BigDecimal::operator = (const BigDecimal& ref)
{
  signum = ref.signum;
  scale = ref.scale;
  mpz_set(unscaled_value, ref.unscaled_value);
  byteArrayLength = ref.byteArrayLength;
  if (byteArrayLength != 0)
  {
    byteArray = new java_byte[byteArrayLength];
    memcpy(byteArray, ref.byteArray, byteArrayLength*sizeof(byteArray[0]));
  }
  else
  {
    byteArray = NULL;
  }
  return *this;
}

bool BigDecimal::operator == (const BigDecimal& ref) const
{
  return (signum == ref.signum
      && scale == ref.scale
      && mpz_cmp(unscaled_value, ref.unscaled_value) == 0);
}

void BigDecimal::toBigInteger(mpz_t result) const
{
  if (scale == 0)
  {
    mpz_set(result, unscaled_value);
    return;
  }
  // TODO: optimization = if scale > number of digits in unscaled_value, return 0

  // TODO: multiply by 10^(-scale) instead of dividing by 10^(scale)
  //we have to apply scale
  mpz_t scaler; mpz_init(scaler);                       //int scaler;
  mpz_ui_pow_ui(scaler, 10, scale);                     //scaler = 10^scale;

  mpz_tdiv_q(result, unscaled_value, scaler);              //ret = unscaled_value / scaler;
  mpz_clear(scaler);
}

int* BigDecimal::toIntArray(mpz_t scaled_int, int* mag_length) const
{
  size_t sizeInInts = mpz_size(scaled_int) * sizeof(mp_limb_t) / sizeof(int);
  int* mag = new int[sizeInInts];
  size_t exported = 0;
  mpz_export(static_cast<void*>(mag), &exported,
      1,/*big-endian*/
      sizeof(int), /*words of integers*/
      0, /*native word endianess*/
      0, /*full words*/
      scaled_int);
  *mag_length = static_cast<int>(exported);
  return mag;
}

BigDecimal::operator wstring() const
{
  return toString(std::locale::classic()); //use user-defined locale
}

// FIXME: this function can be optimized: lots of front-inserts can be avoided
std::wstring BigDecimal::toString(const std::locale& loc) const
{
  wchar_t decimalPoint = std::use_facet< std::numpunct<wchar_t> >(loc).decimal_point();
  wchar_t thousandsSep = std::use_facet< std::numpunct<wchar_t> >(loc).thousands_sep();
  bool groupIt = std::use_facet< std::numpunct<wchar_t> >(loc).grouping() != "";
  //convert to string, then to wstring (no direct mpz->wstring converter)
  // First compute space needed by the resulting string
  // number of digits in the unscaled value
  size_t nbDigits = mpz_sizeinbase(unscaled_value, 10);
  // the real size needed by the string
  int size_needed = static_cast<int>(nbDigits);
  // compute room for heading zeros
  if (scale > size_needed)
    size_needed = scale;
  // add room for the decimal point
  if (scale != 0)
    //TODO: rather take decimal point length in case of multi char decimal point
    size_needed++;
  // space for the heading '-' (minus) sign
  if (signum < 0)
    size_needed++;
  // and room for the possible thousand separators
  int nbDec = nbDigits-scale;
  if (nbDec > 0 && groupIt)
    size_needed += nbDec/3 - (nbDec%3 == 0 ? 1 : 0);
  char* as_string = new char[size_needed+1]; // +1 for the \0

  gmp_snprintf(as_string, nbDigits+1, "%Zd", unscaled_value);

  wstring sRet = fromString(as_string);
  delete[] as_string;
  int sRetLength = static_cast<int>(sRet.length());
  if (scale != 0)
  {
    if (sRetLength <= scale)
    {
      // add heading zeros as needed
      sRet.insert(static_cast<size_t>(0),
        static_cast<size_t>(scale - sRetLength + 1), L'0');
    }
    // insert the decimal point
    sRet.insert(sRet.length()-scale, 1, decimalPoint);
  }
  //insert thousand separators if needed
  if (nbDec > 0 && groupIt)
  {
    int tCount = 0; // counts the number of digits starting from the least significant decimal
    int idx = sRet.length();
    if (scale > 0)
      idx = idx - scale-1; // don't count fractional part and separator
    while (idx > 0)
    {
      if (tCount>0 && tCount%3 == 0)
        sRet.insert(idx, 1, thousandsSep);
      tCount++;
      idx--;
    }
  }
  // and the optional sign
  if (signum < 0)
    sRet.insert(0, L"-");

  return sRet;
}

// FIXME: this operator is possibly bugged for 64bits machines
BigDecimal::operator int32_t() const throw (ConversionException)
{
  if (signum == 0)
    return 0;
  mpz_t scaled_int; mpz_init(scaled_int);
  toBigInteger(scaled_int); // truncate value
  if (mpz_size(scaled_int) == 0) // ie. if (scaled_int == 0)
  {
    mpz_clear(scaled_int);
    return 0;
  }
  // Check that we can convert to int
  // TODO: the same with only one "mpz_fits" and a test on int min
  if (signum>0)
  {
    if (!mpz_fits_sint_p(scaled_int))
    {
      mpz_clear(scaled_int);
      throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
          + L" is to big to be converted into an int", CONV_NUM_RANGE));
    }
  }
  else
  {
    // We have to negate the value for the test
    mpz_t tmp; mpz_init(tmp);
    mpz_neg(tmp, scaled_int);
    if (!mpz_fits_sint_p(tmp))
    {
      mpz_clear(tmp);
      mpz_clear(scaled_int);
      throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
          + L" is to big to be converted into an int", CONV_NUM_RANGE));
    }
    mpz_clear(tmp);
  }
  int mag_length = -1;
  int* mag = toIntArray(scaled_int, &mag_length);
  if (mag_length <= 0)
  {
    mpz_clear(scaled_int);
    throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
        + L" could not be converted to int", L"FIXME"));
  }
  int res = mag[mag_length-1];
  delete[] mag;
  mpz_clear(scaled_int);

  if (signum < 0)
    res = -res;
  return res;
}


//TODO: optimize this horribly slow stuff
BigDecimal::operator int64_t() const throw (ConversionException)
{
  if (signum == 0)
    return 0;
  mpz_t scaled_int; mpz_init(scaled_int);
  toBigInteger(scaled_int);
  if (mpz_size(scaled_int) == 0) // ie. if scaled_int == 0
  {
    mpz_clear(scaled_int);
    return 0;
  }
  // check size
  size_t nbBitsNeeded = mpz_sizeinbase(scaled_int, 2);
  if (nbBitsNeeded+1 > sizeof(int64_t)*8) //+1 for the sign bit
  {
    bool numTooBig = true;
    // Handle special "min int64" value:
    if (signum < 0 && nbBitsNeeded == sizeof(int64_t)*8)
    {
      // min int64 is the only value that requires 64bits and which can be
      // represented by a signed int64 (which will be 10000...(63 zeros total)
      // So we have to compare the unscaled_value to abs(numeric_limits<long long>::min())
      // in order to know if we can cast the number
      // TODO: is there an optimized way to do it ?
      BigDecimal minInt64(std::numeric_limits<int64_t>::min());
      if (minInt64 == *this)
        numTooBig = false;
    }
    if (numTooBig)
    {
      mpz_clear(scaled_int);
      throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
          + L" is too big to be converted into an int64_t", CONV_NUM_RANGE));
    }
  }
  int mag_length = -1;
  int* mag = toIntArray(scaled_int, &mag_length);
  if (mag_length <= 0) //should not happen, but safer
  {
    mpz_clear(scaled_int);
    throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
        + L" could not be converted to int64", L"FIXME"));
  }
  int64_t result = 0;
  for (int i = 0; i < mag_length; i++)
  {
    // min int64 special value will do an overflow that magically falls back
    // to 0x80000000
    result = (result<<(8*sizeof(int))) + (mag[i]&0xffffffffLL);
  }
  delete[] mag;
  mpz_clear(scaled_int);

  if (signum < 0)
    result = -result;
  return result;
}

//TODO: optimize this horribly slow stuff
BigDecimal::operator uint64_t() const throw (ConversionException)
{
  if (signum == 0)
    return 0;
  if (signum < 0)
    throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
        +L" is negative ! it cannot be converted into an int64_t", CONV_NUM_RANGE));
  mpz_t scaled_int; mpz_init(scaled_int);
  toBigInteger(scaled_int);
  if (mpz_size(scaled_int) == 0) // ie. if scaled_int == 0
  {
    mpz_clear(scaled_int);
    return 0;
  }
  // check size
  size_t nbBitsNeeded = mpz_sizeinbase(scaled_int, 2);
  if (nbBitsNeeded > (sizeof(int64_t)*8))
  {
    mpz_clear(scaled_int);
    throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
        + L" is too big to be converted into an int64_t", CONV_NUM_RANGE));
  }
  int mag_length = -1;
  int* mag = toIntArray(scaled_int, &mag_length);
  if (mag_length <= 0) //should not happen, but safer
  {
    mpz_clear(scaled_int);
    throw (ConversionException(L"BigDecimal " + static_cast<wstring>(*this)
        + L" could not be converted to int64", L"FIXME"));
  }

  uint64_t result = 0;
  for (int i = 0; i < mag_length; i++)
  {
    result = (result<<(8*sizeof(int))) + (mag[i]&0xffffffffLL);
  }
  delete[] mag;
  mpz_clear(scaled_int);
  return result;
}

BigDecimal::operator float() const throw (ConversionException)
{
  // TODO: check that we can represent the big decimal by comparing:
  // log2(this) < log2(std::numeric_limits<float>::max())
  // idem with min, -min and max negative values
    return (static_cast<float>(operator double()));
}

BigDecimal::operator double() const throw (ConversionException)
{
  // TODO: check that we can represent the big decimal by comparing:
  // log2(this) < log2(std::numeric_limits<double>::max())
  // idem with min, -min and max negative values

                                                      // understandable c-style translation...

  mpf_t res; mpf_init(res);                           // float res;
  mpf_t unscaled; mpf_init(unscaled);                 // float unscaled;
  mpf_t scaler; mpf_init(scaler);                     // int scaler;

  // initialize res and scale to double values so the mpf_t will have the right precision
  mpf_set_d(res, static_cast<double>(0));             // res = (double)0;
  mpf_set_d(scaler, static_cast<double>(10));         // scaler = (double)10;
  mpf_pow_ui(scaler, scaler, scale);                  // scaler = scaler^scale;
  mpf_set_z(unscaled, unscaled_value);                // unscaled = unscaled_value;
  mpf_div(res, unscaled, scaler);                     // ret = unscaled/scaler;
  if (signum < 0)
    mpf_neg(res, res);                                // ret = -ret;
  double ret = mpf_get_d(res);

  mpf_clear(scaler);
  mpf_clear(unscaled);
  mpf_clear(res);

  return ret;
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
