/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2007 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#include "BufferedSocket.hpp"
#include "SystemDependantDefs.hpp"

#ifdef __MINGW32__
  #include <ws2tcpip.h>
#else
  #include <sys/types.h>   //  socket(), send(), etc.
  #include <sys/socket.h>  //         "
#endif

using namespace CarobNS;

/** Returns the number of octets read or the (negative) error from recv() */
int BufferedSocket::recvFully(void *buf, const int len, const int flags)
    const throw (SocketIOException, UnexpectedException)
{
  int alreadyRead = 0; // Keeps the whole size read as the offset in
                       // user buffer
  int remainingBytesToRead = len;
  while (!canceled && (remainingBytesToRead > 0))
  {
    int availableBytes = read_cache.getDataSize();

    // Empty cache? (always true except at first iteration)
    if (availableBytes < 1)
    {
      // Application has eaten everything we received so far in the
      // read cache and it wants even more: perfect time to flush()
      // the write cache, make the controller react and send us some
      // new data.
      flush();

      if (remainingBytesToRead > READ_BUFFER_SIZE-1)
        // 1. our read cache is empty
        // 2. we add a memcpy()
        // 3. our recv(READ_BUFFER_SIZE) would be LESS eager that the
        //    non-buffered one!
        // => our purpose is defeated, let's give up and let the
        //    simpler, NON-buffered code do the job
        return alreadyRead
          + JavaSocket::recvFully(static_cast<java_byte*>(buf)+alreadyRead, remainingBytesToRead, flags);

      availableBytes = fillReadBuffer(flags);
    }

    // toBeReadThisTime = min (remainingBytesToRead, availableBytes)
    int toBeReadThisTime =
      remainingBytesToRead < availableBytes
      ? remainingBytesToRead // we have enough in stock.
      : availableBytes;      // not enough: copy all what we already have and loop

    memcpy(static_cast<java_byte*>(static_cast<java_byte*>(buf)+alreadyRead),
           read_cache.getData(),
           toBeReadThisTime);
    read_cache.discard(toBeReadThisTime);
    alreadyRead += toBeReadThisTime;
    remainingBytesToRead -= toBeReadThisTime;
  }
  // We were shutdown()
  if (!connected)
    throw SocketIOException(L"JavaSocket has been shutdown, read aborted");
  return alreadyRead;
}

int BufferedSocket::fillReadBuffer(int recvFlags) const throw (SocketIOException, UnexpectedException)
{
  // Wait for incoming data
  int pollRetVal = pollOnSingleFd(socket_fd, 1000); // 1000 = 1 second timeout
  int readThisTime = 0;
  if (pollRetVal > 0) // data is ready to be read
  // if (pollRetVal == 0) => timeout, no response, do nothing
  {
    read_cache.reset();
    int readThisTime = recv(socket_fd, read_cache.getData(), READ_BUFFER_SIZE, recvFlags);
    read_cache.setDataSize(readThisTime);
    if (readThisTime < 0)
      throw SocketIOException(L"Could not read from socket.");
    if (readThisTime == 0)
      throw SocketIOException(L"Peer reset connection while we were receiving data");
  }
  return readThisTime;
}

void BufferedSocket::sendToSocket(const wchar_t fctName[],
    const void* buf, int len, int flags) const
    throw (SocketIOException, UnexpectedException)
{
  // not enough room in the cache?
  if (len > write_cache.getRemainingSize())
    // then make some room!
    flush();

  // still not enough room in the cache?
  if (len > WRITE_BUFFER_SIZE-1)
  { // Caching so big data would cost:
    // 1. more system calls 2. memcpy() cost
    // => just don't do cache.
    JavaSocket::sendToSocket(fctName, buf, len, flags);
    return;
  }

  // Here we have enough cache room. Put data in the cache (and delay
  // system send())
  write_cache.append(static_cast<const java_byte*>(buf), len);

}

void BufferedSocket::flush() const throw (SocketIOException, UnexpectedException)
{
  const wchar_t fctName[] = L"BufferedSocket::flush";
  
  if (write_cache.getDataSize() == 0)
  {
    return;
  }
  int ret = send(socket_fd,
            static_cast<const void*>(write_cache.getData()),
            write_cache.getDataSize(),
            SOCKET_SEND_FLAGS);
  if ( ret == -1 )
  {
    throw SocketIOException(std::wstring(fctName) + L": could not write to socket");
  }
  if (ret != write_cache.getDataSize())
  {
    // should never happen
    throw SocketIOException(std::wstring(fctName) + L": only " + toUserString(ret)
        + L"/" + toWString(write_cache.getDataSize())
        + L" bytes could be send to socket");
  }
  write_cache.reset();
}

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
