/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#include "CarobException.hpp"

#include "Common.hpp"
#include "DriverSocket.hpp"

#include <sstream>

#include <exception>

#ifdef __GLIBC__  // from features.h
#define GLIB_BACKTRACE
#include <execinfo.h>
#include <stdlib.h>
#endif // __GLIBC__

using std::wstring;

using namespace CarobNS;

/** constructs the BackTrace of the current place */
BackTrace::BackTrace()
{
#ifdef GLIB_BACKTRACE
  void * stack_pointers[100];
  size_t size = backtrace (stack_pointers /* set it */, 100);
  char **cstrings = backtrace_symbols (stack_pointers /* use it */, size);
  for (size_t i=0; i<size; i++)
    this->push_back(cstrings[i]);
  free(cstrings);
#endif
}


CarobException::~CarobException()
{
  if (causePtr != NULL)
  {
    delete causePtr;
    causePtr = NULL;
  }
}

/** Copy constructor. NOT chaining! */
CarobException::CarobException(const CarobException& ce) :
  message(ce.message),
  stackTrace(ce.stackTrace),
  SQLState(ce.SQLState),
  errorCode(ce.errorCode),
  backTrace(ce.backTrace)
{
  if (ce.causePtr != NULL)
  { // copy chain by recursively calling copy constructor
    causePtr = new CarobException(*(ce.causePtr));
  }
  else
  {
    causePtr = NULL;
  }
}

CarobException::CarobException(const DriverSocket& sock)
{
  bool hasNextException;

  sock>>message;
  sock>>hasNextException;
  if (hasNextException)
    causePtr = new CarobException(sock);
  else
    causePtr = NULL;
  
  // receive stackTrace
  int stLength;
  sock>>stLength;
  for (int i = 0; i < stLength; i++)
  {
    stackTrace.push_back(StackTraceElement(sock));
  }
  sock>>SQLState;
  sock>>errorCode;
}

CarobException::operator std::wstring() const
{
  std::wostringstream woss;
  woss << *this;
  return woss.str();
}

void UnexpectedException::convertUnexpected()
{
  try
  {
    if (isErrorEnabled())
      logError(L"convertUnexpectedException",
               L"Unexpected exception caught...");
    throw;
  }
  catch(CarobException& ce)
  {
    std::wostringstream woss;
    woss << ce;
    if (isErrorEnabled())
        logError(L"convertUnexpectedException", woss.str());
    throw UnexpectedException(std::wstring(L"Unexpected CarobException: ")
                                           + ce.description(),
                              ce.getSQLState(), &ce);
  }
  catch(std::exception& stde)
  {
    if (isFatalEnabled())
        logFatal(L"convertUnexpectedException",
                 L"std::exception caught");
    throw UnexpectedException(L"Unexpected std::exception "
                              + CarobNS::StaticCodecs::fromASCII(stde.what()));
  }
  catch(...)
  {
    if (isFatalEnabled())
        logFatal(L"convertUnexpectedException",
                 L"non-CarobException caught");
    throw UnexpectedException(L"Unknown exception type");
  }
}

StackTraceElement::StackTraceElement(const DriverSocket& ds)
    throw (SocketIOException, UnexpectedException)
{
  ds>>declaringClass;
  ds>>methodName;
  ds>>fileName;
  ds>>lineNumber;
}


std::wostream& CarobNS::operator<< (std::wostream& oss, const CarobNS::StackTraceElement& ste)
{
  oss << L"at " << ste.declaringClass << L"." << ste.methodName
      << L"(" << ste.fileName << L":" << ste.lineNumber << L")";
  return oss;
}

std::wostream& CarobNS::operator<< (std::wostream& outstream, const CarobNS::StackTrace& st)
{
  if (st.size() == 0)
    outstream << L"  (empty/no remote stack trace)" << std::endl;
  else
  {
    for (CarobNS::StackTrace::const_iterator s = st.begin(); s != st.end(); s++)
        outstream << *s << std::endl;
  }

  return outstream;
}

std::wostream& CarobNS::operator<< (std::wostream& outstream, const CarobNS::BackTrace& stvect)
{
  for (CarobNS::BackTrace::const_iterator btframe = stvect.begin();
      btframe != stvect.end(); btframe++)
  {

    std::string sframe(*btframe);

#ifdef CAROB_CXX_DEMANGLING
    typedef std::string::size_type ssize;
  
    ssize start(sframe.find('('));
    ssize end(sframe.find('+'));
  
    if (std::string::npos != start && std::string::npos != end)
    {   
      start++;
      ssize len(end-start);
      
      std::string mangled(sframe.substr(start, len));
  
      int status;
      char *demangled(abi::__cxa_demangle(mangled.c_str(), 0, 0, &status));
  
      if (demangled)
      {
        sframe.replace(start, len, demangled);
        free(demangled);
      }
    }
#endif
    // not demangled frames is better than nothing: always print
    outstream << CarobNS::StaticCodecs::fromASCII(sframe) << std::endl;
  }

  return outstream;
}


std::wostream& CarobNS::operator<<(std::wostream& woss, const CarobNS::CarobException& ce)
{
  woss << ce.description();
  woss << L"; SQL state: " << ce.getSQLState() << std::endl;
  woss << ce.getBackTrace();
  woss << L"      -- remote stack -- " << std::endl;
  woss <<  ce.getStackTrace();

  const CarobNS::CarobException* nextex_p = ce.getNext();
  
  if (nextex_p != NULL)
  {
    woss << L"   ---  chained CarobException  --- ";
    woss << *nextex_p;
  }

  return woss;
}


// temporary fix for CAROB-56
namespace
{
  class STC
  {
    std::unexpected_handler old;
  public:
    STC()
    {
      old = std::set_unexpected(UnexpectedException::convertUnexpected);
    }
    ~STC()
    {
      std::set_unexpected(old);
    }
  };

  // brutal
  STC install_our_global_handler;
}

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
