/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Csaba Simon
 */

#include "Common.hpp"

#include "StringCodecs.hpp"

//LOG4CXX includes
#ifdef CAROB_LOG4CXX_NAME
#include "StringCodecs.hpp" //for toString
#include <log4cxx/logger.h>
#endif

#ifndef __MINGW32__
#include <stdint.h> // for uint8_t, int32_t, uint32_t, int64_t, uint64_t
// Alternative:
// #include <cstdint>

// Warning: this header and all the types it defines (including "long
// long") are part of C99 and are MISSING from C++98.

// See:
// - "Incompatibilities Between ISO C and ISO C++" - David R. Tribble
// - C++ Technical Report 1 (TR1)
#endif
#include <sys/time.h> // for gettimeofday()
#include <iostream>
#include <sstream> //for wostringstream
#include <stdexcept>
#include <locale>


using std::wstring;
using std::string;
using std::wcerr;
using std::endl;


#ifdef CAROB_LOG4CXX_NAME

using namespace log4cxx;

// need two steps to delay expansion of argument
#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

namespace
{
  LoggerPtr logger(Logger::getLogger(TOSTRING(CAROB_LOG4CXX_NAME)));
}

namespace CarobNS {
LoggerPtr& getLogger()
{
  return logger;
}
}

#else // CAROB_LOG4CXX_NAME

namespace {
  /** The current log level (set by setLogLevel) */
  CarobNS::LogLevel currentLogLevel = CarobNS::LOG_LEVEL_WARN;
}
#endif //  CAROB_LOG4CXX_NAME



void CarobNS::setLogLevel(const LogLevel l)
{
#ifdef CAROB_LOG4CXX_NAME
  if (l == LOG_LEVEL_DEBUG)
    logger->setLevel(log4cxx::Level::DEBUG);
  else if (l == LOG_LEVEL_INFO)
    logger->setLevel(log4cxx::Level::INFO);
  else if (l == LOG_LEVEL_WARN)
    logger->setLevel(log4cxx::Level::WARN);
  else if (l == LOG_LEVEL_ERROR)
    logger->setLevel(log4cxx::Level::ERROR);
  else if (l == LOG_LEVEL_FATAL)
    logger->setLevel(log4cxx::Level::FATAL);
  else if (l == LOG_LEVEL_OFF)
    logger->setLevel(log4cxx::Level::OFF);
#else
  currentLogLevel = l;
#endif
}

void CarobNS::logDebug(const wstring& fctName, const wstring& s)
{
#ifdef CAROB_LOG4CXX_NAME
  // Log4cxx <= 0.9.7 doesn't accept wstrings - we have to convert the string
  LOG4CXX_DEBUG(logger, StaticCodecs::toString(fctName + L": " + s));
#else
  if (isDebugEnabled())
    wcerr << DEBUG_PRE << L'(' << fctName << L") " << s << endl;
#endif
}
void CarobNS::logInfo(const wstring& fctName, const wstring& s)
{
#ifdef CAROB_LOG4CXX_NAME
  LOG4CXX_INFO(logger, StaticCodecs::toString(fctName + L": " + s));
#else
  if (isInfoEnabled())
    wcerr << INFO_PRE << L'(' << fctName << L") " << s << endl;
#endif
}
void CarobNS::logWarn(const wstring& fctName, const wstring& s)
{
#ifdef CAROB_LOG4CXX_NAME
  LOG4CXX_WARN(logger, StaticCodecs::toString(fctName + L": " + s));
#else
  if (isWarnEnabled())
    wcerr << WARN_PRE << L'(' << fctName << L") " << s << endl;
#endif
}
void CarobNS::logError(const wstring& fctName, const wstring& s)
{
#ifdef CAROB_LOG4CXX_NAME
  LOG4CXX_ERROR(logger, StaticCodecs::toString(fctName + L": " + s));
#else
  if (isErrorEnabled())
    wcerr << ERROR_PRE << L'(' << fctName << L") " << s << endl;
#endif
}
void CarobNS::logFatal(const wstring& fctName, const wstring& s)
{
#ifdef CAROB_LOG4CXX_NAME
  LOG4CXX_FATAL(logger, StaticCodecs::toString(fctName + L": " + s));
#else
  if (isFatalEnabled())
    wcerr << FATAL_PRE << L'(' << fctName << L") " << s << endl;
#endif
}

bool CarobNS::isDebugEnabled()
{
#ifdef CAROB_LOG4CXX_NAME
  return log4cxx::Level::DEBUG->isGreaterOrEqual(logger->getLevel());
#else
  return currentLogLevel <= LOG_LEVEL_DEBUG;
#endif
}
bool CarobNS::isInfoEnabled()
{
#ifdef CAROB_LOG4CXX_NAME
  return log4cxx::Level::INFO->isGreaterOrEqual(logger->getLevel());
#else
  return currentLogLevel <= LOG_LEVEL_INFO;
#endif
}
bool CarobNS::isWarnEnabled()
{
#ifdef CAROB_LOG4CXX_NAME
  return log4cxx::Level::WARN->isGreaterOrEqual(logger->getLevel());
#else
  return currentLogLevel <= LOG_LEVEL_WARN;
#endif
}
bool CarobNS::isErrorEnabled()
{
#ifdef CAROB_LOG4CXX_NAME
  return log4cxx::Level::ERROR->isGreaterOrEqual(logger->getLevel());
#else
  return currentLogLevel <= LOG_LEVEL_ERROR;
#endif
}
bool CarobNS::isFatalEnabled()
{
#ifdef CAROB_LOG4CXX_NAME
  return log4cxx::Level::FATAL->isGreaterOrEqual(logger->getLevel());
#else
  return currentLogLevel <= LOG_LEVEL_FATAL;
#endif
}

wstring CarobNS::trim(const wstring& source, 
             const wchar_t* delims)
{
  wstring result(source);
  string::size_type index = result.find_last_not_of(delims);
  if(index != string::npos)
    result.erase(++index);

  index = result.find_first_not_of(delims);
  if(index != string::npos)
    result.erase(0, index);
  else
    result.erase();
  return result;
}

namespace {

  std::locale tryuserlocale()
  {
    try {
      return std::locale("");
    } catch (std::runtime_error& ) {

      // Cannot use log4cxx logging functions here since they use the
      // codec we are currently trying to initialize. Moreover log4cxx
      // static may not have been yet initialized.
      std::wcerr << __WFILE__ L":tryuserlocale(): "
        L"Missing user-preferred locale (check LANG). "
        "Falling back on current global locale." << std::endl;
      return std::locale();
    }
  }
}

const CarobNS::MBSCodec CarobNS::StaticCodecs::ascii_codec(std::locale::classic());

// init order matters, since user_codec may be used to print error
// messages in case of failures
const CarobNS::MBSCodec CarobNS::StaticCodecs::user_codec(tryuserlocale());

// decimal point initialization
wchar_t CarobNS::userDecimalPoint =
    std::use_facet<std::numpunct<wchar_t> >(tryuserlocale()).decimal_point();
wchar_t CarobNS::controllerDecimalPoint = 
    std::use_facet<std::numpunct<wchar_t> >(std::locale::classic()).decimal_point();

//Can this be factorized with ParameterStatement ?
template <class T> wstring CarobNS::toWString(const T& t)
{
  std::wostringstream buffer;
  buffer.imbue(std::locale::classic());
  buffer << t;
  return buffer.str();        
}
// Explicit instantiation to avoid Borland/Cfront/whatever else
// non-portable solution to the template instantiation problem.
// Don't instantiate fuzzy types like (unsigned?) "char" or "long"
// they cause portability headaches (see CAROB-116).
// Don't forget to update javadoc when changing this
template wstring CarobNS::toWString<int16_t>(const int16_t&);
template wstring CarobNS::toWString<uint16_t>(const uint16_t&);
template wstring CarobNS::toWString<int32_t>(const int32_t&);
template wstring CarobNS::toWString<uint32_t>(const uint32_t&);
template wstring CarobNS::toWString<int64_t>(const int64_t&);
template wstring CarobNS::toWString<uint64_t>(const uint64_t&);
template wstring CarobNS::toWString<float>(const float&);
template wstring CarobNS::toWString<double>(const double&);

// locale dependant conversions
template <class T> wstring CarobNS::toUserString(const T& t)
{
  std::wostringstream buffer;
  // imbue with user's locale
  buffer.imbue(std::locale());
  buffer << t;
  return buffer.str();        
}

// Explicit instantiation, see toWString() above
template wstring CarobNS::toUserString<int16_t>(const int16_t&);
template wstring CarobNS::toUserString<uint16_t>(const uint16_t&);
template wstring CarobNS::toUserString<int32_t>(const int32_t&);
template wstring CarobNS::toUserString<uint32_t>(const uint32_t&);
template wstring CarobNS::toUserString<int64_t>(const int64_t&);
template wstring CarobNS::toUserString<uint64_t>(const uint64_t&);
template wstring CarobNS::toUserString<float>(const float&);
template wstring CarobNS::toUserString<double>(const double&);

template <class T> bool CarobNS::wstringTo(T& t, const std::wstring& s)
{
  std::wistringstream iss(s);
  return !(iss >> std::skipws >> t).fail();
}
// Explicit instantiation, see toWString() above
template bool CarobNS::wstringTo<int16_t>(int16_t&, const std::wstring&);
template bool CarobNS::wstringTo<uint16_t>(uint16_t&, const std::wstring&);
template bool CarobNS::wstringTo<int32_t>(int32_t&, const std::wstring&);
template bool CarobNS::wstringTo<uint32_t>(uint32_t&, const std::wstring&);
template bool CarobNS::wstringTo<int64_t>(int64_t&, const std::wstring&);
template bool CarobNS::wstringTo<uint64_t>(uint64_t&, const std::wstring&);
template bool CarobNS::wstringTo<float>(float&, const std::wstring&);
template bool CarobNS::wstringTo<double>(double&, const std::wstring&);

wstring CarobNS::replaceAll(wstring src, const wstring &from, const wstring &to)
{
  wstring::size_type found = src.find(from);
  while (found != wstring::npos)
  {
    src.replace(found, from.length(), to);
    found = src.find(from, found+to.length());
  }
  return src;
}

int64_t CarobNS::currentTimeMillis()
{
  struct timeval now;
  struct timezone ignored;
  if (gettimeofday(&now, &ignored) != 0)
  {
    if (isErrorEnabled())
      logError(L"getCurrentTimeInMs", L"gettimeofday failed!");
    return -1;
  }
  return ((int64_t) now.tv_sec)*1000 + now.tv_usec/1000;
}


int CarobNS::currentTimeSeconds()
{
  struct timeval now;
  struct timezone ignored;
  if (gettimeofday(&now, &ignored) != 0)
  {
    if (isErrorEnabled())
      logError(L"getCurrentTimeInMs", L"gettimeofday failed!");
    return -1;
  }
  return now.tv_sec;
}

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
