/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert, Zsolt Simon
 */

#include "ControllerPool.hpp"

#include "DriverSocket.hpp"
#include "Connection.hpp"
#include "Statement.hpp"
#include "ParameterStatement.hpp"
#include "TypeTag.hpp"
#include "Request.hpp"
#include "RequestWithResultSetParameters.hpp"

#include "ConnectionParameters.hpp"
#include "CarobException.hpp"

#include <string>
#include <sstream>


#ifdef __MINGW32__
  #define sleep _sleep
#else
  #include <unistd.h> // sleep()
#endif

using std::wstring;

using namespace CarobNS;

#define RECONNECT_RETRIES 10
#define CONNECT_RETRIES 3

// The reconnection macro:
// When a SocketIOException occurs between these to macros, sleep and try to
// reconnect(). Sleeping time depends on the retry #. Only N retries are
// performed

#define FO_TRY_NTIMES(N)\
for (int foCnt=0; foCnt<N; foCnt++)\
{\
  try\
  {\
    /* Code to monitor will go here */


// TODO: catching non-socket exceptions should probably be done
// outside of the retry for loop. Like this we could 
// try to reconnect() after corruptedConnectionClose()
// WITHOUT trying to re-execute the failed query! See CAROB-125


#define FO_CATCH_NTIMES\
    break; /* no exception means success: get out of the retry for loop */\
  }\
  catch (const SocketIOException& sioe)\
  {\
    if (isInfoEnabled())\
      logInfo(fctName, L"SocketIOException received");\
    ::sleep(static_cast<int>(foCnt*0.5)); /* sleep incrementally */\
    reconnectOrDie();\
    /* All controllers down? Maybe not, maybe it is OUR network	   \
       connection which is down! So let's sleep and try NTIMES */  \
  }\
  catch (const ProtocolException& pe)\
  {\
    corruptedConnectionClose(pe); throw; /* TODO: type rethrown? */ 	\
  }\
  catch (const DriverException& pe)\
  { /* If carob has locally created an exception while communicating \
       with the controller, then the connection is very likely to be
       corrupted */	\
    corruptedConnectionClose(pe); throw; \
  }\
  catch (const UnexpectedException& ue) /* because of CAROB-101 */ \
  {\
    corruptedConnectionClose(ue); throw;	\
  }\
}

Connection::Connection(ConnectionParameters& prms)
    throw (AuthenticationException, NoMoreControllerException,
    VirtualDatabaseUnavailableException, UnexpectedException) :
  driverSocketPtr(NULL),
  autoCommit(true),
  readOnly(false),
  writeExecutedInTransaction(false),
  transactionId(-1),
  mustBeginTransaction(false),
  persistent_connection(false),
  retrieve_sql_warnings(false),
  first_warning(NULL),
  controller_pool(prms.getControllerPool()),
  parameters(prms),
  auto_delete_statements(true)
{
  const wchar_t fctName[] = L"Connection::Connection";
  persistent_connection = parameters.getPersistentConnection();
  retrieve_sql_warnings = parameters.getRetrieveSQLWarnings();
  try
  {
    getConnectionToNewController();
    return;
  }
  // exceptions must be reported to the user
  catch (const CarobException& ce)
  {
    logError(fctName, ce.description());

    // Clean all: we are in constructor, destructor won't be called !
    if (driverSocketPtr)
    {
      controller_pool.deleteRegisteredSocket(driverSocketPtr);
      driverSocketPtr = NULL;
    }
    throw;
  }
  catch (const std::exception& se)
  {
    logFatal(fctName, StaticCodecs::fromASCII(se.what()));

    if (driverSocketPtr)
    {
      controller_pool.deleteRegisteredSocket(driverSocketPtr);
      driverSocketPtr = NULL;
    }
    throw;
  }
  catch (...)
  {
    logFatal(fctName, L"unknown exception type");

    if (driverSocketPtr)
    {
      controller_pool.deleteRegisteredSocket(driverSocketPtr);
      driverSocketPtr = NULL;
    }
    throw;
  }

}

Connection::~Connection()
{
  close();
  // driverSocketPtr is deleted in close()
}

void Connection::sendCommand(const DriverSocket& socket, int command)
    throw (SocketIOException, ProtocolException, UnexpectedException)
{
  int ack = -1;
  socket>>ack;
  if (ack != ControllerPrompt)
    throw ProtocolException(
        L"Protocol corruption while trying to send command: "
        + toUserString(command) + L". Check the previous command");
  socket<<CommandPrefix;
  socket<<static_cast<int32_t>(command);
}

void Connection::getConnectionToNewController() throw (AuthenticationException,
    NoMoreControllerException, VirtualDatabaseUnavailableException, UnexpectedException)
{
  // This will count the number of controllers left to connect to upon
  // VDB not available exceptions
  // No need to synchronize with getController(): at worth, we will not try
  // one controller that just reappeared, or we will try two time the same
  // controller. This is harmless, we just want to prevent endless retry loops
  int numberOfCtrlsLeft = controller_pool.numberOfAliveControllers();
  while (numberOfCtrlsLeft > 0)
  {
    try
    {
      connectToNextController();
      return;
    }
    catch (VirtualDatabaseUnavailableException vdbue)
    {
      numberOfCtrlsLeft--;
    }
  }
  // at this point, we have tried all controllers
  throw VirtualDatabaseUnavailableException(L"Virtual database "
      + parameters.getDatabaseName() + L" not found on any of the controllers");
}

void Connection::connectToNextController()
    throw (AuthenticationException, NoMoreControllerException,
    VirtualDatabaseUnavailableException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::initConnectionToNextController";
  try
  {
    // This must lock the pool synchro to avoid controller list operations
    // between getController() and connect() calls
    {
      LockScope(&controller_pool.getLock());
      connected_controller = controller_pool.getController();
      if (isInfoEnabled())
        logInfo(fctName, L"Authenticating with controller "
                          + static_cast<wstring>(connected_controller));
      //Here is the connection protocol...
      // 1. connect to the controller
      driverSocketPtr = connected_controller.connect(controller_pool);
    }
    // 2. sent the Protocol version information
    *driverSocketPtr<<ProtocolVersion;
    // 3. send the database name
    *driverSocketPtr<<parameters.getDatabaseName();
    // 4. send user/pass
    *driverSocketPtr<<parameters.getUserName();
    *driverSocketPtr<<parameters.getUserPass();
    finalizeConnect();
  }
  catch (ConnectionException connExcpt)
  {
    wstring msg = L"Unable to connect to controller "
                  + static_cast<wstring>(connected_controller)
                  + L" (" + connExcpt.description() + L").";
    if (isErrorEnabled())
      logError(fctName, msg);
    //suspect failing controller
    controller_pool.forceControllerDown(connected_controller);
    closeSocket();
    // Recurse until no more controller
    return connectToNextController();
  }
  catch (SocketIOException sockIOExcpt)
  {
    wstring msg = L"Could not authentify to " + static_cast<wstring>(connected_controller);
    if (isErrorEnabled())
      logError(fctName, msg);
    //suspect failing controller
    controller_pool.forceControllerDown(connected_controller);
    closeSocket();
    // Recurse until no more controller
    return connectToNextController();
  }
  catch (VirtualDatabaseUnavailableException vdbue)
  {
    if (isInfoEnabled())
      logInfo(fctName, L"Virtual database down on controller " + static_cast<wstring>(connected_controller));
    closeSocket();
    controller_pool.setVdbDownOnController(connected_controller);
    throw;
  }
}

void Connection::finalizeConnect() throw (AuthenticationException,
    SocketIOException, VirtualDatabaseUnavailableException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::finalizeConnect";
  bool vdbFound = false,
       authenticated = false;
  *driverSocketPtr >> vdbFound;
  if (!vdbFound)
  {
    wstring vdbReason;
    *driverSocketPtr >> vdbReason;
    throw VirtualDatabaseUnavailableException(L"Connection failed: "+vdbReason);
  }
  *driverSocketPtr >> authenticated;
  if (!authenticated)
  {
    wstring authReason;
    *driverSocketPtr >> authReason;
    throw AuthenticationException(L"Authentication failed: "+authReason);
  }
  // So the controller can correctly parse our requests
  *driverSocketPtr << LINE_SEPARATOR;
  *driverSocketPtr << persistent_connection;
  if (persistent_connection)
  {
    bool success = false;
    *driverSocketPtr >> success;
    if (success)
      *driverSocketPtr >> persistent_connection_id;
    else
      throw AuthenticationException(std::wstring(fctName) + 
          L" No more persistent connections available for given virtual database");
  }
  *driverSocketPtr << retrieve_sql_warnings;
}

void Connection::deleteStatement(Statement* stPtr) throw (DriverException, UnexpectedException)
{
  if (stPtr != NULL) // if null, we don't do anything
  {
    for (std::list<Statement*>::iterator iter = created_statements.begin();
        iter != created_statements.end(); iter++)
    {
      if (*iter == stPtr)
      {
        created_statements.erase(iter);
        delete stPtr;
        return;
      }
    }
    // We shouldn't come here. It means we have been asked to delete a statement
    // we don't own
    throw DriverException(L"deleteStatement: the given statement is not owned by this connection!");
  }
}

void Connection::corruptedConnectionClose(const CarobException& reason)
{
  const wchar_t fctName[] = L"Connection::corruptedConnectionClose(const CarobException&)";

  std::wostringstream buffer;
  buffer << reason;
  logError(fctName, buffer.str());

  // Brutally close network socket FIRST, so other concurrent close() will not
  // try to do the job twice
  try
  {
    logInfo(fctName, L"closing corrupted connection because of" + reason.description());
    this->closeSocket();
  }
  catch (const SocketIOException& sioe)
  {
    std::wostringstream buffer;
    buffer << L"failed to force ->closeSocket(), probably already closed? " << sioe.description();
    logError(fctName, buffer.str());
  }

  // Then call the regular close() to clean-up objects.  It will
  // notice that the socket is already closed and refrain to send the
  // Close command to the controller.
  close();
}


bool Connection::close()
{
  const wchar_t fctName[] = L"Connection::close";

  if (isDebugEnabled())
    logDebug(fctName, L"Cleaning connection to controller " + static_cast<wstring>(connected_controller));

  //Wait until other methods/Commands are done
  LockScope ls(&connectionCS);

  // Clean-up: delete all statements created by this connection unless
  // told not to
  if (auto_delete_statements)
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Deleting statements created by this connection");

    for (std::list<Statement*>::iterator iter = created_statements.begin();
        iter != created_statements.end(); iter++)
    {
      delete *iter;
    }
    created_statements.clear();
  }

  if (driverSocketPtr == NULL)
  {
    if (isInfoEnabled())
      logInfo(fctName, L"driverSocketPtr was NULL (already closed)");
    return false;
  }

  bool closeOK = false;
  try
  {
    // Try to cleanly close the connection (also on controller side)
    if (isDebugEnabled())
      logDebug(fctName, L"Sending Close command to controller " + static_cast<wstring>(connected_controller));
    sendCommand(*driverSocketPtr, Close);
    closeOK = receiveBoolOrException();
  }
  catch (...)
  {
    if (isInfoEnabled())
      logInfo(fctName, L"Could not inform controller we are closing connection");
  }

  // actual system socket closing
  closeSocket();
  return closeOK;
}

void Connection::closeSocket()
{
  const wchar_t fctName[] = L"Connection::closeSocket";

  // this will allow delayed deletion: we will just set driverSocketPtr to NULL
  // in a synchronized block, so concurrent calls will see this NULL and return
  // asap. Effective deletion will be done outside the synchronized block
  DriverSocket* toDelete = driverSocketPtr;
  {
    LockScope ls(&socketCS);
  
    if (isDebugEnabled())
      logDebug(fctName, L"Closing system socket to controller " + static_cast<wstring>(connected_controller));
  
    if (driverSocketPtr == NULL)
    {
      if (isInfoEnabled())
        logInfo(fctName, L"... was NULL (already closed)");
      return;
    }
    // will have to delete it but will do that outside synchronized block
    toDelete = driverSocketPtr;
    // it's dying, so hide it from anyone else
    driverSocketPtr = NULL;
  }
  // Now we can call the destructor without slowing down concurrent calls
  // The system socket closing is done in this destructor
  controller_pool.deleteRegisteredSocket(toDelete);
}

void Connection::setAutoCommit(bool autoCommitPrm)
    throw (SocketIOException, BackendException, ControllerException, 
    ProtocolException, ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::setAutoCommit";

  LockScope ls(&connectionCS);
  
  if (mustBeginTransaction) // so this.autoCommit is false
  {
    if (autoCommitPrm)
    { // Just cancel the need to start a transaction
      autoCommit = true;
      mustBeginTransaction = false;
      return;
    }
  }
    
  // nothing to change
  if (autoCommit == autoCommitPrm)
  {
    return;
  }
  // autocommit false -> true
  if (autoCommitPrm)
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Setting connection in autocommit mode");

    doCommit();
    autoCommit = true;
    return;
  }
  else // autocommit true -> false
  {
    
    mustBeginTransaction = true;
    autoCommit = false;

    if (isDebugEnabled())
      logDebug(fctName, L"autocommit set to false, next transaction will do a 'Begin'");
  }
}

void Connection::beginTransactionIfNeeded() throw (SocketIOException,
    BackendException, ControllerException, DriverException, ProtocolException,
    ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::beginTransactionIfNeeded";
  if (mustBeginTransaction)
  {
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      sendCommand(*driverSocketPtr, Begin);
      transactionId = receiveLongOrException();
      mustBeginTransaction = false;
      if (isDebugEnabled())
        logDebug(fctName, L"Begin transaction done. Transaction id=" + toWString(transactionId));
    FO_CATCH_NTIMES
  }
}

void Connection::setReadOnly(bool readOnlyPrm) throw (SocketIOException,
DriverException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::setReadOnly";
  checkIfConnected();
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    sendCommand(*driverSocketPtr, SetReadOnly);
    *driverSocketPtr << readOnlyPrm;
    receiveBoolOrException();
    readOnly = readOnlyPrm;
  FO_CATCH_NTIMES
}

void Connection::commit() throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::commit";
  
  LockScope ls(&connectionCS);
  checkIfConnected();

  if (autoCommit)
    throw DriverException(std::wstring(fctName)
			  + L": trying to commit a connection in autocommit mode");

  // Check if we are not committing an empty transaction (not started yet)
  if (mustBeginTransaction)
    return;

  doCommit();
}

void Connection::doCommit()  throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  try
  {
    sendCommand(*driverSocketPtr, Commit);
    int64_t acknowledgedTransactionId = receiveLongOrException();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (ProtocolException(
          L"Protocol error during commit (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = "
          + toUserString(transactionId) + L")"));
    }
  }
  catch (SocketIOException e)
  {
    // Connection failed, try to reconnect and re-exec the commit
    reconnectOrDie();
    // Try to get the commit result from the newly connected controller: if
    // this fails again, give up
    int64_t acknowledgedTransactionId = retrieveCommitResult();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (DriverException(
          L"Error during commit failover (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = " + toUserString(transactionId) + L")"));
    }

    // The controller will automatically redo the commit if it was not done
    // earlier so we can safely return here, this is a success.
  }

  // Success (with or without reconnection)

  // Commit must be followed by a BEGIN
  mustBeginTransaction = true;
  writeExecutedInTransaction = false;

}

void Connection::rollback() throw (SocketIOException, DriverException,
    ProtocolException, ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::rollback";
  
  LockScope ls(&connectionCS);

  checkIfConnected();

  if (autoCommit)
    throw DriverException(std::wstring(fctName)
			  + L": trying to rollback a connection in autocommit mode");

  // Check if we are not rollbacking an empty transaction (not started yet)
  if (mustBeginTransaction)
    return;

  try
  {
    sendCommand(*driverSocketPtr, Rollback);
    int64_t acknowledgedTransactionId = receiveLongOrException();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (ProtocolException(
          L"Protocol error during rollback (acknowledge transaction ID = "
            + toUserString(acknowledgedTransactionId)
            + L", expected transaction ID = "
            + toUserString(transactionId) + L")"));
    }
    // Rollback is followed by a BEGIN
    mustBeginTransaction = true;
  }
  catch (SocketIOException e)
  {
    // Connection failed, try to reconnect and re-exec the rollback
    reconnectOrDie();
    // Try to get the rollback result from the newly connected controller: if
    // this fails again, give up
    int64_t acknowledgedTransactionId = retrieveRollbackResult();
    if (acknowledgedTransactionId != transactionId)
    {
      throw (DriverException(
          L"Error during rollback failover (acknowledge transaction ID = "
          + toUserString(acknowledgedTransactionId)
          + L", expected transaction ID = " + toUserString(transactionId) + L")"));
    }
    // The controller will automatically redo the rollback if it was not
    // done earlier so we can safely return here, this is a success.
    return;
  }
}

Statement* Connection::createStatement() throw (SocketIOException,
    UnexpectedException)
{
  checkIfConnected();
  Statement* stPtr = new Statement(this);
  created_statements.push_back(stPtr);
  return stPtr;
}

ParameterStatement* Connection::createParameterStatement(const std::wstring &query) 
    throw (SocketIOException, UnexpectedException)
{
  checkIfConnected();
  ParameterStatement* pstPtr = new ParameterStatement(this, query);
  created_statements.push_back(pstPtr);
  return pstPtr;
}


bool Connection::isClosed() const throw (UnexpectedException)
{
  return (driverSocketPtr == NULL) || !driverSocketPtr->isConnected();
}

void Connection::checkIfConnected() const throw (SocketIOException,
    UnexpectedException)
{
  if (isClosed())
    throw SocketIOException(
        L"Request cannot be processed : connection is closed!");
}

DriverResultSet* Connection::preparedStatementGetMetaData(const wstring &sqlTemplate)
    throw (SocketIOException, BackendException, ControllerException, DriverException, 
    ProtocolException, NotImplementedException, ConnectionException,
    UnexpectedException)
{
  checkIfConnected();
  const wchar_t fctName[] = L"Connection::PreparedStatementGetMetaData";

  LockScope ls(&connectionCS);

  FO_TRY_NTIMES(RECONNECT_RETRIES)
    sendCommand(*driverSocketPtr, PreparedStatementGetMetaData);
    *driverSocketPtr << sqlTemplate;
  
    if (isDebugEnabled())
      logDebug(fctName, L"Statement sent successfully");
  
    return receiveResultSet();
  FO_CATCH_NTIMES

  return NULL; //avoid compiler warning
}

void Connection::setConnectionParametersOnRequest(Request &request)
{
  request.setIsAutoCommit(autoCommit);
}

DriverResultSet* Connection::statementExecuteQuery(RequestWithResultSetParameters& request)
    throw (SocketIOException, BackendException, ControllerException,
    DriverException, ProtocolException, NotImplementedException,
    ConnectionException, UnexpectedException)
{

  const wchar_t fctName[] = L"Connection::statementExecuteQuery";

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  DriverResultSet* retVal = NULL;
  SQLWarning* sqlwPtr = NULL;
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Sending read request " + static_cast<wstring>(request));
    sendCommand(*driverSocketPtr, StatementExecuteQuery);
    request.sendToStream(*driverSocketPtr);
    // clean up (in case of failover)
    if (retVal != NULL)
      delete retVal;
    if (sqlwPtr != NULL)
      delete sqlwPtr;
    SQLWarning* sqlwPtr = receiveSQLWarnings();
    retVal = receiveResultSet();
    if (retVal != NULL)
      retVal->setStatementWarnings(sqlwPtr);
  FO_CATCH_NTIMES
  return retVal;
}

ResultAndWarnings Connection::statementExecuteUpdate(Request& request)
    throw (SocketIOException, BackendException, ControllerException,
    DriverException, ProtocolException, ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::statementExecuteUpdate";

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  int controllerResponse = -1;

  // this for-loop is in the case we were disconnected and the controller did
  // not get the update request
  for (int failuresCnt=0; failuresCnt<RECONNECT_RETRIES; failuresCnt++)
  {
    ResultAndWarnings ret;
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      if (isDebugEnabled())
        logDebug(fctName, L"Sending write request " + static_cast<wstring>(request));
      sendCommand(*driverSocketPtr, StatementExecuteUpdate);
      writeRequestOnStream(request);
      request.setId(receiveLongOrException());
    FO_CATCH_NTIMES
  
    // At this point, the request id has been received, let's retrieve the
    // update count
    try
    {
      sqlwPtr = receiveSQLWarnings();    
      controllerResponse = static_cast<int>(receiveIntOrException());
      break; // get out of the big for-loop
    }
    catch (SocketIOException sioe)
    {
      reconnectOrDie();
      FO_TRY_NTIMES(RECONNECT_RETRIES)
        ret = retrieveExecuteUpdateResult(request);
      FO_CATCH_NTIMES
      if (ret.getUpdateCount() != -1)
      {
        return ret;
      }
      // otherwise, the update was not done on the controller, we go back at
      // the beginning of the function, cleaning up results
      delete ret.getWarnings();
    }
  }

  return ResultAndWarnings(controllerResponse, sqlwPtr);
}

ResultAndWarnings Connection::statementExecute(
    RequestWithResultSetParameters& request) throw (SocketIOException,
    BackendException, ControllerException, DriverException, ProtocolException,
    ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::statementExecute";

  LockScope ls(&connectionCS);
  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  std::list<ResultSetOrUpdateCount> results;

  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Sending request " + static_cast<wstring>(request));
    if (!autoCommit)
      writeExecutedInTransaction = true;
    sendCommand(*driverSocketPtr, StatementExecute);
    request.sendToStream(*driverSocketPtr);
    request.setId(receiveLongOrException());
  FO_CATCH_NTIMES

  try
  {
    sqlwPtr = receiveSQLWarnings();
    results = fetchMultipleResultsFromStream();
  }
  catch (SocketIOException sioe)
  {
    reconnectOrDie();
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      // Clean-up old results if any
      if (results.size() > 0)
      {
        for (std::list<ResultSetOrUpdateCount>::iterator iter = results.begin();
            iter != results.end(); iter++)
        {
          if (iter->isResultSet)
          {
            delete iter->resultSetPtr;
          }
        }
        results.clear();
      }
      if (sqlwPtr != NULL)
      {
        delete sqlwPtr;
        sqlwPtr = NULL;
      }
      // Then retrieve results on controller
      return (retrieveExecuteResult(request));
    FO_CATCH_NTIMES
  }

  return ResultAndWarnings(results, sqlwPtr);
}

ResultAndWarnings Connection::statementExecuteUpdateWithKeys(
    Request &request) throw (SocketIOException, BackendException,
    ControllerException, DriverException, ProtocolException, ConnectionException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::statementExecuteUpdateWithKeys";

  LockScope ls(&connectionCS);

  checkIfConnected();
  beginTransactionIfNeeded();
  setConnectionParametersOnRequest(request);

  SQLWarning* sqlwPtr = NULL;
  std::list<ResultSetOrUpdateCount> result;

  // this for-loop is in the case we were disconnected and the controller did
  // not get the update request
  for (int failuresCnt=0; failuresCnt<RECONNECT_RETRIES; failuresCnt++)
  {
    ResultAndWarnings ret;
    FO_TRY_NTIMES(RECONNECT_RETRIES)
      if (isDebugEnabled())
        logDebug(fctName, L"Executing write request with keys: "
            + static_cast<wstring>(request));
      sendCommand(*driverSocketPtr, StatementExecuteUpdateWithKeys);
      request.sendToStream(*driverSocketPtr);
      request.setId(receiveLongOrException());
    FO_CATCH_NTIMES
    try
    {
      sqlwPtr = receiveSQLWarnings();
      // Receive first the update count or an exception
      ResultSetOrUpdateCount uc;
      uc.isResultSet = false;
      uc.updateCount = static_cast<int>(receiveIntOrException());
      result.push_back(uc);
      // Then the ResultSet containing the autogenerated keys
      ResultSetOrUpdateCount rs;
      rs.isResultSet = true;
      rs.resultSetPtr = receiveResultSet();
      result.push_back(rs);
      break; //no exception, get out of the big for-loop
    }
    catch (SocketIOException sioe)
    {
      reconnectOrDie();
      FO_TRY_NTIMES(RECONNECT_RETRIES)
        // Clean-up old result if any
        if (result.size() > 0)
        {
          for (std::list<ResultSetOrUpdateCount>::iterator iter = result.begin();
              iter != result.end(); iter++)
          {
            if (iter->isResultSet)
            {
              delete iter->resultSetPtr;
            }
          }
          result.clear();
        }
        if (sqlwPtr != NULL)
        {
          delete sqlwPtr;
          sqlwPtr = NULL;
        }
        // Then retrieve result on controller
        ret = retrieveExecuteUpdateWithKeysResult(request);
      FO_CATCH_NTIMES
      if (ret.getResultList().size() > 1)
        return ret;
      // otherwise, the update was not done on the controller, we go back at
      // the beginning of the function, cleaning up results
      ret.getResultList().clear();
      delete ret.getWarnings();
    }
  }
  return ResultAndWarnings(result, sqlwPtr);
}

void Connection::writeRequestOnStream(const Request& request)
    throw (SocketIOException, UnexpectedException )
{
  if (!autoCommit)
    writeExecutedInTransaction = true;

  request.sendToStream(*driverSocketPtr);

}
void Connection::tryFetchNext(const wstring& cursorName, int fetchSize)
    throw (SocketIOException, ControllerException, ProtocolException,
    BackendException, UnexpectedException)
{   
  const wchar_t fctName[] = L"Connection::tryFetchNext";

  LockScope ls(&connectionCS);

  checkIfConnected();
  sendCommand(*driverSocketPtr, FetchNextResultSetRows);
  *driverSocketPtr<<cursorName;
  *driverSocketPtr<<fetchSize;
  if (isDebugEnabled())
    logDebug(fctName, L"Fetching next " + toUserString(fetchSize) + L" rows.");
  TypeTag tag(*driverSocketPtr);

  if (tag == TT_EXCEPTION)
    receiveException();

  if (tag != TT_NOT_EXCEPTION)
    throw ProtocolException(L"Protocol exception while trying fetchNext");

  // success, now we can let the DriverResultSet caller receive its data.
}


void Connection::closeRemoteResultSet(wstring cursorName)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::closeRemoteResultSet";
  
  LockScope ls(&connectionCS);

  checkIfConnected();
  sendCommand(*driverSocketPtr, CloseRemoteResultSet);
  *driverSocketPtr<<cursorName;
  if (isDebugEnabled())
    logDebug(fctName, L"Closing remote ResultSet");

  receiveBoolOrException();
}

bool Connection::receiveBoolOrException() throw (SocketIOException,
                                                 BackendException,
                                                 ControllerException,
                                                 ProtocolException,
                                                 UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    bool answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return false;
}

int32_t Connection::receiveIntOrException() throw (SocketIOException,
                                                   BackendException,
                                                   ControllerException,
                                                   ProtocolException,
                                                   UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    int32_t answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return 0;
}

int64_t Connection::receiveLongOrException() throw (SocketIOException,
                                                    BackendException,
                                                    ControllerException,
                                                    ProtocolException,
                                                    UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    int64_t answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if it's an exception, deserialize and throw it (at the same time)
  receiveException();
  //just to avoid compiler warnings
  return 0;
}

wstring Connection::receiveStringOrException() throw (SocketIOException,
                                                      BackendException,
                                                      ControllerException,
                                                      ProtocolException,
                                                      UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);
  //if it's not an exception, receive the bool normally
  if (tag == TT_NOT_EXCEPTION)
  {
    wstring answer;
    *driverSocketPtr>>answer;
    return answer;
  }
  //if we are here, it's an exception, deserialize and throw it (at the same
  //time)
  receiveException();
  //just to avoid compiler warnings
  return L"";
}

SQLWarning* Connection::receiveSQLWarnings() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  if (!receiveBoolOrException())
    // no warning
    return NULL;
  TypeTag exceptionType(*driverSocketPtr);
  if (exceptionType != TT_BACKEND_EXCEPTION)
    throw ProtocolException(
        L"Unknown exception received instead of SQLWarning");
  // Receive the warning as a BackendDriverException
  BackendException* e = new BackendException(*driverSocketPtr);
  return convertToSQLWarnings(e);
}

SQLWarning* Connection::getWarnings() throw (DriverException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::getWarnings";

  throwExceptionIfClosed(L"Cannot getWarnings() on a closed connection");
  if (!persistent_connection || !retrieve_sql_warnings)
    return first_warning;
  // else...
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Getting persistent connection warnings");
    sendCommand(*driverSocketPtr, ConnectionGetWarnings);
    *driverSocketPtr << persistent_connection_id;
    return receiveSQLWarnings();
  FO_CATCH_NTIMES
  return NULL; //avoid compiler warning
}

void Connection::clearWarnings() throw (DriverException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::clearWarnings";

  if (!persistent_connection || !retrieve_sql_warnings)
    return;
  // else...
  FO_TRY_NTIMES(RECONNECT_RETRIES)
    if (isDebugEnabled())
      logDebug(fctName, L"Getting persistent connection warnings");
    sendCommand(*driverSocketPtr, ConnectionClearWarnings);
    *driverSocketPtr << persistent_connection_id;
    receiveBoolOrException();
  FO_CATCH_NTIMES
}

SQLWarning* Connection::convertToSQLWarnings(BackendException* bde)
{
  if (bde == NULL)
    return NULL;
  // recursion here:
  SQLWarning* sqlw = new SQLWarning(bde->description(), bde->getSQLState(),
      convertToSQLWarnings((BackendException*) bde->getNext()),
      bde->getErrorCode());
  return sqlw;
}

void Connection::throwExceptionIfClosed(const wstring& msg)
    throw (DriverException, UnexpectedException)
{
  if (driverSocketPtr == NULL)
    throw DriverException(msg);
}

void Connection::receiveException() throw (SocketIOException,
                                           BackendException,
                                           ControllerException,
                                           ProtocolException,
                                           UnexpectedException)
{
  TypeTag exceptionType(*driverSocketPtr);
  if (exceptionType == TT_BACKEND_EXCEPTION)
    throw BackendException(*driverSocketPtr);
  else if (exceptionType == TT_CORE_EXCEPTION)
    throw ControllerException(*driverSocketPtr);
  else throw ProtocolException(L"Received unknown exception type");
}

DriverResultSet* Connection::receiveResultSet()
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, NotImplementedException, UnexpectedException)
{
  TypeTag tag(*driverSocketPtr);

  if (tag == TT_NULL_RESULTSET)
    return NULL;

  if (tag == TT_RESULTSET)
  {
    DriverResultSet* drs = new DriverResultSet(this);
    return drs;
  }

  if (tag == TT_EXCEPTION)
    receiveException();

  throw ProtocolException(L"Expected a resultset, received unexpected tag: "
      + toUserString(static_cast<int>(tag)));
  //just to avoid compiler warnings
  return NULL;
}

int64_t Connection::retrieveCommitResult() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveCommitResult);
  return receiveLongOrException();
}

int64_t Connection::retrieveRollbackResult() throw (SocketIOException,
    BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveRollbackResult);
  return receiveLongOrException();
}

ResultAndWarnings Connection::retrieveExecuteUpdateResult(const Request &request)
    throw (SocketIOException, BackendException, ControllerException,
    ProtocolException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::retrieveExecuteUpdateResult";
  if (isDebugEnabled())
    logDebug(fctName, L"Retrieving previous result of query "
        + toWString(request.getId()));
  sendCommand(*driverSocketPtr, RetrieveExecuteUpdateResult);
  *driverSocketPtr << request.getId();
  SQLWarning* sqlwPtr = receiveSQLWarnings();
  int result = static_cast<int>(receiveIntOrException());
  return ResultAndWarnings(result, sqlwPtr);
}

ResultAndWarnings Connection::retrieveExecuteResult(
    const Request &request) throw (SocketIOException, BackendException,
    ControllerException, ProtocolException, UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveExecuteResult);
  *driverSocketPtr << request.getId();
  SQLWarning* sqlwPtr = receiveSQLWarnings();
  std::list<ResultSetOrUpdateCount> result = fetchMultipleResultsFromStream();
  return ResultAndWarnings(result, sqlwPtr);
}

ResultAndWarnings Connection::retrieveExecuteUpdateWithKeysResult(
                        const Request &request) throw (SocketIOException,
                        BackendException, ControllerException,
                        ProtocolException, UnexpectedException)
{
  sendCommand(*driverSocketPtr, RetrieveExecuteUpdateWithKeysResult);
  *driverSocketPtr << request.getId();

  SQLWarning* sqlwPtr = receiveSQLWarnings();
  std::list<ResultSetOrUpdateCount> result;
  // Receive first the update count or an exception
  ResultSetOrUpdateCount uc;
  uc.isResultSet = false;
  uc.updateCount = static_cast<int>(receiveIntOrException());
  if (uc.updateCount != -1)
  {
    result.push_back(uc);
    // Then the ResultSet containing the autogenerated keys
    ResultSetOrUpdateCount rs;
    rs.isResultSet = true;
    rs.resultSetPtr = receiveResultSet();
    result.push_back(rs);
  }
  //else return an empty list
  return ResultAndWarnings(result, sqlwPtr);
}

std::list<ResultSetOrUpdateCount> Connection::fetchMultipleResultsFromStream()
                        throw (SocketIOException, BackendException,
                        ControllerException, ProtocolException,
                        UnexpectedException)
{
  std::list<ResultSetOrUpdateCount> results;
  bool hasResult = false;
  ResultSetOrUpdateCount resultReceived;
  do
  {
    hasResult = receiveBoolOrException();
    if (hasResult)
    {
      resultReceived.isResultSet = true;
      resultReceived.resultSetPtr = receiveResultSet();
//JavaReflecting FIXME: remove this
      results.push_back(resultReceived);
    }
    else
    {
      resultReceived.isResultSet = false;
      resultReceived.updateCount = receiveIntOrException();
//JavaReflecting FIXME: remove this
      if (resultReceived.updateCount != -1)
        results.push_back(resultReceived);
    }
//JavaReflecting FIXME: add this
//    results.push_back(resultReceived);
    
  }
  while (hasResult || resultReceived.updateCount != -1);
  return results;
}

void Connection::reconnectOrDie() throw (ConnectionException,
    DriverException, UnexpectedException)
{
  bool reconnected = false;
  while (!reconnected)
  {
    try
    {
      tryToReconnect();
      reconnected = true;
    }
    catch (NoMoreControllerException nmce)
    {
      throw (ConnectionException(nmce.description()));
    }
// See CAROB-108
//    catch (DriverException de)
//    {
//      logError(fctName, L"Unexpected DriverException: " +
//	       de.description() + L", wrapping ");
//      throw (ConnectionException(de.description()));
//    }
    // other exceptions are forwarded to caller...
  }
}

void Connection::tryToReconnect() throw (NoMoreControllerException,
    DriverException, UnexpectedException)
{
  const wchar_t fctName[] = L"Connection::tryToReconnect";
  //Wait until other methods/Commands are done
  LockScope ls(&connectionCS);

  if (isWarnEnabled())
    logWarn(fctName, L"Connection to controller lost - trying to reconnect");

  // close everything
  if (isDebugEnabled())
    logDebug(fctName, L"Closing socket of lost connection");

  // inform controller pool of the failure: even if the info came from the
  // controller pool (ie. controller pool closed our socket), it is safe to tell
  // him again about the failure
  // This must be done *before* closeSocket because forceControllerDown() will
  // switch the socket to cancelled state
  if (isDebugEnabled())
    logDebug(fctName, L"Controller " + static_cast<wstring>(connected_controller)
        + L" is not responding, suspecting it of failure");
  controller_pool.forceControllerDown(connected_controller);

  // It is now safe to close network socket
  closeSocket();
  
  // ...And try next one in the list

  // CAROB-125 : we may need to extract the code below in a new method
  // so corruptedConnectionClose() can use it.
  try
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Getting a new controller from the list...");
    getConnectionToNewController();
  }
  catch (AuthenticationException ae)
  {
    // Should not happen, this probably mean an inconsistency in controller
    // configuration but safely ignore (see below)
    wstring msg = L"Warning! Authentication exception received on connection retry, controller configuration might be inconsistent.";
    if (isWarnEnabled())
      logWarn(fctName, msg);
    throw DriverException(msg+ L" Exception:" + ae.description());
  }

  // Connection succeeded, restore the previous connection state on the
  // controller
  try
  {
    if (isInfoEnabled())
      logInfo(fctName, L"Restoring connection state on the new controller "
          + static_cast<wstring>(connected_controller));
    sendCommand(*driverSocketPtr, RestoreConnectionState);
    *driverSocketPtr<<writeExecutedInTransaction;
    *driverSocketPtr<<autoCommit;
    if (!autoCommit)
      *driverSocketPtr<<transactionId;
    *driverSocketPtr<<persistent_connection;
    if (persistent_connection)
      *driverSocketPtr<<persistent_connection_id;
    *driverSocketPtr<<retrieve_sql_warnings;
    // read ack
    bool ack = false;
    *driverSocketPtr>>ack;
    // Restore read-only state (not part of RestoreConnectionState command for
    // backward compatibility)
    setReadOnly(readOnly);
    if (isInfoEnabled())
      logInfo(fctName, L"Transparent reconnection succeeded");
  }
  catch (SocketIOException sioe)
  {
    throw (ConnectionException(L"Error in reconnection attempt: "
        + sioe.description()));
  }
}
