/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "ConnectionParameters.hpp"
#include "ControllerPool.hpp"

#include "Common.hpp"

using std::wstring;

using namespace CarobNS;

/**
 * Constants for connection defaults
 */
const std::wstring    ConnectionParameters::DEFAULT_USER                   = L"user"; 
const std::wstring    ConnectionParameters::DEFAULT_PASSWD                 = L"" ;
const std::wstring    ConnectionParameters::DEFAULT_DB                     = L"myDB" ;
const ConnectPolicy   ConnectionParameters::DEFAULT_POLICY                 = ROUND_ROBIN; 
const int             ConnectionParameters::DEFAULT_PING_DELAY             = 1000;
const int             ConnectionParameters::DEFAULT_CONTROLLER_TIMEOUT     = 25000;
const bool            ConnectionParameters::DEFAULT_CONNECTION_PERSISTENCY = false ;
const bool            ConnectionParameters::DEFAULT_RETRIEVE_SQL_WARNINGS  = false ;

ConnectionParameters::ConnectionParameters(
    const std::vector<ControllerInfo>& ctrls,
    const std::wstring& db /*= DEFAULT_DB*/,
    const std::wstring& uname /*= DEFAULT_USER*/,
    const std::wstring& upass /*= DEFAULT_PASSWD*/,
    ConnectPolicy cp /*= DEFAULT_POLICY*/,
    int pingDelayInMs /* = DEFAULT_PING_DELAY */,
    int controllerTimeoutInMs /* = DEFAULT_CONTROLLER_TIMEOUT */,
    bool persistentConnections /*= DEFAULT_CONNECTION_PERSISTENCY */,
    bool retrieveSQLWarnings /*= DEFAULT_RETRIEVE_SQL_WARNINGS */)

    throw (ConnectionException, UnexpectedException) :

    database_name(checkDatabaseName(db)),
    user_name(uname),
    user_pass(upass),
    connect_policy(cp),
    controller_pool_ptr(NULL),
    ping_delay_in_ms(pingDelayInMs),
    controller_timeout_in_ms(controllerTimeoutInMs),
    persistent_connections(persistentConnections),
    retrieve_sql_warnings(retrieveSQLWarnings)
{
  controller_list = ctrls;
  // Put default values if given delay/timeout are unrelevant
  if (ping_delay_in_ms <= 0)
    ping_delay_in_ms = DEFAULT_PING_DELAY;
  // -3210 = disable pingers, see stillborn_hack in ControllerWatcher.cpp
  if (controller_timeout_in_ms <= 0 && controller_timeout_in_ms != -3210)
    controller_timeout_in_ms = DEFAULT_CONTROLLER_TIMEOUT;
}

ConnectionParameters::ConnectionParameters(const ConnectionParameters& from) :
    controller_list(from.controller_list),
    database_name(from.database_name),
    user_name(from.user_name),
    user_pass(from.user_pass),
    connect_policy(from.connect_policy),
    controller_pool_ptr(from.controller_pool_ptr),
    ping_delay_in_ms(from.ping_delay_in_ms),
    controller_timeout_in_ms(from.controller_timeout_in_ms),
    persistent_connections(from.persistent_connections),
    retrieve_sql_warnings(from.retrieve_sql_warnings)
{
  if (controller_pool_ptr != NULL)
    // we are now pointing to the controller pool, so we have to add a reference
    // to it via the controller pool manager
    controller_pool_ptr = &ControllerPoolManager::getPool(controller_list,
        connect_policy, ping_delay_in_ms, controller_timeout_in_ms);
}

ConnectionParameters::~ConnectionParameters()
{
  if (controller_pool_ptr != NULL)
    ControllerPoolManager::freePool(controller_pool_ptr);
}

AbstractControllerPool& ConnectionParameters::getControllerPool()
{
  if (controller_pool_ptr == NULL)
    controller_pool_ptr = &ControllerPoolManager::getPool(controller_list,
        connect_policy, ping_delay_in_ms, controller_timeout_in_ms);
  return *controller_pool_ptr;
}

const wstring& ConnectionParameters::checkDatabaseName(const wstring& dbname) const
    throw (ConnectionException, UnexpectedException)
{
  wchar_t pbChar;
  if (!isValidDatabaseName(dbname, pbChar))
      throw ConnectionException(L"Invalid database name "+dbname+L" at character '"+pbChar+L"'.");
  return dbname;
}

bool ConnectionParameters::isValidDatabaseName(const wstring& databaseName, wchar_t& pbChar) const
{
  int size = databaseName.length();
  wchar_t c;

  for (int i = 0; i < size; i++)
  {
    c = databaseName[i];
    if ((c < L'-') || (c > L'z') || (c == L'/') || (c == L'.') || (c == L'`')
        || ((c > L'9') && (c < L'A')) || ((c > L'Z') && (c < '_')))
    {
      pbChar = c;
      return false;
    }
  }
  return true;
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
