/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#include "ControllerInfo.hpp"
#include "SocketAddress.hpp"
#include "DriverSocket.hpp"
#include "ControllerPool.hpp"
#include "StringCodecs.hpp" //for toString()
#include "Common.hpp" //for logging
#include <sstream> //for wostringstream

#include <vector>

using namespace CarobNS;

using std::wstring;

// defined below
namespace CarobNS
{
std::wostream& operator<< (std::wostream& ostr, const addrinfo& toprint);
}

ControllerInfo::ControllerInfo(const std::wstring &name, in_port_t port)
    throw (ConnectionException, UnexpectedException) :
    original_hostname(name)
{
  // get the addresses
  addrinfo* addrs = getAddrs(name, port);
  // addrs always contain at least one address, otherwise an exception would
  // have been thrown
  // We cannot know for sure which of the returned addresses is valid, we take
  // arbitrarly the first one
  socket_address.setAddress(addrs->ai_addr, addrs->ai_addrlen);
  // discard the allocated addresses
  freeaddrinfo(addrs);
}

ControllerInfo::ControllerInfo(const ControllerInfo& ci)
    throw (ConnectionException, UnexpectedException)
{
  // make use of the = operator
  *this = ci;
}

ControllerInfo::~ControllerInfo()
{
}

ControllerInfo& ControllerInfo::operator = (const ControllerInfo& ci)
{
  socket_address = ci.socket_address;
  original_hostname = ci.original_hostname;
  return *this;
}

/* static */
addrinfo* ControllerInfo::getAddrs(const std::wstring& name, in_port_t port)
    throw (ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"ControllerInfo::getAddrs";
  char service[NI_MAXSERV];
  addrinfo*  addressList = NULL;
  addrinfo   hint;
  memset(&hint, 0, sizeof(hint));
  hint.ai_socktype = SOCK_STREAM;
  hint.ai_family = AF_UNSPEC;
  hint.ai_flags = AI_ADDRCONFIG; // will prevent receiving addresses we can't talk to
  #ifdef AI_NUMERICSERV
  // prevent name resolution (only if AI_NUMERICSERV is supported) 
  hint.ai_flags |= AI_NUMERICSERV; // service is a port number
  #endif
  // convert given port to char*
  snprintf(service, NI_MAXSERV, "%d", port);
  int error = getaddrinfo(toString(name).c_str(), service, &hint, &addressList);
  if (isInfoEnabled())
  {
    std::wostringstream msg;
    msg<<L"Gathered following addresses for "<<name<<":"<<port<<" : ";
    for (addrinfo *addri = addressList; addri != NULL; addri = addri->ai_next)
    {
      msg << *addri;
      if (addri->ai_next)
        msg << L"; ";
    }
    logInfo(fctName, msg.str());
  }
  // not sure addressList can be NULL without error, but it is cheap to check
  if (error || addressList == NULL)
  {
    wstring wMsg(L"Error while executing getaddrinfo() on hostname \"" + name + L"\"");
#ifndef __MINGW32__
    // get the error from getaddrinfo
    const char* gaiMsg = gai_strerror(error);
    wMsg += L". Error was: " + fromString(gaiMsg);
    if (addressList != NULL)
    {
      freeaddrinfo(addressList);
    }
#endif
    if (isDebugEnabled())
      logDebug(fctName, wMsg);
    throw ConnectionException(wMsg);
  }
  return addressList;
}

DriverSocket* ControllerInfo::connect(AbstractControllerPool& pool)
    throw (ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"ControllerInfo::connect()";

    // register the socket to our policy so it can be killed during connection
  DriverSocket *socketPtr = pool.newRegisteredSocket(*this);

  // set the correct protocol family
  int family = PF_INET;
  if (socket_address.getFamily() == AF_INET6)
  {
    family = PF_INET6;
    if (isWarnEnabled())
      logWarn(fctName, L"IP v6 host detected. This feature is untested");
  }
  try
  {
    socketPtr->create(family);
  }
  catch (ConnectionException ce)
  {
    pool.deleteRegisteredSocket(socketPtr);
    throw;
  }

  if (isDebugEnabled())
    logDebug(fctName, L"Connecting...");

  try
  {
    if (socketPtr->connectTo(socket_address))
    {
      if (isInfoEnabled())
        logInfo(fctName, L"Connection to " + static_cast<wstring>(*this)
            + L" succeeded");
      return socketPtr;
    }
  }
  catch (ConnectionException ce)
  {
    pool.deleteRegisteredSocket(socketPtr);
    throw ce;
  }
  if (isDebugEnabled())
    logDebug(fctName, L"Could not connect to " + static_cast<wstring>(*this));

  // could not connect
  pool.deleteRegisteredSocket(socketPtr);
  throw ConnectionException(L"Connection failed");
}

bool ControllerInfo::operator < (const ControllerInfo& ci) const
{
  return (socket_address < ci.socket_address);
}

bool ControllerInfo::operator == (const ControllerInfo& ci) const
{
  // if none if less that the other, they are equal!
  return (!(*this < ci)
       && !(ci < *this));
}

ControllerInfo::operator std::wstring() const
{
  std::wostringstream msg;
  msg<< *this;
  return msg.str();
}

std::wostream& CarobNS::operator<< (std::wostream& ostr, const ControllerInfo& ci)
{
  return ostr << ci.original_hostname<<L" ("<<*(ci.socket_address.getAddressPtr())<<L")";
}

std::wostream& CarobNS::operator<< (std::wostream& ostr, const std::vector<ControllerInfo>& clist)
{
  ostr << L"[";
  for (std::vector<ControllerInfo>::const_iterator iter = clist.begin();
      iter != clist.end(); iter++)
  {
    ostr << *iter;
    if (iter+1 != clist.end())
      ostr << L", ";
  }
  ostr << L"]";
  return ostr;
}

/*
 Typically used like this:

  for (addrinfo *addri = addressList; addri != NULL; addri = addri->ai_next)
  {
    std::wcerr << *addri;
    if (addri->ai_next)
      std::wcerr << L"; ";
  }
*/
namespace CarobNS
{
std::wostream&
operator<< (std::wostream& oss, const addrinfo& addri)
{

    switch(addri.ai_family) {
    case AF_INET:
        oss << L"AF_INET"; break;
    case AF_INET6:
        oss << L"AF_INET6"; break;
    case AF_UNIX:
        oss << L"AF_UNIX"; break;
    case AF_UNSPEC:
        oss << L"AF_UNSPEC"; break;
    default:
        oss << L"unknown ai_family number=" << addri.ai_family ; break;
    }
    oss << L":";

    if (addri.ai_canonname)
        oss << StaticCodecs::fromASCII(addri.ai_canonname) << L"/";
    else
        oss << L"<unresolved>" << L"/";

    oss << *(addri.ai_addr);

    oss << L",";
    switch(addri.ai_socktype) {
    case SOCK_STREAM:
        oss << L"SOCK_STREAM"; break;
    case SOCK_DGRAM:
        oss << L"SOCK_DGRAM"; break;
    default:
        oss << L"unknown socket type=" << addri.ai_socktype ; break;
    }

    oss << L"/";
    switch(addri.ai_protocol) {
    case IPPROTO_TCP:
        oss << L"IPPROTO_TCP"; break;
    case IPPROTO_UDP:
        oss << L"IPPROTO_UDP"; break;
    default:
        oss << L"unknown protocol=" << addri.ai_protocol ; break;
    }


    oss << L",flags=";

    oss << addri.ai_flags;

    // hex
    std::wostringstream wss;
    wss.setf(std::ios_base::hex, std::ios_base::basefield);
    wss.setf(std::ios_base::showbase);
    wss << L"(" << addri.ai_flags << L")";

    oss << wss.str();

    return oss;
}
}
