/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#include "ControllerPingSender.hpp"

#include "SocketAddress.hpp"
#include "WatchedControllers.hpp"
#include "Common.hpp"

#include <unistd.h> // usleep()
#include <sys/socket.h>
#include <netinet/in.h> // in_port_t
#include <pthread.h>
#include <vector>

using namespace CarobNS;

using std::wstring;

// extern C thread function that launches the ping
void *PingSenderThread(void * thisPtr)
{
  ControllerPingSender* pt = (ControllerPingSender*)thisPtr;
  pt->run();
  pthread_exit(NULL);
  return NULL;
}

ControllerPingSender::ControllerPingSender(WatchedControllers& controllerList,
    int socket, int pingDelayInMs) : controllers(controllerList),
    send_socket(socket), ping_delay_in_ms(pingDelayInMs),
    // max shift is 1.5 times ping delay
    max_sleep_shift(static_cast<int>(ping_delay_in_ms*1.5)), is_stopped(false)
{
}

void ControllerPingSender::run()
{
  const wchar_t fctName[] = L"ControllerPingSender::run";
  if (isDebugEnabled())
    logDebug(fctName, L"Pinger started");
  java_byte packet[1];
  packet[0] = CONTROLLER_PING_VERSION;
  while (!is_stopped)
  {
    std::vector<SocketAddress> ctrls = controllers.getControllerAddresses();
    for (std::vector<SocketAddress>::const_iterator iter = ctrls.begin(); iter != ctrls.end(); iter++)
    {
      sendto(send_socket, static_cast<void*>(packet), sizeof(packet), 0,
          iter->getAddressPtr(), iter->getAddressLength());
      // This is really too much logs...
    }
    // detect violent clock changes (user hardly set new date/time)
    int64_t before = currentTimeMillis();
    usleep((ping_delay_in_ms)*1000);
    int64_t after = currentTimeMillis();
    int diffMs = static_cast<int>(after-before);
    // usleep calls last at about 10ms. Don't report error on very short delays  
    if (diffMs > max_sleep_shift && ping_delay_in_ms>10)
      if (isErrorEnabled())
      {
        logError(fctName, 
            wstring(L"Time shift detected. Manually changing system date/time can lead to erroneous controller failure detections!")
            + L"\nSleep lasted " + toUserString(diffMs)
            + L"ms instead of "+ toUserString(ping_delay_in_ms));
      }
  }
}
