/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "ControllerPool.hpp"
#include "ControllerWatcher.hpp"
#include "ControllerStateChangedCallback.hpp"
#include "DriverSocket.hpp"
#include "Connection.hpp"

#include "ConnectionParameters.hpp"
#include "Common.hpp"

#include <map>
#include <sstream>

using std::wstring;
using std::vector;

using namespace CarobNS;

////////////////////////////////////////////////////////////////////////////////
//                           ABSTRACT CONTROLLER POOL
////////////////////////////////////////////////////////////////////////////////

AbstractControllerPool::AbstractControllerPool(
    const vector<ControllerInfo>& controllerList, int pingDelayInMs,
    int controllerTimeoutInMs) throw (DriverException, UnexpectedException) :
    ref_counter(0)
{
  if (controllerList.size() == 0)
    throw DriverException(
        L"Controller list must at least contain one controller!");
  if (pingDelayInMs == 0)
    throw DriverException(
        L"Controller ping delay cannot be 0!");
  alive_controllers.reserve(controllerList.size());
  for (vector<ControllerInfo>::const_iterator iter = controllerList.begin();
      iter != controllerList.end(); iter++)
  {
    alive_controllers.push_back(ControllerAndVdbState(*iter));
  }

  // Create callback for the controller watcher. Ownership is transfered to the
  // WatchedControllers instance of the controller watcher
  callback_ptr = new SocketKillerCallback(*this);
  try
  {
    // Since the destruction of the watcher can be long (waiting for select
    // timeout), the property of this watcher instance + the callback instance
    // is transfered to the ControllerWatcherThread C function (unless a 
    // connection exception is thrown)
    watcher_ptr = new ControllerWatcher(controllerList, callback_ptr,
        pingDelayInMs, controllerTimeoutInMs);
    pthread_create(&watcher_thread, NULL, ControllerWatcherThread, watcher_ptr);
    pthread_detach(watcher_thread);
  }
  catch (ConnectionException cause)
  {
    alive_controllers.clear();
    delete watcher_ptr;
    throw DriverException(
        L"Controller watcher creation failed!");
  }
}

AbstractControllerPool::~AbstractControllerPool()
{
  watcher_ptr->stop();
  alive_controllers.clear();
}

void AbstractControllerPool::setVdbDownOnController(const ControllerInfo& controller)
{
  LockScope ls(&pool_CS);
  // remove all occurences of the given controller
  for (vector<ControllerAndVdbState>::iterator iter = alive_controllers.begin();
      iter != alive_controllers.end();)
  {
    if (iter->controller == controller)
    {
      iter->lastVdbFailure = currentTimeSeconds();
      return;
    }
  }
}

bool AbstractControllerPool::isVdbUpOnController(const ControllerInfo& controller)
{
  LockScope ls(&pool_CS);
  // remove all occurences of the given controller
  for (vector<ControllerAndVdbState>::iterator iter = alive_controllers.begin();
      iter != alive_controllers.end();)
  {
    if (iter->controller == controller)
      return iter->isVdbUp(currentTimeSeconds());
  }
  return false;
}

int AbstractControllerPool::numberOfAliveControllers()
{
  LockScope ls(&pool_CS);
  return alive_controllers.size();
}

DriverSocket *AbstractControllerPool::newRegisteredSocket(const ControllerInfo& controller)
  throw (ConnectionException)
{
  LockScope ls(&pool_CS);

  DriverSocket *socket;
  try {
    socket = new DriverSocket(controller);
  } catch (CodecException ce) {
    std::wostringstream woss;
    woss << ce;
    logError(L"AbstractControllerPool::newRegisteredSocket", woss.str() );
    throw ConnectionException(ce.description(), ce.getSQLState(), &ce);
  }

  callback_ptr->registerSocket(controller, socket);
  return socket;
}

void AbstractControllerPool::deleteRegisteredSocket(DriverSocket *socket)
{
  LockScope ls(&pool_CS);
  callback_ptr->unRegisterSocket(socket->controller, socket);
  delete socket;
}

void AbstractControllerPool::forceControllerDown(const ControllerInfo& controller)
{
  LockScope ls(&pool_CS);
  watcher_ptr->forceControllerDown(controller);
}

void AbstractControllerPool::controllerDown(const ControllerInfo& controller)
{
  LockScope ls(&pool_CS);
  // remove all occurences of the given controller
  for (vector<ControllerAndVdbState>::iterator iter = alive_controllers.begin();
      iter != alive_controllers.end();)
  {
    if (iter->controller == controller)
      iter = alive_controllers.erase(iter);
    else
      iter++;
  }
}

void AbstractControllerPool::controllerUp(const ControllerInfo& controller)
{
  LockScope ls(&pool_CS);
  alive_controllers.push_back(ControllerAndVdbState(controller));
}

AbstractControllerPool::operator wstring()
{
  LockScope ls(&pool_CS);
  wstring sRet(L"controller list: ");
  for (size_t i = 0; i < alive_controllers.size(); i++)
  {
    sRet += L"[" + static_cast<wstring>(alive_controllers[i]) + L"]";
  }
  sRet += L" - Ref counter=" + toUserString(ref_counter);
  return sRet;
}

int AbstractControllerPool::removeRef()
{
  ref_counter--;
  if (ref_counter < 1)
    last_time_zero_ref = currentTimeMillis();
  return ref_counter;
}

////////////////////////////////////////////////////////////////////////////////
//                             ROUND ROBIN POLICY
////////////////////////////////////////////////////////////////////////////////

RoundRobinControllerPool::RoundRobinControllerPool(
    const vector<ControllerInfo>& controllerList, int pingDelayInMs,
    int controllerTimeoutInMs) throw (DriverException, UnexpectedException) :
AbstractControllerPool(controllerList, pingDelayInMs, controllerTimeoutInMs),
index(-1)
{
}

ControllerInfo RoundRobinControllerPool::getController()
    throw (NoMoreControllerException, UnexpectedException)
{
  const wchar_t fctName[] = L"RoundRobinControllerPool::getController";
  LockScope ls(&pool_CS);
  int nbOfControllers = alive_controllers.size();
  if (nbOfControllers == 0)
    throw NoMoreControllerException(L"All controllers down");
  index++;
  if (index >= nbOfControllers)
    index = 0;
  // to stop iteration when all controllers have been tested
  int nbTested = 0;
  // if the vdb was not available for the whished controller, we get the next
  // one in a round robin manner
  ControllerAndVdbState ctrl = alive_controllers[index];
  while (nbTested < nbOfControllers)
  {
    if (ctrl.isVdbUp(currentTimeSeconds()))
    {
      if (isDebugEnabled())
        logDebug(fctName, L"Selected controller[" + toUserString(index) + L"]:"
            + static_cast<wstring>(alive_controllers[index]));
      return ctrl.controller;
    }
    // try next controller
    index++;
    if (index >= nbOfControllers)
      index = 0;
    nbTested++;
    ctrl = alive_controllers[index];
  }
  // all controllers tested, no more available
  throw NoMoreControllerException(L"All controllers down");
}

////////////////////////////////////////////////////////////////////////////////
//                           CONTROLLER POOL MANAGER
////////////////////////////////////////////////////////////////////////////////

/* Pool index less than operator */
bool ltPoolIndex::operator()(const PoolIndex& pi1, const PoolIndex& pi2) const
{
  if (pi1.policy < pi2.policy)
    return true;
  if (pi1.policy > pi2.policy)
    return false;
  if (pi1.ping_delay_in_ms < pi2.ping_delay_in_ms)
    return true;
  if (pi1.ping_delay_in_ms > pi2.ping_delay_in_ms)
    return false;
  if (pi1.controller_timeout_in_ms < pi2.controller_timeout_in_ms)
    return true;
  if (pi1.controller_timeout_in_ms > pi2.controller_timeout_in_ms)
    return false;
  // same policy, compare number of controllers
  if (pi1.controllers.size() < pi2.controllers.size())
    return true;
  else if (pi1.controllers.size() > pi2.controllers.size())
    return false;
  // same size, compare controllers themselves
  vector<ControllerInfo>::const_iterator iter_pi1 = pi1.controllers.begin();
  vector<ControllerInfo>::const_iterator iter_pi2 = pi2.controllers.begin();
  while (iter_pi1 != pi1.controllers.end())
  // we can safely check only pi1, vectors have same size
  {
    if (*iter_pi1 < *iter_pi2)
      return true;
    if (*iter_pi2 < *iter_pi1)
      return false;
    iter_pi1++;
    iter_pi2++;
  } 
  // identical
  return false;
}

/* static members declaration */
CriticalSection ControllerPoolManager::pool_map_CS;
std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>  ControllerPoolManager::pool_map;

/*static*/
AbstractControllerPool& ControllerPoolManager::getPool(vector<ControllerInfo>& ctrls,
    ConnectPolicy cp, int pingDelayInMs, int controllerTimeoutInMs)
    throw (DriverException, UnexpectedException)
{
  const wchar_t fctName[] = L"ControllerPoolManager::getPool()";

  // iterate through the list to find the wanted pool, if present
  PoolIndex idx;
  idx.controllers = ctrls;
  idx.policy = cp;
  idx.ping_delay_in_ms = pingDelayInMs;
  idx.controller_timeout_in_ms = controllerTimeoutInMs;
  LockScope ls(&pool_map_CS);
  std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator pool
      = pool_map.find(idx);
  if (pool != pool_map.end())
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Reusing pool \""
          + static_cast<wstring>(*(pool->second)) + L"\"");
    pool->second->addRef();
    tryTocleanUpPools();
    return *(pool->second);
  }
  // if the pool was not found, let's create a new one
  AbstractControllerPool* new_pool;
  switch (cp)
  {
    case ROUND_ROBIN:
      new_pool = new RoundRobinControllerPool(ctrls, pingDelayInMs, controllerTimeoutInMs);
    break;
    default:
      tryTocleanUpPools();
      throw DriverException(L"Unsupported connection policy #" + toUserString(static_cast<int>(cp)));
  }
  new_pool->addRef();
  pool_map[idx] = static_cast<AbstractControllerPool*>(new_pool);
  if (isDebugEnabled())
    logDebug(fctName, L"Creating new pool \"" + static_cast<wstring>(*new_pool) + L"\"");
  tryTocleanUpPools();
  return *new_pool;
}

/*static*/
void ControllerPoolManager::tryTocleanUpPools()
{
  const wchar_t fctName[] = L"ControllerPoolManager::tryTocleanUpPools";
  LockScope ls(&pool_map_CS);
  for (std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator iter
      = pool_map.begin(); iter != pool_map.end();)
  {
    std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator toRemove = pool_map.end();

    // has this controller pool been freed since a long time?
    if (iter->second->getRefCount() == 0
	&& (currentTimeMillis() - POOL_TIMEOUT_IN_MS) > iter->second->getLastTimeZeroRef())
    {
        toRemove = iter;
        if (isInfoEnabled())
          logInfo(fctName, L"Pool \"" + static_cast<wstring>(*iter->second)
              + L"\" freed " + toWString(static_cast<int64_t>(currentTimeMillis() - iter->second->getLastTimeZeroRef()))
              + L"ms ago - Deleting it");
    }

    // must increment BEFORE removal
    iter++;

    if (toRemove != pool_map.end())
    {
      delete toRemove->second;
      pool_map.erase(toRemove);
    }
  }
}

/*static*/
void ControllerPoolManager::freePool(AbstractControllerPool* pool)
{
  const wchar_t fctName[] = L"ControllerPoolManager::freePool";
  if (pool == NULL)
  {
    if (isErrorEnabled())
      logError(fctName, L"Null pool can't be freed!");
    return;
  }
  LockScope ls(&pool_map_CS);
  std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator iter = pool_map.begin();
  while (iter != pool_map.end())
  {
    if (iter->second == pool)
    {
      if (iter->second->removeRef() == 0)
      {
        if (isDebugEnabled())
          logDebug(fctName, L"No more references to pool \""
              + static_cast<wstring>(*iter->second) + L"\" - keeping it at least "
              + toWString(POOL_TIMEOUT_IN_MS) + L" ms");
      }
      break; // pool found, stop search
    }
    iter++;
  }
  // No match => log error and continue
  if (iter == pool_map.end() && isErrorEnabled())
  {
    wstring msg(L"Pool " + static_cast<wstring>(*pool)
        + L" not found in pool map! Map contains ");
    for (std::map<PoolIndex, AbstractControllerPool*, ltPoolIndex>::iterator iter
        = pool_map.begin(); iter != pool_map.end(); iter++)
    {
      msg += L"[" + static_cast<wstring>(*(iter->second)) + L"]";
    }
    logError(fctName, msg);
  }
  // look for pools to delete
  tryTocleanUpPools();
}
