/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */


#include "ControllerStateChangedCallback.hpp"

#include "ControllerInfo.hpp"
#include "JavaSocket.hpp"
#include "ControllerPool.hpp"
#include "Common.hpp"

#include <sys/socket.h>
#include <unistd.h>
#include <string>

using namespace CarobNS;
using std::map;
using std::vector;
using std::wstring;

/* static members declaration */
CriticalSection SocketKillerCallback::controllers_and_sockets_CS;

SocketKillerCallback::~SocketKillerCallback()
{
  for (map<ControllerInfo, vector<JavaSocket*> >::iterator iter = controllers_and_sockets.begin();
      iter != controllers_and_sockets.end(); iter++)
  {
    iter->second.clear();
  }
  controllers_and_sockets.clear();  
}

void SocketKillerCallback::registerSocket(const ControllerInfo& ctrl, JavaSocket* s)
{
  LockScope ls(&controllers_and_sockets_CS);
  vector<JavaSocket*> sockets;
  map<ControllerInfo, vector<JavaSocket*> >::iterator iter = controllers_and_sockets.find(ctrl);
  if (iter != controllers_and_sockets.end())
  {
    sockets = iter->second;
  }
  sockets.push_back(s);
  controllers_and_sockets[ctrl] = sockets;
}

void SocketKillerCallback::unRegisterSocket(const ControllerInfo& ctrl,
    JavaSocket* s)
{
  LockScope ls(&controllers_and_sockets_CS);
  map<ControllerInfo, vector<JavaSocket*> >::iterator ctrlIter = 
      controllers_and_sockets.find(ctrl);
  if (ctrlIter != controllers_and_sockets.end())
  {
    vector<JavaSocket*> sockets = ctrlIter->second;
    vector<JavaSocket*>::iterator socketIter = sockets.begin();
    while (socketIter != sockets.end())
    {
      if (*socketIter == s)
        socketIter = sockets.erase(socketIter);
      else
        socketIter++;
    }
    controllers_and_sockets[ctrl] = sockets;
  }
}

void SocketKillerCallback::onControllerDown(const ControllerInfo& ctrl)
{
  const wchar_t fctName[] = L"SocketKillerCallback::OnControllerDown";
  // tell policy asap
  policy.controllerDown(ctrl);
  LockScope ls(&controllers_and_sockets_CS);
  map<ControllerInfo, vector<JavaSocket*> >::iterator iter = controllers_and_sockets.find(ctrl);
  if (iter != controllers_and_sockets.end())
  {
    if (isDebugEnabled())
      logDebug(fctName, L"Shutting down sockets connected to controller "
          + static_cast<wstring>(ctrl));
    for (vector<JavaSocket*>::iterator socketIter = iter->second.begin();
        socketIter != iter->second.end(); socketIter++)
    {
      (*socketIter)->shutdown();
    }
    // Clean-up shut down sockets
    iter->second.clear();
  }
}

void SocketKillerCallback::onControllerUp(const ControllerInfo& ctrl)
{
  // tell policy asap
  policy.controllerUp(ctrl);
}
