/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#include "ControllerWatcher.hpp"

#include "SocketAddress.hpp"
#include "ControllerPingSender.hpp"
#include "ControllerStateChangedCallback.hpp"
#include "WatchedControllers.hpp"
#include "JavaSocket.hpp"
#include "ControllerInfo.hpp"
#include "Common.hpp"

#include <fcntl.h> // setsockopt
#include <sstream> //for wostringstream

#include <unistd.h> // usleep()

using namespace CarobNS;
using std::vector;
using std::wstring;

namespace {
  const int stillborn_hack(-3210);
}

// extern C thread function that launches the ping
void* ControllerWatcherThread(void * thisPtr)
{
  ControllerWatcher* self = (ControllerWatcher*)thisPtr;
  self->run();
  
  // Function has finished, it means that we were stopped, so we have to clean
  // everything:

  delete self;
  pthread_exit(0);
  return NULL;
}

ControllerWatcher::ControllerWatcher(const vector<ControllerInfo>& controllerList,
      ControllerStateChangedCallback* callbackPtr, int pingDelayInMs, int controllerTimeout)
      throw (ConnectionException, UnexpectedException) :
      ping_delay_in_ms(pingDelayInMs), controller_timeout(controllerTimeout),
      is_stopped(false)
      
{
  const wchar_t fctName[] = L"ControllerWatcher::ControllerWatcher";

  controllers_ptr = new WatchedControllers(controllerList, currentTimeMillis(),
      controllerTimeout, callbackPtr);

  // Get the ip family of the first controller, this will be the default one
  socket_fd = socket(controllerList.front().getSocketAddress().getFamily(), SOCK_DGRAM, 0);
  // set socket non blocking
#ifdef __MINGW32__
  int var = 1;
  ioctlsocket(socket_fd, FIONBIO, (u_long FAR*) &var);
  if (var == 0)
  {
    if (isErrorEnabled())
    {
      logError(fctName, L"Could not turn ping socket to non-blocking mode.");
    }
    throw ConnectionException(L"Could not turn ping socket to non-blocking mode.");
  }
#else
    // *nix:
    int flags;
    flags = fcntl(socket_fd, F_GETFL, 0);
    //Set non blocking
    if (fcntl(socket_fd, F_SETFL, flags | O_NONBLOCK) < 0)
    //if above fcntl doesn't work, try this old way of doing it
    //flags = 1;
    //if (ioctl(socketFd, FIOBIO, &flags) < 0)
    {
      if (isErrorEnabled())
        logError(fctName, L"Could not turn ping socket to non-blocking mode.");
    throw ConnectionException(L"Could not turn ping socket to non-blocking mode.");
    }
#endif

  pinger_ptr = new ControllerPingSender(*controllers_ptr, socket_fd, ping_delay_in_ms);

  if (controller_timeout == stillborn_hack)
  {
    pinger_ptr->stop();
    std::wostringstream ostr;
    ostr << L"ControllerPinger of Watcher " << this << L", to: "<< controllerList
	 << L" stillborn by " << stillborn_hack;
    logWarn(fctName, ostr.str());
  }

  pthread_create(&pinger_thread, NULL, PingSenderThread, pinger_ptr);
}

ControllerWatcher::~ControllerWatcher()
{
  // we should be called only when run() is over:
  // objects in use should not be deleted!

  // stop the pinger thread
  getControllerPingSenderPtr()->stop();
  // wait for the pinger to finish
  pthread_join(getPingerThread(), NULL);
  // pinger has finished, we can close socket...
  close(getSocketFd());
  // ...and delete everything
  delete(getControllerPingSenderPtr());
  delete(getControllersPtr());

}

void ControllerWatcher::processAnswers()
{
  const wchar_t fctName[] = L"ControllerWatcher::processAnswers";
  java_byte buf[1];
  sockaddr_storage respondingCtrlAddr;
  socklen_t respondingCtrlAddrLength = sizeof(sockaddr_storage);
  ssize_t recLength = recvfrom(socket_fd, static_cast<void*>(buf), 1,
       0, reinterpret_cast<sockaddr*>(&respondingCtrlAddr), &respondingCtrlAddrLength);
  while (recLength > 0)
  {
    try
    {
      SocketAddress ctrlAddr(reinterpret_cast<sockaddr*>(&respondingCtrlAddr), respondingCtrlAddrLength);
      // This is really too much logs...
//      if (isDebugEnabled())
//        logDebug(fctName, L"Received response from " + static_cast<wstring>(ctrl));
      if (buf[0] == CONTROLLER_PING_VERSION)
      {
        controllers_ptr->setControllerResponsed(ctrlAddr, currentTimeMillis());
      }
      else if (isDebugEnabled())
      {
        std::wostringstream buffer;
        buffer<<L"Wrong ping version in response from "//<<FIXME: ctrlAddr
              <<+ L" was: "<<buf[0]
              <<L", expected:"<<CONTROLLER_PING_VERSION;
        logDebug(fctName, buffer.str());
      }
      recLength = recvfrom(socket_fd, static_cast<void*>(buf), 1,
           0, reinterpret_cast<sockaddr*>(&respondingCtrlAddr), &respondingCtrlAddrLength);
    }
    catch (ConnectionException ignored)
    {
    }
  }
}

void ControllerWatcher::run()
{
  const wchar_t fctName[] = L"ControllerWatcher::run";
  if (isDebugEnabled())
  {
    logDebug(fctName, L"Watcher started");
  }

  // secret code to disable the ping
  if (controller_timeout == stillborn_hack)
  {
    std::wostringstream ostr; ostr << L"ControllerWatcher " << this << L" disabled by " << stillborn_hack;
    logWarn(fctName, ostr.str());

    while (!is_stopped)
      sleep(2);
  }

  while (!is_stopped)
  {
    // return of poll/select function
    int retVal = JavaSocket::pollOnSingleFd(socket_fd, ping_delay_in_ms);
    // if (retVal == 0) => timeout, no response, do nothing
    if (retVal > 0)
    {
      // process all responses received until now
      processAnswers();
    }
    // This will start callbacks according to new states
    controllers_ptr->lookForDeadControllers(currentTimeMillis());
  }
  pthread_exit(0);
}

/**
 * Forces a given controller to be considered as not responding
 * @see WatchedControllers#setControllerDown(ControllerInfo)
 */
void ControllerWatcher::forceControllerDown(const ControllerInfo& c)
{
  controllers_ptr->setControllerDown(c.getSocketAddress());
}
