/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "DriverResultSet.hpp"

#include "BigDecimal.hpp"
#include "Connection.hpp"
#include "TypeTag.hpp"
#include "Statement.hpp"
#include "Field.hpp"

#include "CarobException.hpp"
#include "Common.hpp"

#include <sstream> //for wostringstream
#include <algorithm> //for transform

using std::wstring;
using std::vector;
using std::transform;

using namespace CarobNS;

DriverResultSet::~DriverResultSet()
{
  clearAllData();
  for (int i=0; i<nbOfColumns; i++)
  {
    delete fields[i];
    fields[i] = NULL;
  }
  nbOfColumns = -1;
  //Inform statement we are beeing destroyed
  if (owningStatementPtr != NULL)
  {
    owningStatementPtr->removeResultSetFromList(this);
    owningStatementPtr = NULL;
  }
}

void DriverResultSet::clearAllData()
{
  //empty all data
  dataIsNull = true;
  if (data.size() > 0)
  {
    size_t cnt = 0;
    //First delete data contained in data[x][y].as_other.
    //These are void* so we need to switch on the column type to delete it
    //Let's make this switch only once per column !
    for (size_t col=0; col<data[0]->size(); col++)
    {
      switch (columnTypeTags[col])
      {
        case TT_STRING:
          for (cnt=0; cnt<data.size(); cnt++)
          {
            if (!(*nulls[cnt])[col])
            {
              delete static_cast<wstring*>((*data[cnt])[col].as_other);
            }
          }
        break;
        case TT_BIGDECIMAL:
          for (cnt=0; cnt<data.size(); cnt++)
          {
            if (!(*nulls[cnt])[col])
            {
              delete static_cast<BigDecimal*>((*data[cnt])[col].as_other);
            }
          }
        break;
        case TT_SQL_TIMESTAMP:
          for (cnt=0; cnt<data.size(); cnt++)
          {
            if (!(*nulls[cnt])[col])
            {
              delete static_cast<SQLTimeStamp*>((*data[cnt])[col].as_other);
            }
          }
        break;
        case TT_BYTE_ARRAY:
        case TT_BLOB:
        case TT_CLOB:
          for (cnt=0; cnt<data.size(); cnt++)
          {
            if (!(*nulls[cnt])[col])
            {
              LargeData* ld = static_cast<LargeData*>((*data[cnt])[col].as_other);
              delete[] ld->data;
              delete ld;
            }
          }
        break;
      }
    }
    //Now remove data from the vector
    for (cnt=0; cnt<data.size(); cnt++)
    {
      data[cnt]->clear();
      delete data[cnt];
    }
    data.clear();
    //The types
    columnTypeTags.clear();
    //And finally the nulls
    for (cnt=0; cnt<nulls.size(); cnt++)
    {
      nulls[cnt]->clear();
      delete nulls[cnt];
    }
    nulls.clear();
    delete warningsPtr;
    warningsPtr = NULL;
    // Don't delete statementWarnings, it will be done by the statement itself
  }  
}

DriverResultSet::DriverResultSet(Connection* conPtr) throw (ProtocolException,
    SocketIOException, NotImplementedException, UnexpectedException) :
currentRow(-1),
nbOfRows(-1),
nbOfColumns(-1),
dataIsNull(true),
fetchSize(0),
wasNullFlag(false),
resultSetType(0),
resultSetConcurrency(0),
warningsPtr(NULL),
statement_warnings_ptr(NULL),
owningStatementPtr(NULL),
isClosed(true)
{
  const wchar_t fctName[] = L"DriverResultSet::DriverResultSet";
  connectionPtr = conPtr;
  const DriverSocket&  socket = conPtr->getDriverSocket();
  bool hasWarnings = false;
  socket>>hasWarnings;
  if (hasWarnings)
  {
    BackendException* e = new BackendException(socket);
    warningsPtr = Connection::convertToSQLWarnings(e);
  }
  // Get the ResultSet metadata, deserialize columns information
  socket>>nbOfColumns;
  for (int i=0; i<nbOfColumns; i++)
  {
    fields.push_back(new Field(socket));
  }
  columnTypeTags.reserve(nbOfColumns);
  deserializers.reserve(nbOfColumns);
  TypeTag colTypes(socket);
  if (colTypes != TT_COL_TYPES)
  {
    throw ProtocolException(std::wstring(fctName) + L"Column types were expected");
  }

  socket>>nbOfRows;

  // Receive java types
  if (nbOfRows > 0)
  {
    for (int col = 0; col < nbOfColumns; col++)
    {
      TypeTag tag(socket);
      deserializers.push_back(SQLDataSerialization::getDeserializer(tag));
      columnTypeTags.push_back(tag);
    }
  }
  receiveRows(socket);
  if (hasMoreData)
    socket>>cursorName;
  isClosed = false;
}

DriverResultSet::operator wstring()
{
  wstring sRet;
  sRet+=toUserString(nbOfRows)+L" rows - ";
  sRet+=toUserString(nbOfColumns)+L" columns - ";
  sRet+=L"current row:" +toUserString(currentRow)+L" - ";
  sRet+=hasMoreData?L"[has more data/":L"[no more data/";
  sRet+=isClosed?L"closed]":L"not closed]";
  return sRet;
}

void DriverResultSet::setStatement(Statement* stmtPtr)
{
  owningStatementPtr = stmtPtr;
  if (stmtPtr != NULL)
  {
    fetchSize = stmtPtr->getFetchSize();
    resultSetConcurrency = stmtPtr->getResultSetConcurrency();
    resultSetType = stmtPtr->getResultSetType();
  }
}

bool DriverResultSet::next() throw (SocketIOException, ControllerException, 
    ProtocolException, BackendException, UnexpectedException)
{
  checkIfClosed();

  if (nbOfRows == 0)
    return false;

  if (currentRow + 1 >= nbOfRows)
  {
    if (hasMoreData)
    {
      // TODO: what happens if user closed this connection? What says JDBC?
      connectionPtr->tryFetchNext(cursorName, fetchSize);
      // no exception from controller, so let's receive our new rows
      receiveRows(connectionPtr->getDriverSocket());
      currentRow = 0;
      if (dataIsNull)
      {
        nbOfRows = 0;
        return false;
      }
      else
      {
        nbOfRows = data.size();
        return true;
      }
    }

    // force scroll past end
    currentRow = nbOfRows;
    return false;
  }
  //TODO:
  //clearWarnings();
  currentRow++;
  return true;
}

wstring DriverResultSet::getString(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);
  if (wasNullFlag)
  {
    throw (NullValueException(L"getString: value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  if (columnTypeTags[columnIndex - 1] != TT_STRING)
  {
    throw (DriverException(L"getString: value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type string"));
  }
  return *(static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
}

wstring DriverResultSet::getAsString(int columnIndex, const std::locale& loc/*=std::locale()*/)
    throw (DriverException, NullValueException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);
  if (wasNullFlag)
  {
    throw (NullValueException(L"getString: value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  std::wostringstream buffer;
  buffer.imbue(loc);

  // big switch on column data type
  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
      buffer << *(static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      buffer << bd->toString(loc);
      break;
    }
    case TT_BOOLEAN:
      buffer << (*data[currentRow])[columnIndex - 1].as_bool;
    break;
    case TT_INTEGER:
      buffer << (*data[currentRow])[columnIndex - 1].as_int;
    break;
    case TT_LONG:
      buffer << (*data[currentRow])[columnIndex - 1].as_long;
    break;
    case TT_FLOAT:
      buffer << (*data[currentRow])[columnIndex - 1].as_float;
    break;
    case TT_DOUBLE:
      buffer << (*data[currentRow])[columnIndex - 1].as_double;
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to string conversion not implemented yet.");
    break;
    case TT_SQL_DATE:
      buffer << (*data[currentRow])[columnIndex - 1].as_long;
    break;
    case TT_SQL_TIME:
      buffer << (*data[currentRow])[columnIndex - 1].as_int;
    break;
    case TT_SQL_TIMESTAMP:
      throw NotImplementedException(L"SQL TimeStamp to string conversion not implemented yet.");
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to string conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to string conversion not implemented yet.");
    break;
    default:
      throw NotImplementedException(L"to string conversion not implemented for this type yet.");
  }
  return buffer.str();
}

int32_t DriverResultSet::getInt32(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getInt: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_INTEGER)
  {
    throw (DriverException(L"getInt: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type integer"));
  }
  
  return (*data[currentRow])[columnIndex - 1].as_int;
}

int DriverResultSet::getAsInt(int columnIndex) throw (DriverException,
    NullValueException, ConversionException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getInt: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  int ret = -1;
  // big switch on column data type
  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
    { //these brackets to be able to declare loc var without warns
      wstring valAsString = trim(*static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
      //Tries to parse the string as an integer
      if(!wstringTo<int>(ret, valAsString))
      {
        //Last chance: is it a boolean ?
        transform(valAsString.begin(),valAsString.end(), valAsString.begin(), (int(*)(int))std::tolower); //see http://lists.debian.org/debian-gcc/2002/04/msg00092.html
        if (L"true" == valAsString)
          ret = 1;
        else if (L"false" == valAsString)
          ret = 0;
        else //we cannot do more... throw an exception
        {
          throw (ConversionException(L"The value (" + valAsString
              + L") cannot be converted to int", CONV_INVALID));
        }
      }
    }
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      ret = static_cast<int>(*bd);
      break;
    }
    case TT_BOOLEAN:
      if ((*data[currentRow])[columnIndex - 1].as_bool)
        ret = 1;
      else
        ret = 0;
    break;
    case TT_INTEGER:
      ret = (*data[currentRow])[columnIndex - 1].as_int;
    break;
    case TT_LONG:
      ret = static_cast<int>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_FLOAT:
      ret = static_cast<int>((*data[currentRow])[columnIndex - 1].as_float);
    break;
    case TT_DOUBLE:
      ret = static_cast<int>((*data[currentRow])[columnIndex - 1].as_double);
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to int conversion not implemented yet.");
    break;
    case TT_SQL_DATE:
      ret = static_cast<int>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_SQL_TIME:
      ret = (*data[currentRow])[columnIndex - 1].as_int;
    break;
    case TT_SQL_TIMESTAMP:
      ret = static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other)->time;
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to int conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to int conversion not implemented yet.");
    break;
    default:
      throw NotImplementedException(L"Int conversion not implemented for this type yet.");
  }
  return ret;
}

int64_t DriverResultSet::getInt64(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getInt64: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_LONG)
  {
    throw (DriverException(L"getInt64: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type integer"));
  }
  
  return (*data[currentRow])[columnIndex - 1].as_long;
}

int64_t DriverResultSet::getAsInt64(int columnIndex) throw (DriverException,
    NullValueException, ConversionException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getAsInt64: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  int64_t ret = -1;
  // big switch on column data type
  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
    { //these bracket to be able to declare loc var without warns
      wstring valAsString = trim(*static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
      //Tries to parse the string as an integer
      if(!wstringTo<int64_t>(ret, valAsString))
      {
        //Last chance: is it a boolean ?
        transform(valAsString.begin(),valAsString.end(), valAsString.begin(), (int(*)(int))std::tolower);
        if (L"true" == valAsString)
          ret = 1;
        else if (L"false" == valAsString)
          ret = 0;
        else //we cannot do more... throw an exception
        {
          throw (ConversionException(L"The value " + valAsString
              + L" is not a valid int64 number", CONV_INVALID));
        }
      }
    }
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      ret = static_cast<int64_t>(*bd);
      break;
    }
    case TT_BOOLEAN:
      if ((*data[currentRow])[columnIndex - 1].as_bool)
        ret = 1;
      else
        ret = 0;
    break;
    case TT_INTEGER:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_LONG:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_FLOAT:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_float);
    break;
    case TT_DOUBLE:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_double);
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to int64 conversion not implemented yet.");
    break;
    case TT_SQL_DATE:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_SQL_TIME:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_SQL_TIMESTAMP:
      ret = static_cast<int64_t>(static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other)->time);
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to int64 conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to int64 conversion not implemented yet.");
    break;
    default:
      throw NotImplementedException(L"Int64 conversion not implemented for this type yet.");
  }
  return ret;
}

uint64_t DriverResultSet::getAsUInt64(int columnIndex) throw (DriverException,
    NullValueException, ConversionException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getAsUInt64: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  uint64_t ret = 0;
  // big switch on column data type
  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
    { //these bracket to be able to declare loc var without warns
      wstring valAsString = trim(*static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
      //Tries to parse the string as an integer
      if(!wstringTo<uint64_t>(ret, valAsString))
      {
        //Last chance: is it a boolean ?
        transform(valAsString.begin(),valAsString.end(), valAsString.begin(), (int(*)(int))std::tolower);
        if (L"true" == valAsString)
          ret = 1;
        else if (L"false" == valAsString)
          ret = 0;
        else //we cannot do more... throw an exception
        {
          throw (ConversionException(L"The value " + valAsString
              + L" is not a valid uint64 number", CONV_INVALID));
        }
      }
    }
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      ret = static_cast<uint64_t>(*bd);
      break;
    }
    case TT_BOOLEAN:
      if ((*data[currentRow])[columnIndex - 1].as_bool)
        ret = 1;
      else
        ret = 0;
    break;
/* TODO: => check that the number is positive before casting it to uint
    case TT_INTEGER:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_LONG:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_FLOAT:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_float);
    break;
    case TT_DOUBLE:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_double);
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to int64 conversion not implemented yet.");
    break;
    case TT_SQL_DATE:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_SQL_TIME:
      ret = static_cast<int64_t>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_SQL_TIMESTAMP:
      ret = static_cast<int64_t>(static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other)->time);
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to int64 conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to int64 conversion not implemented yet.");
    break;
*/
    default:
      throw NotImplementedException(L"Int64 conversion not implemented for this type yet.");
  }
  return ret;
}

float DriverResultSet::getFloat(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getFloat: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_FLOAT)
  {
    throw (DriverException(L"getFloat: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type float"));
  }
  
  return (*data[currentRow])[columnIndex - 1].as_float;
}

float DriverResultSet::getAsFloat(int columnIndex) throw (DriverException,
    NullValueException, ConversionException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getAsFloat: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  float ret = -1;

  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
    {
      wstring valAsString = trim(*static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
      //Tries to parse the string as an integer
      if(!wstringTo<float>(ret, valAsString))
      {
        //Last chance: is it a boolean ?
        transform(valAsString.begin(),valAsString.end(), valAsString.begin(), (int(*)(int))std::tolower);
        if (L"true" == valAsString)
          ret = 1;
        else if (L"false" == valAsString)
          ret = 0;
        else //we cannot do more... throw an exception
        {
          throw (ConversionException(L"The value " + valAsString
              + L" is not a valid float number", CONV_INVALID));
        }
      }
    }
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      ret = static_cast<float>(*bd);
      break;
    }
    case TT_BOOLEAN:
      if ((*data[currentRow])[columnIndex - 1].as_bool)
        ret = 1;
      else
        ret = 0;
    break;
    case TT_INTEGER:
      ret = static_cast<float>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_LONG:
      ret = static_cast<float>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_FLOAT:
      ret = (*data[currentRow])[columnIndex - 1].as_float;
    break;
    case TT_DOUBLE:
      ret = static_cast<float>((*data[currentRow])[columnIndex - 1].as_double);
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to float conversion not implemented yet.");
    break;
    case TT_SQL_DATE:
      ret = static_cast<float>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_SQL_TIME:
      ret = static_cast<float>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_SQL_TIMESTAMP:
      ret = static_cast<float>(static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other)->time);
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to float conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to float conversion not implemented yet.");
    break;
    default:
      throw NotImplementedException(L"Float conversion not implemented for this type yet.");
  }
  return ret;
}

double DriverResultSet::getDouble(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getDouble: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_DOUBLE)
  {
    throw (DriverException(L"getDouble: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type double"));
  }
  
  return (*data[currentRow])[columnIndex - 1].as_double;
}

double DriverResultSet::getAsDouble(int columnIndex) throw (DriverException,
    NullValueException, ConversionException, NotImplementedException,
    UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getAsDouble: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }

  double ret = -1;

  switch (columnTypeTags[columnIndex - 1])
  {
    case TT_STRING:
    {
      wstring valAsString = trim(*static_cast<wstring*>((*data[currentRow])[columnIndex - 1].as_other));
      //Tries to parse the string as an integer
      if(!wstringTo<double>(ret, valAsString))
      {
        //Last chance: is it a boolean ?
        transform(valAsString.begin(),valAsString.end(), valAsString.begin(), (int(*)(int))std::tolower);
        if (L"true" == valAsString)
          ret = 1;
        else if (L"false" == valAsString)
          ret = 0;
        else //we cannot do more... throw an exception
        {
          throw (ConversionException(L"The value " + valAsString
              + L" is not a valid double number", CONV_INVALID));
        }
      }
    }
    break;
    case TT_BIGDECIMAL:
    {
      BigDecimal* bd = static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
      ret = static_cast<double>(*bd);
      break;
    }
    case TT_BOOLEAN:
      if ((*data[currentRow])[columnIndex - 1].as_bool)
        ret = 1;
      else
        ret = 0;
    break;
    case TT_INTEGER:
      ret = static_cast<double>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_LONG:
      ret = static_cast<double>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_FLOAT:
      ret = static_cast<double>((*data[currentRow])[columnIndex - 1].as_float);
    case TT_DOUBLE:
      ret = (*data[currentRow])[columnIndex - 1].as_double;
    break;
    case TT_BYTE_ARRAY:
      throw NotImplementedException(L"ByteArray to double conversion not implemented yet.");
    break;
    break;
    case TT_SQL_DATE:
      ret = static_cast<double>((*data[currentRow])[columnIndex - 1].as_long);
    break;
    case TT_SQL_TIME:
      ret = static_cast<double>((*data[currentRow])[columnIndex - 1].as_int);
    break;
    case TT_SQL_TIMESTAMP:
      ret = static_cast<double>(static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other)->time);
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob to double conversion not implemented yet.");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Java Serializable to double conversion not implemented yet.");
    break;
    default:
      throw NotImplementedException(L"Double conversion not implemented for this type yet.");
  }
  return ret;
}

SQLTimeStamp DriverResultSet::getTimeStamp(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getTimeStamp: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_SQL_TIMESTAMP)
  {
    throw (DriverException(L"getTimeStamp: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type timestamp"));
  }
  
  return *static_cast<SQLTimeStamp*>((*data[currentRow])[columnIndex - 1].as_other);
}

LargeData& DriverResultSet::getLargeObject(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getLargeData: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (!(columnTypeTags[columnIndex - 1] == TT_BLOB || columnTypeTags[columnIndex - 1] == TT_CLOB || 
      columnTypeTags[columnIndex - 1] == TT_BYTE_ARRAY))
  {
    throw (DriverException(L"getLargeData: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type blob, clob or byte array"));
  }
  
  return *static_cast<LargeData*>((*data[currentRow])[columnIndex - 1].as_other);
}

const BigDecimal& DriverResultSet::getBigDecimal(int columnIndex) throw (DriverException,
    NullValueException, NotImplementedException, UnexpectedException)
{
  checkRowAndColPosAndSetNullFlag(columnIndex);

  if (wasNullFlag)
  {
    throw (NullValueException(L"getBigDecimal: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is NULL"));
  }
  
  if (columnTypeTags[columnIndex - 1] != TT_BIGDECIMAL)
  {
    throw (DriverException(L"getBigDecimal: Value at row " + toUserString(currentRow)
        + L" column " + toUserString(columnIndex) + L" is not of type BigDecimal"));
  }
  
  return *static_cast<BigDecimal*>((*data[currentRow])[columnIndex - 1].as_other);
}

void DriverResultSet::close() throw (SocketIOException, BackendException,
    ControllerException, ProtocolException, UnexpectedException)
{
  if (hasMoreData)
    connectionPtr->closeRemoteResultSet(cursorName);
  isClosed = true;
}

void DriverResultSet::receiveRows(const DriverSocket& socket)
    throw (SocketIOException, ProtocolException,
    UnexpectedException)
{
  clearAllData();
  socket>>nbOfRows;

  //Declare empty result which is used in case of null values
  ResultSetDataType emptyResult = {0};
  //Reserve memory now so it won't be done during iterations
  nulls.reserve(nbOfRows);
  data.reserve(nbOfRows);

  // Receive the actual data
  for (int rowCnt = 0; rowCnt<nbOfRows; rowCnt++)
  {
    vector<ResultSetDataType>* rowOfData = new vector<ResultSetDataType>;
    vector<bool>* nullsForThisRow = new vector<bool>;
    rowOfData->reserve(nbOfColumns);
    nullsForThisRow->reserve(nbOfColumns);
    TypeTag tt(socket);
    if (tt != TT_ROW)
    {
      throw ProtocolException(L"A row was expected");
    }

    // First let's flag null values using a burst of booleans
    for (int colCnt = 0; colCnt < nbOfColumns; colCnt++)
    {
      bool b;
      socket>>b;
      nullsForThisRow->push_back(b);
    }
    nulls.push_back(nullsForThisRow);
    
    for (int colCnt2 = 0; colCnt2 < nbOfColumns; colCnt2++)
    {
      if (!(*nullsForThisRow)[colCnt2])
        rowOfData->push_back(deserializers[colCnt2](socket));
      else 
        rowOfData->push_back(emptyResult);
    }
    data.push_back(rowOfData);
    //now we have some data
    if (dataIsNull)
      dataIsNull = false;
  }
  socket>>hasMoreData;
}

void DriverResultSet::checkRowPos() const throw (DriverException,
    UnexpectedException)
{
  if (currentRow < 0)
    throw DriverException(L"Before start of result set");

  if (currentRow == nbOfRows)
    throw DriverException(L"After end of result set");
}

void DriverResultSet::checkIfClosed() const throw (DriverException, 
    UnexpectedException)
{
  if (isClosed)
    throw DriverException(L"Trying to access a closed ResultSet");
}

bool DriverResultSet::isNull(int columnIndex) const throw (DriverException,
    UnexpectedException)
{
  bool answer = false;

  checkIfClosed();

  checkRowPos();

  if (fields.size() == 0)
    throw DriverException(L"Query generated no fields for ResultSet");

  if (columnIndex < 1 || columnIndex > nbOfColumns)
    throw DriverException(L"Column Index out of range ( "
        + toUserString(columnIndex) + L" > " + toUserString(nbOfColumns) + L").");

  if (static_cast<int>(data[currentRow]->size())<columnIndex-1)
    answer = true;
  else
    answer = (*nulls[currentRow])[columnIndex - 1];
  return answer;
}

SQLWarning* DriverResultSet::getWarnings() throw (DriverException, UnexpectedException)
{
  checkIfClosed();
  return warningsPtr;
}
  
void DriverResultSet::checkRowAndColPosAndSetNullFlag(int columnIndex)
    throw (DriverException, UnexpectedException)
{
    wasNullFlag = isNull(columnIndex);
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
