/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): 
 */

#include "DriverSocket.hpp"

#include "CarobException.hpp"

#include <sstream>

using std::wstring;

using namespace CarobNS;

const DriverSocket& DriverSocket::operator <<(const wstring & s) const
    throw (SocketIOException, UnexpectedException)
{
  // First write "true" (for "non null string"), then the size of the upcoming
  // string, then the string itself
  *this<<true;
  size_t strLen = s.size();
  *this<<static_cast<int32_t>(strLen);
  try
  {
    for (size_t idx=0; idx < strLen; idx+=STRING_CHUNK_SIZE)
      writeJavaUTF(s.substr(idx, STRING_CHUNK_SIZE));
  }
  catch (CodecException ce)
  {
    // not so bad fix for CAROB-101
    if (isErrorEnabled())
    {
      std::wostringstream buffer;
      buffer<<L"CodecException while UTF8 encoding string to send:'"<< s << L"'. " << ce.description();;
      logError(L"DriverSocket::operator<<(wstring)", buffer.str());
    }
    throw UnexpectedException(ce.description(), ce.getSQLState(), &ce);
  }
  return *this;
}

const DriverSocket& DriverSocket::operator <<(const wchar_t *s) const
    throw (SocketIOException, UnexpectedException)
{
  wstring wobject(s);
  return *this<<wobject;
}

const DriverSocket & DriverSocket::operator >>(wstring & s) const
    throw (SocketIOException, UnexpectedException)
{
  bool ack = false;
  *this>>ack;
  if (ack)
  {
    //First reads the size of the upcoming string, then the string itself.
    int strSize;
    *this>>strSize;
    if (strSize > 0)
    {
      s.empty();
      s.reserve(strSize+1);
      size_t sizeRead = 0;
      //We must read strSize character, which can be done in several passes
      std::wstring chunk;
      while (static_cast<int>(sizeRead)<strSize)
      {
        try
        {
          sizeRead += readJavaUTF(chunk);
        }
        catch (CodecException ce)
        {
	  // not so bad fix for CAROB-101
          if (isErrorEnabled())
          {
	    std::wostringstream buffer;
	    // Here GNU libstdc++ 4.0.2 ostream.tcc will just widen the bytes to wchar_t
	    // Not sure this is portable
	    buffer<<L"CodecException while UTF8-decoding byte sequence received'"<< chunk << L"'. " << ce.description();
	    logError(L"DriverSocket::operator<<(wstring)", buffer.str());
	  }
	  throw UnexpectedException(ce.description(), ce.getSQLState(), &ce);
	}
        s += chunk;
      }
    }
  }
  return *this;
}

const DriverSocket & DriverSocket::operator <<(const int32_t &i) const
    throw (SocketIOException, UnexpectedException)
{
  writeJavaInt(i);
  return *this;
}

const DriverSocket & DriverSocket::operator >>(int32_t &i) const
    throw (SocketIOException, UnexpectedException)
{
  readJavaInt(i);
  return *this;
}

const DriverSocket & DriverSocket::operator <<(const int64_t &i) const
    throw (SocketIOException, UnexpectedException)
{
  writeJavaLong(i);
  return *this;
}

const DriverSocket & DriverSocket::operator >>(int64_t &i) const
    throw (SocketIOException, UnexpectedException)
{
  readJavaLong(i);
  return *this;
}

const DriverSocket & DriverSocket::operator <<(const bool &b) const
    throw (SocketIOException, UnexpectedException)
{
  writeJavaBool(b);
  return *this;
}

const DriverSocket & DriverSocket::operator >>(bool &b) const
    throw (SocketIOException, UnexpectedException)
{
  readJavaBool(b);
  return *this;
}

void DriverSocket::readBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  readJavaBytes(length, data);
}

void DriverSocket::writeBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  writeJavaBytes(length, data);
}
