/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon, Marc Herbert
 */

#include "JavaSocket.hpp"
#include "SocketAddress.hpp"

#include "Common.hpp"
#include "SystemDependantDefs.hpp"

#include <locale>
#include <string>
#include <stdexcept>

#ifdef __MINGW32__
  #include <ws2tcpip.h>
#else
  #include <sys/types.h>   //  socket(), send(), etc.
  #include <sys/socket.h>  //         "
  #include <netinet/tcp.h> // TCP_NODELAY etc.
#endif
#include <errno.h>
#include <fcntl.h>

#ifdef CAROB_USE_SELECT
#include <sys/select.h>
#define POLL_FUNCTION L"Select"
#else
#include <sys/poll.h>
#define POLL_FUNCTION L"Poll"
#endif

#include <unistd.h>  // close()

using std::wstring;
using std::string;

using namespace CarobNS;

namespace {
inline uint64_t ntohll(const uint64_t &n)
{
  static const bool bigEndian = (ntohl(1) == 1);

  if (bigEndian)
    return n;
  else
    return (((uint64_t)ntohl(n)) << 32) | ntohl(n >> 32);
}

inline uint64_t htonll(const uint64_t &n)
{
  static const bool bigEndian = (ntohl(1) == 1);

  if (bigEndian)
    return n;
  else
    return (((uint64_t)htonl(n)) << 32) | htonl(n >> 32);
}
}


// See carob/doc/locales

// An empty string "" will get the "default" C++ locale, typically set
// at run-time using environnement variables like LANG/LC_XXX. This
// will work fine as long as this "default" is set to some UTF8 locale.

// Using "C" should work fine as long as every string involved is pure
// ASCII.

#define NAME_OF_ANY_UTF8_LOCALE_AVAILABLE "en_US.utf8"

namespace CarobNS { // tryUTF8locale() exported only for testing purposes

  // TODO: add a *run-time* UTF8 locale definition option. See
  //       CAROB-74 and related
  // TODO: when available, should use POSIX'
  //        compare(uppercase(nl_langinfo(CODESET)), "UTF-?8")

  // For windows we may:
  // - try codecvt_byname(".65001")
  // - implement our own codecvt on top of
  //     MultiByteToWideChar(CP_UTF8,...) like this:
  //     http://groups.google.com/group/comp.os.ms-windows.programmer.win32/browse_thread/thread/257dc6357fbc6b48/6d448d9e36c89514
  // - use iconv...

  std::locale tryUTF8locale()
  {
    // __func__ is unfortunately C99
    const wchar_t funcname[] = __WFILE__ L":tryUTF8locale()";

    // 1. trying hardwired macro
    try {
      logInfo(funcname,
              L"1. trying to use locale: " NAME_OF_ANY_UTF8_LOCALE_AVAILABLE);
      return trylocale(NAME_OF_ANY_UTF8_LOCALE_AVAILABLE);
    } catch (CodecException& ce) {
      logWarn(funcname,
              L"please install the " NAME_OF_ANY_UTF8_LOCALE_AVAILABLE " locale");
    }

    // 2. try to fallback on the user defined locale
    string user_loc;
    try{
      user_loc = std::locale("").name();
    } catch (std::runtime_error&) {
      user_loc = std::locale().name();
    }
    logInfo(funcname,
            std::wstring(L"2. trying to fallback on locale: ")
            + fromString(user_loc));

    try { // search for "utf8" or variants

      string::size_type utf8pos;

      // sorry for this
      (utf8pos = user_loc.rfind("UTF")) != string::npos 
        || (utf8pos = user_loc.rfind("Utf")) != string::npos 
        || ( utf8pos = user_loc.rfind("utf")) != string::npos;

      if (utf8pos != string::npos)
        if (user_loc.at(utf8pos+3) == '8'
            || user_loc.at(utf8pos+4) == '8')
          return std::locale(user_loc.c_str());

    } catch (std::out_of_range& oor) {
      // out_of_range => not found
    }

    logWarn(funcname, fromString(user_loc)
            + L" does not seem to be an utf8 locale either");


    // 3. Else try to hack the user locale name a bit
    string::size_type delimpos;
    string nocodeset;
    if ((delimpos = user_loc.find_last_of('.')) != string::npos)
      nocodeset = user_loc.substr(0, delimpos);
    else if ((delimpos = user_loc.find_last_of('@')) != string::npos)
      nocodeset = user_loc.substr(0, delimpos);
    else
      nocodeset = user_loc;

    string hacked_names[2];
    hacked_names[0] = nocodeset + ".UTF-8";
    hacked_names[1] = nocodeset + ".utf8";

    for (int i=0; i<2; i++)
      try {
        logInfo(funcname, std::wstring(L"3n. trying to fall back on locale ")
                + fromString(hacked_names[i]));
        return trylocale(hacked_names[i].c_str());
      } catch (CodecException&) {
        logWarn(funcname,
                std::wstring(L"or please install locale ")
                + fromString(hacked_names[i]));
      }

    // 4. nothing found
    logFatal(funcname,
             std::wstring(L"Please install one of the following locales: "
                          NAME_OF_ANY_UTF8_LOCALE_AVAILABLE ", ")
             + fromString(user_loc) + L", "
             + fromString(nocodeset)
             + L".utf8"
      );
    throw CodecException(L"no UTF8 codec found");

    // TODO: fall back on:
    // - configuration file
    // - iconv
    // - MultiByteToWideChar(CP_UTF8,...)
    // - others?
    // See CAROB-74

  } // tryUTF8locale()
} // CarobNS namespace



JavaSocket::JavaSocket() throw (CodecException) :
socket_fd(-1),
connected(false),
canceled(false),
#ifdef CAROB_USE_ICONV // cpp-based factory :-]
  utf8_codec(*new IconvCodec("UTF-8"))
#else
  utf8_codec(*new MBSCodec(tryUTF8locale()))
#endif
{
}

JavaSocket::~JavaSocket()
{
  const wchar_t fctName[] = L"JavaSocket::~JavaSocket";

  delete &utf8_codec;

  if (isDebugEnabled())
    logDebug(fctName, L"About to close socket...");
  try
  {
    closeSystemSocket();
  }
  catch (SocketIOException e)
  {
      logInfo(fctName, L"Socket closure failed. Exception is: " + e.description());
  }
}

void JavaSocket::create(int domain)
    throw (ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::Create";
  if (isDebugEnabled())
    logDebug(fctName, L"Creating socket...");

  socket_fd = socket(domain, SOCK_STREAM, 0);

  if (socket_fd == -1)
  {
    if (isErrorEnabled())
      logError(fctName, L"Socket creation failed");
    throw ConnectionException(L"Socket creation failed");
  }

  // Set socket options
  // 1. non-blocking
  setBlockingMode(fctName, false);
  // 2. Disable nagle algorithm
  int opt_value = 1;
  if (setsockopt(socket_fd, IPPROTO_TCP, TCP_NODELAY,
                 reinterpret_cast<char*>(&opt_value), sizeof opt_value) == -1
#ifdef SO_NOSIGPIPE
  // 3. Disable sigpipe, see SystemDependantDefs.hpp
      || setsockopt(socket_fd, SOL_SOCKET, SO_NOSIGPIPE,
                    reinterpret_cast<char*>(&opt_value), sizeof opt_value) == -1
#endif
    )
  {
    wstring msg = L"Could not set socket options: setsockopt returned error: ";
    msg += StaticCodecs::fromString(strerror(errno));
    if (isErrorEnabled())
      logError(fctName, msg);
    throw ConnectionException(msg);
  }

  if (isDebugEnabled())
    logDebug(fctName, L"Socket created.");
}

bool JavaSocket::connectTo(const SocketAddress& addr)
    throw (ConnectionException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::connectTo";
  connected = false;
  if (!isValid())
  {
    if (isErrorEnabled())
      logError(fctName, L"Invalid socket. Did you create() it ?");
    throw ConnectionException(L"Invalid socket for connection");
  }

  // try non blocking connect
  int resp = ::connect(socket_fd, addr.getAddressPtr(), addr.getAddressLength());
  if (resp==0)
  {
    // connection succeeded, set back to blocking mode
    setBlockingMode(fctName, true);
    connected = true;
    return true;
  }
  if (errno == EINPROGRESS)
  {
    // connection is not established yet. Loop until connection ok, error or cancelation

    if (isDebugEnabled())
      logDebug(fctName, L"Socket connect() in progress... polling");
    while (!canceled)
    {
      int retVal = pollOnSingleFd(socket_fd, 1000, true); // 1000 = 1 second timeout
      if (retVal > 0)
      {
        // Check that the socket is not in error:
        int32_t sockErr;
        socklen_t sockErrLen = sizeof(sockErr);
        if (getsockopt(socket_fd, SOL_SOCKET, SO_ERROR, (void*)(&sockErr), &sockErrLen) < 0)
        {
          wstring msg = L"Could not get socket option: getsockopt returned error: ";
          msg += StaticCodecs::fromString(strerror(errno));
          if (isErrorEnabled())
            logError(fctName, msg);
          throw ConnectionException(msg);
        }
        // If the error code returned is 0 => connect ok
        if (sockErr == 0)
        {
          if (isDebugEnabled())
            logDebug(fctName, L"Connection succeeded");
          // connection succeeded, set back to blocking mode
          setBlockingMode(fctName, true);
          connected = true;
          return true;
        }
        else
        {
          wstring msg = L"connect() returned error: ";
          msg += StaticCodecs::fromString(strerror(sockErr));
          if (isErrorEnabled())
            logError(fctName, msg);
          throw ConnectionException(msg);
        }
      }
    }
  }
  if (canceled && isDebugEnabled())
    logDebug(fctName, L"Connection interrupted.");

  // At this point, we either got an error or were canceled. Return an error
  // anyway
  return false;
}

namespace {
inline int portable_close(int socket_fd)
{
#ifdef __MINGW32__
  return closesocket(socket_fd);
#else
  return close(socket_fd);
#endif
}
}

bool JavaSocket::closeSystemSocket() throw (SocketIOException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::closeSocket";
  if (isValid())
  {
    if (portable_close(this->socket_fd) != 0)
    {
      wstring msg(L"Could not close socket. Error is ");
#ifdef __MINGW32__
      msg += toUserString(WSAGetLastError());
#else
      msg += StaticCodecs::fromString(strerror(errno));;
#endif
      throw SocketIOException(msg);
      return false;
    }
    if (isDebugEnabled())
      logDebug(fctName, L"Socket closed.");
  }
  socket_fd = -1;
  connected = false;
  return true;
}

// Localization (Chapter 22) in
// /usr/share/doc/gcc-4.0-base/libstdc++/html/documentation.html

size_t JavaSocket::writeJavaUTF(const wstring& str) const
  throw (SocketIOException, CodecException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::writeJavaUTF";

  std::string utf8str;

  try
  {
    utf8str = utf8_codec.nomutex_encode(str);
  }
  catch (const CodecException&)
  { // Don't leave connection in such a rotten state, see CAROB-101
    portable_close(this->socket_fd);
    throw;
  }

  int32_t netlen = htonl(utf8str.length());

  // First write number of bytes to follow as
  sendToSocket(fctName, &netlen, sizeof(netlen), SOCKET_SEND_FLAGS);
  sendToSocket(fctName, utf8str.data(), utf8str.length(), SOCKET_SEND_FLAGS);

  return utf8str.length();
}

size_t JavaSocket::readJavaUTF(wstring& s) const
  throw (SocketIOException, CodecException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::readJavaUTF";

  int32_t lenRecNet;
  //the converted size
  uint16_t lenRec;
  size_t sizeRead = 0;
  //First get size
  receiveFromSocket(fctName, &lenRecNet, sizeof(lenRecNet), 0);
  lenRec = ntohl(lenRecNet); // number of bytes to come

  uint8_t* utfStr = new uint8_t[lenRec];
  receiveFromSocket(fctName, utfStr, lenRec, 0);
  const std::string received(reinterpret_cast<char*>(utfStr), lenRec);

  try
  {
    s = utf8_codec.nomutex_decode(received);
  }
  catch (const CodecException&)
  { 
    delete[] utfStr;
    // Don't leave connection in such a rotten state, see CAROB-101
    portable_close(this->socket_fd);
    throw;
  }
  sizeRead = (size_t)lenRec;

  delete[] utfStr;
  return sizeRead;
}

void JavaSocket::writeJavaInt(int32_t i) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::WriteInt32";
  // we have to convert the byte order
  int32_t conv = htonl(i);
  sendToSocket(fctName, &conv, sizeof(conv), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaInt(int32_t& i) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Socket::ReadInt32";
  // the data will be received with network byte order.
  // so we must convert it after reception
  int32_t rec = 0;

  receiveFromSocket(fctName, &rec, sizeof(rec), 0);
  i = ntohl(rec);
}

void JavaSocket::writeJavaLong(int64_t i) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::writeJavaLong";
  // we have to convert the byte order
  int64_t conv = htonll(i);
  sendToSocket(fctName, &conv, sizeof(conv), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaLong(int64_t& i) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Socket::readJavaLong";
  // the data will be received with network byte order.
  // so we must convert it after reception
  int64_t rec = 0;

  receiveFromSocket(fctName, &rec, sizeof(rec), 0);
  i = ntohll(rec);
}

void JavaSocket::writeJavaBool(bool b) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::writeJavaBool";
  uint8_t byteToSend = b; // the implicit casts do it right
  //No byte order problem here: the only value read is '0' for false
  //all other values are true
  sendToSocket(fctName, &byteToSend, sizeof(byteToSend), SOCKET_SEND_FLAGS);
}

void JavaSocket::readJavaBool(bool& b) const throw (SocketIOException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::readJavaBool";

  uint8_t rec;
  receiveFromSocket(fctName, &rec, sizeof(rec), 0);
  b = rec; // the implicit cast does it right
}

void JavaSocket::readJavaBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::readJavaBytes";
  receiveFromSocket(fctName, data, length, 0);
}

void JavaSocket::writeJavaBytes(int32_t length, java_byte* data) const
    throw (SocketIOException, UnexpectedException)
{
  const wchar_t fctName[] = L"JavaSocket::writeJavaBytes";
  sendToSocket(fctName, data, length, SOCKET_SEND_FLAGS);
}

/* static */
int JavaSocket::pollOnSingleFd(int socketFd, int pollTimeoutInMs, bool pollOnWrites/*=false*/)
{
  // return of poll/select function
  int retVal = 0;
#ifndef CAROB_USE_SELECT
  // create poll fd
  struct pollfd pfd;
  pfd.fd = socketFd;
  if (pollOnWrites)
    pfd.events = POLLOUT;
  else
    pfd.events = POLLIN|POLLPRI;
  // The pfd.revents is is going to be modified
  retVal = poll(&pfd, 1, pollTimeoutInMs);
#else
  // create select fd
  fd_set sfd;
  FD_ZERO(&sfd);
  struct timeval tv;
  tv.tv_sec = pollTimeoutInMs / 1000;
  tv.tv_usec = (pollTimeoutInMs % 1000) *1000; // milli -> micro seconds conversion
  FD_SET(socketFd, &sfd);
  if (pollOnWrites)
    retVal = select(socketFd+1, NULL, &sfd, NULL, &tv);
  else
    retVal = select(socketFd+1, &sfd, NULL, NULL, &tv);
#endif
  if (retVal == -1)
  {
    if (isWarnEnabled())
    {
      logWarn(L"JavaSocket::pollOnSingleFd",
          wstring(POLL_FUNCTION) + L" returned error: "
          + StaticCodecs::fromString(strerror(errno)));
    }
  }
  return retVal;
}

/** Returns the number of octets read or the (negative) error from recv() */
int JavaSocket::recvFully(void *buf, const int len, const int flags)
    const throw (SocketIOException, UnexpectedException)
{
  int alreadyRead = 0; //Keeps the whole size read
  int readThisTime = 0; //The size read at each receive
  while (!canceled && alreadyRead < len)
  {
    int retVal = pollOnSingleFd(socket_fd, 1000); // 1000 = 1 second timeout
    // if (retVal == 0) => timeout, no response, do nothing
    if (retVal > 0) // data is ready to be read
    {
      readThisTime = recv(socket_fd,
                          (char*)(static_cast<uint8_t*>(
                                        ((static_cast<uint8_t*>(buf))+alreadyRead))),
                          len-alreadyRead,
                          flags);
      if (readThisTime <= 0)
      {
        //this is an error
        return readThisTime;
      }
      alreadyRead += readThisTime;
    }
  }
  // We were shutdown()
  if (!connected)
    throw SocketIOException(L"JavaSocket has been shutdown, read aborted");
  return alreadyRead;
}

void JavaSocket::sendToSocket(const wchar_t fctName[],
    const void* buf, int len, int flags) const
    throw (SocketIOException, UnexpectedException)
{
  // in non blocking mode, send will block until len bytes are send. So no need
  // to loop here
  int ret = send(socket_fd, static_cast<const char*>(buf), len, flags);
  if ( ret == -1 )
  {
    throw SocketIOException(std::wstring(L"(") + fctName + L"): could not write"
        L" data to socket");
  }
  if (ret != len)
  {
    // should never happen
    throw SocketIOException(std::wstring(L"(") + fctName + L"): only " + toUserString(ret)
        + L"/" + toUserString(len) + L" bytes send while sending data to socket");
  }
}

void JavaSocket::receiveFromSocket(const wchar_t fctName[],
    void *buf, int len, int flags) const
    throw (SocketIOException, UnexpectedException)
{
  if (0 == len)
    return;

  int status = recvFully(buf, len, flags);

  if (status == -1)
    throw SocketIOException(std::wstring(L"(") + fctName + L"): Could not read from socket.");

  if (status == 0)
    throw SocketIOException(std::wstring(L"(") + fctName + L"): Peer reset connection while receiving");

  // else it went fine
  // if (status != len)
  //  throw SocketIOException(std::wstring(L"(") + fctName + L"): something is really wrong!");
}

void JavaSocket::setBlockingMode(const wchar_t fctName[], bool blocking)
    throw (ConnectionException, UnexpectedException)
{
  long fcntlopt = fcntl(socket_fd, F_GETFL, NULL);
  if (fcntlopt < 0)
  {
    wstring msg = L"Could not get socket options: fcntl returned error: ";
    msg += StaticCodecs::fromString(strerror(errno));
    if (isErrorEnabled())
      logError(fctName, msg);
    throw ConnectionException(msg);
  }
  if (blocking)
    fcntlopt &= (~O_NONBLOCK);
  else
    fcntlopt |= O_NONBLOCK;
  if(fcntl(socket_fd, F_SETFL, fcntlopt) < 0)
  {
    wstring msg = L"Could not set socket options: fcntl returned error: ";
    msg += StaticCodecs::fromString(strerror(errno));
    if (isErrorEnabled())
      logError(fctName, msg);
    throw ConnectionException(msg);
  }
}


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 */
