/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Zsolt Simon
 * Contributor(s): Gilles Rayrat, Marc Herbert
 */
#include "ParameterStatement.hpp"

#include "Connection.hpp"
#include "BigDecimal.hpp"

#include "Common.hpp"
#include "CarobException.hpp"

#include <sstream>

#ifdef __MINGW32__
  #include "math.h" // C
#else
  #include <cmath>  // C++
#endif

#include "C99Math.h" // signbit, isinf, isnan


namespace CarobNS {
/**
 * All tags must have the same length
 */
const std::wstring BYTE_TAG                          = L"b|";
const std::wstring BYTES_TAG                         = L"B|";
const std::wstring BLOB_TAG                          = L"c|";
const std::wstring CLOB_TAG                          = L"C|";
const std::wstring BOOLEAN_TAG                       = L"0|";
const std::wstring BIG_DECIMAL_TAG                   = L"1|";
const std::wstring DATE_TAG                          = L"d|";
const std::wstring DOUBLE_TAG                        = L"D|";
const std::wstring FLOAT_TAG                         = L"F|";
const std::wstring INTEGER_TAG                       = L"I|";
const std::wstring LONG_TAG                          = L"L|";
const std::wstring NULL_TAG                          = L"N|";
const std::wstring NULL_STRING_TAG                   = L"n|";
const std::wstring OBJECT_TAG                        = L"O|";
const std::wstring REF_TAG                           = L"R|";
const std::wstring SHORT_TAG                         = L"s|";
const std::wstring STRING_TAG                        = L"S|";
const std::wstring TIME_TAG                          = L"t|";
const std::wstring TIMESTAMP_TAG                     = L"T|";
const std::wstring URL_TAG                           = L"U|";
const std::wstring REGISTER_OUT_PARAMETER            = L"o|";
const std::wstring REGISTER_OUT_PARAMETER_WITH_SCALE = L"w|";
const std::wstring REGISTER_OUT_PARAMETER_WITH_NAME  = L"W|";

/** Tag maker for parameters */
const std::wstring TAG_MARKER                        = L"!%";
/** Escape for tag maker */
const std::wstring TAG_MARKER_ESCAPE                 = TAG_MARKER
                                                     + L";";

/** Tag for parameters start delimiter */
const std::wstring START_PARAM_TAG                   = L"<"
                                                     + TAG_MARKER;
/** Tag for parameters end delimiter */
const std::wstring END_PARAM_TAG                     = L"|"
                                                     + TAG_MARKER
                                                     + L">";

}

using std::wstring;

using namespace CarobNS;

ParameterStatement::ParameterStatement(Connection *c, const wstring &sqlStatement) :
Statement(c)
{
  sql_request = trim(sqlStatement);

  // The following two boolean switches are used to make sure we're not
  // counting "?" in either strings or metadata strings. For instance the
  // following query:
  // select '?' "A ? value" from dual
  // doesn't have any parameters.
  bool inString = false, inMetaString = false; 

  std::wstring::const_iterator isql = sql_request.begin();
  std::wstring::const_iterator isql_end = sql_request.end();
  std::wstring::const_iterator ilast = isql;
  
  inStrings.clear();

  for (; isql != isql_end; ++isql)
  {
    switch (*isql)
    {
      case L'\'':
        inString = !inString;
        break;
      case L'"':
        inMetaString = !inMetaString;
        break;
      case L'?':
        if (!inString && !inMetaString)
        {
         inStrings.push_back(wstring());
         ilast=isql+1;
        }
    }
  }
}

ParameterStatement::~ParameterStatement()
{
}

void ParameterStatement::close()
{
  Statement::close();
  
  sql_request.clear();
  inStrings.clear();
}

void ParameterStatement::closeStatement()
{
  Statement::close();
  
  for (size_t i = 0; i < inStrings.size(); i++ )
  {
    inStrings[i] = L"";
  }
}

void ParameterStatement::clearParameters()
{
  std::vector<wstring>::iterator iparam = inStrings.begin();
  std::vector<wstring>::iterator iparam_end = inStrings.end();
  
  for (; iparam != iparam_end; (*iparam++).clear()) ;
}

const wchar_t base64[] = L"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

wstring ParameterStatement::base64Encode(const void *buffer, size_t len)
{
  unsigned char rc[3];
  int usedChars;
  wstring result;
  char *p = (char *)buffer;

  result.reserve(len*4/3+4);

  while(len > 0)
  {
    for (int i = usedChars = 0; i < 3; i++)
    {
      if(len) {
        usedChars++;
        rc[i] = *p++;
        len--;
      }
      else
        rc[i] = 0;
    }

    result.push_back(base64[(rc[0]&0xFC) >> 2]);
    result.push_back(base64[((rc[0]&0x03) << 4)|((rc[1]&0xF0) >> 4)]);
    result.push_back(usedChars >= 2 ? base64[((rc[1]&0x0F) << 2) | ((rc[2]&0xC0) >> 6)] : L'=');
    result.push_back(usedChars >= 3 ? base64[rc[2]&0x3F] : L'=');
  }

  return result;
}

namespace {
  std::locale locC = std::locale::classic();
}


  /////  Template & specialized filters /////

//This code is very similar to Common.cpp toWString(<>) => factorize ?

namespace CarobNS {

using std::basic_ostream;

  /// template filters do nothing by default ///

/**
 * Stream filter: intercept (only) one value, modify it before printing it, and returns the
 * filtered stream before disappearing.
 */
template <class T, class Ch = wchar_t>
class oneShotFilter
{
  basic_ostream<Ch>& wrapped_stream;
public:
  oneShotFilter(basic_ostream<Ch>& s) : wrapped_stream(s) { } ;
  basic_ostream<Ch>&
 // this is the actual filtering operator
  operator<< (T value)
  { // by default filter nothing, just output the value
    wrapped_stream << value;
    // then return the underlying stream (we filter only once!)
    return wrapped_stream;
  }
};

template <class T, class Ch = wchar_t>
class streamSettings
{
public:
  void doManip(basic_ostream<Ch>& s) const { /* do nothing by default */ };  
};

/**
 * Operator<< doing the actual filtering work: modify the settings
 * of the underlying standard stream, then returns a one-shot filter.
 */
template <class T, class Ch>
oneShotFilter<T, Ch>
operator<< (basic_ostream<Ch>& filtered_std_stream, const streamSettings<T, Ch>& setfunctor)
{
  // modify settings of underlying stream
  setfunctor.doManip(filtered_std_stream);
  // then return a filter instance
  return oneShotFilter<T, Ch>(filtered_std_stream);
}

  /////  Filter specializations ////

// Partial specialization code shared by float & double

// std::numeric_limits<...>::digits10 is NOT the precision we want, cause
// digits10 is the maximum number of base10 digits that (float|double)
// can hold without loss, whereas we want the opposite: the minimum
// number of base10 digits required to describe any (float|double
// value) without loss.
// Let's put it in another way: digits10 matters when you want to go
// from base10 to floating-point and back to base10 without loss.
// What we do here is to go from floating-point to base10 and back to
// floating point.

// g++ 4.0.2 defines for instance:
// <float>::digits10 = 6 and <double>::digits10 = 15

// For IEEE754's floats the maximum required number of base10 digits is 9 (nine),
// and for IEEE754's doubles                                 it's 17 (seventeen)
// "What Every Computer Scientist Should Know About Floating Point Arithmetic"
// 1991 - David Goldberg
// We add one digit for some safety wrt to buggy converters (but if
// they are really too buggy this won't be enough to save us)

// In scientific mode, we have:
// ios_base::precision() digits _after_ comma, +1 digit _before_ comma

// See paragraph "Conversions" at:
// http://www2.hursley.ibm.com/decimal/ for more details and proofs.

template<int precision, class Ch>
class FPSettings
{
public:
  void doManip(basic_ostream<Ch>& s) const
  {
    s.imbue(locC);
    s.setf(std::ios_base::scientific, std::ios_base::floatfield);
    s.precision(precision);
  }
};

// settings specialization for double
template<class Ch>
class streamSettings<double, Ch> : public FPSettings<17, Ch>  { };

// settings specialization for float
template<class Ch>
class streamSettings<float, Ch> : public FPSettings<9, Ch>  { };



template<typename FP>
class FPoneShotFilter
{
  basic_ostream<wchar_t>& wrapped_stream;
public:
  FPoneShotFilter(basic_ostream<wchar_t>& s) : wrapped_stream(s) { } ;

  basic_ostream<wchar_t>&
  // This is the actual filtering operator
  operator<< (FP value)
  { 
    // Warning: isnan() and isinf() are C99, not C++98!

    // first the common case
    if (!isnan(value) && !isinf(value)) {
      wrapped_stream << value;
      return wrapped_stream;
    }

    // now the exceptional cases
    // Bug: java.lang.FloatingDecimal#readJavaFormatString()
    // ignores negative sign of NaN
    if (signbit(value))
      wrapped_stream << L"-";
    
    // Too bad Standard C99 and Java could not agree on the same strings :-(
    if (isinf(value))
      wrapped_stream << L"Infinity"; // see java.lang.(Float|Double)#toString()
    else if (isnan(value))  // forget about quiet vs signaling
      wrapped_stream << L"NaN";

    return wrapped_stream;
  }
};

// filter specialization for double
template<>
class oneShotFilter<double, wchar_t> : public FPoneShotFilter<double>
{
public:
  oneShotFilter(basic_ostream<wchar_t>& s) : FPoneShotFilter<double>(s) { } ;
};

// filter specialization for float
template<>
class oneShotFilter<float, wchar_t> : public FPoneShotFilter<float>
{
public:
  oneShotFilter(basic_ostream<wchar_t>& s) : FPoneShotFilter<float>(s) { } ;
};

// template specialization for bool
template<class Ch>
class streamSettings<bool, Ch>
{
public:
  void doManip(basic_ostream<Ch>& s) const
  {
    // see: java.lang.Boolean#toBoolean()
    s.imbue(locC);
    s << std::boolalpha;
  };
};

// template specialization for wstring
template <>
class oneShotFilter<wstring>
{
  basic_ostream<wchar_t>& wrapped_stream;
public:
  oneShotFilter(basic_ostream<wchar_t>& s) : wrapped_stream(s) { } ;
  basic_ostream<wchar_t>&
 // this is the actual filtering operator
  operator<< (wstring value)
  { 
    wrapped_stream <<
      (value == NULL_TAG ? NULL_STRING_TAG : replaceAll(value, TAG_MARKER, TAG_MARKER_ESCAPE));
    // then return the underlying stream (we filter only once!)
    return wrapped_stream;
  }
};



} // namespace CarobNS

  //// end of template filters ////


template <class T> void ParameterStatement::set(const int paramIndex, const std::wstring &typeTag, 
        const T& value) 
    throw (DriverException, UnexpectedException)
{
  if (paramIndex < 1 || paramIndex > static_cast<int>(inStrings.size()))
    throw DriverException(L"Parameter index out of range.");
    
  // As an optimization, we could make this stream a member.
  // That would require
  // - a member lock for thread-safety
  // - carefully checking that streamSettings<T1>() does not step on streamSettings<T2>() toes
  std::wostringstream buffer;

  // streamSettings<T>() is a function object that will:

  // 1. Modify the default stream settings of 'buffer'
  // For instance it will set the minimum precision() to avoid
  // rounding when T is a floating-point type.

  // 2. Return a oneShotFilter<T>() that may transform the value
  // printed. For instance END_PARAM_TAG will be escaped in value when
  // T is std::wstring.

  // streamSettings<defaultT>() and oneShotFilter<defaultT>() do nothing.
  buffer << START_PARAM_TAG << typeTag << streamSettings<T>() << value << END_PARAM_TAG;

  inStrings[paramIndex-1] = buffer.str();
}

wstring ParameterStatement::compileQuery(bool throwException) 
    throw (DriverException)
{
  std::vector<wstring>::const_iterator iparam = inStrings.begin();
  std::vector<wstring>::const_iterator iparam_end = inStrings.end();
  std::wostringstream sbuf;

  for (int i=1; iparam != iparam_end; ++iparam, ++i)
  {
    if (iparam->empty())
    {
      if (throwException)
        throw DriverException(L"Parameter " + toUserString(i) + L" is incorrect");
      else
        sbuf << L"?";
    }
    else
      sbuf << *iparam;
  }
  
  return sbuf.str();
}

void ParameterStatement::setNull(const int parameterIndex, const int sqlType)
    throw (DriverException, UnexpectedException)
{
  set<std::wstring>(parameterIndex, NULL_TAG, toWString(sqlType));
}


template<typename CParamType> void
ParameterStatement::setBoolean(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, BOOLEAN_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setByte(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, BYTE_TAG,  x);
}

template<typename CParamType> void
ParameterStatement::setShort(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, SHORT_TAG,  x);
}

template<typename CParamType> void
ParameterStatement::setInt(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, INTEGER_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setLong(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, LONG_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setFloat(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, FLOAT_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setDouble(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, DOUBLE_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setBigDecimal(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, BIG_DECIMAL_TAG, x);
}

template<typename CParamType> void
ParameterStatement::setString(int parameterIndex, const CParamType& x)
    throw (DriverException, UnexpectedException)
{
    set<CParamType>(parameterIndex, STRING_TAG, x);
}

void ParameterStatement::setByteArray(int parameterIndex, const void *buffer, size_t len)
    throw (DriverException, UnexpectedException)
{
  set<std::wstring>(parameterIndex, BYTES_TAG, base64Encode(buffer, len));  
}

void ParameterStatement::setBlob(int parameterIndex, const void *buffer, size_t len)
    throw (DriverException, UnexpectedException)
{
  set<std::wstring>(parameterIndex, BLOB_TAG, base64Encode(buffer, len));
}

/*
 * Explicit instantiation of supported/approved setters
 */

/*
 * 1. First the zero-converting setters. The C-types instantiated here
 * are bit-compatible with the corresponding java type that will be
 * deserialized by the controller.
 *
 */

template void
ParameterStatement::setBoolean<bool>(int parameterIndex, const bool& x);
template void
ParameterStatement::setByte<int8_t>(int parameterIndex, const int8_t& x);
template void
ParameterStatement::setShort<int16_t>(int parameterIndex, const int16_t& x);
template void
ParameterStatement::setInt<int32_t>(int parameterIndex, const int32_t& x);
template void
ParameterStatement::setLong<int64_t>(int parameterIndex, const int64_t& x);
template void
ParameterStatement::setFloat<float>(int parameterIndex, const float& x);
template void
ParameterStatement::setDouble<double>(int parameterIndex, const double& x);
template void
ParameterStatement::setBigDecimal<BigDecimal>(int parameterIndex, const BigDecimal& x);
template void
ParameterStatement::setString<std::wstring>(int parameterIndex, const std::wstring& x);
template void // for convenience. See limitation in .hpp doxygen
ParameterStatement::setString<const wchar_t *>(int parameterIndex, const wchar_t * const& x);


/*
 * 2. Then the "sstream-converting" setters. Add new lines here only
 * if you really understand what you are doing. -- MH.
 *
 * Related:
 * - "ODBC Programmer's Reference: Converting Data from C to SQL Data Types"
 *   <http://msdn.microsoft.com/library/default.asp?url=/library/en-us/odbc/htm/odbcconverting_data_from_c_to_sql_data_types.asp>
 * - Sun's "Getting Started" JDBC online guide, chapter "Mapping SQL and Java Types"
 *   <http://java.sun.com/j2se/1.4.2/docs/guide/jdbc/getstart/mapping.html>
 */

/*
 * Unfortunately we currently lose the original type, and the controller
 * cannot/does not make the difference between:
 * "setFloat<float>(f1)" vs "setFloat<otherType>(convertedf2)"
 * Ideally we would be much more transparent than that, performing NO
 * type conversion at all and letting the backend driver doing it
 * alone by himself.
 */

// Current semantic: "true".equalsIgnoreCase("xxxx")
// See: java.lang.Boolean#toBoolean()
template void
ParameterStatement::setBoolean<std::wstring>(int parameterIndex, const std::wstring& x);


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
