/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Emic Networks
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Marc Herbert
 */

#include "SQLDataSerialization.hpp"

#include "BigDecimal.hpp"
#include "TypeTag.hpp"

#include "CarobException.hpp"
#include "Common.hpp"

#include <cmath>


// What IEEE 754 grants that matters to us:
// - sizeof(float) = 32 && sizeof(double) = 64
// - same standard binary representation as the one sent
//   by Java's DataOutput class

#if !defined __STDC_IEC_559__
#warning "__STDC_IEC_559__ is not defined: only IEEE 754 platforms are supported"
#endif
 // run-time equivalent is: numeric_limits<double|float>.is_iec559()

#if (CHAR_BIT != 8)
#error "CHAR_BIT != 8 is not supported"
#endif

using std::wstring;

using namespace CarobNS;

namespace {
    // forward
    extern const bool floats_inverted_endianness;
}


// String
ResultSetDataType stringDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  wstring* strRead = new wstring;
  input>>*strRead;
  ResultSetDataType res;
  res.as_other = static_cast<void*>(strRead);
  return res;
}

// BigDecimal
ResultSetDataType bigDecimalDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  BigDecimal* bd = new BigDecimal(input);
  res.as_other = static_cast<void*>(bd);
  return res;
}

// Int
ResultSetDataType integerDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  input>>res.as_int;
  return res;
}

// Bool
ResultSetDataType booleanDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  input>>res.as_bool;
  return res;
}

// Long
ResultSetDataType longDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  input>>res.as_long;
  return res;
}

namespace {

// Float
  /**
   * Relatively safe "reinterpret_cast" of a native IEEE 754 float to
   * its uint32_t big endian representation. Usually returns a _big
   * endian_ uint32_t since endianness of floats is usually the same
   * as for integer types, even if not explicitely granted.
   *
   * @param f value to convert
   */
inline uint32_t
floatToNetworkU32Bits(float f)
{
    // "union casting" using { float; uint32_t; } or the simpler:
    //        return * (uint32_t *) &f;
    // are "undefined behaviour" because of aliasing rules, alignement issues, etc.
    // See discussion thread "reinterpret_cast to extract bits" thread in comp.lang.c++
    // in may 2004
    // http://groups.google.fr/group/comp.std.c++/browse_frm/thread/272f1199322af93d/7d3368013656d923
    // The more portable code below should not cost more than a couple
    // of CPU cycles/converted float
    const unsigned char *bytes = reinterpret_cast<const unsigned char *>(&f);

    // If the arch is all big endian (both for integers and floats),
    // then there is no subtility and we return a native integer
    // (big endian)

    // If the arch is all little endian, then we returned a "reversed"
    // result: big endian again.

    return (uint32_t) bytes[0] << 24
        | (uint32_t) bytes[1] << 16
        | (uint32_t) bytes[2] << 8
        | (uint32_t) bytes[3];

    // if the arch is mixed-up, we fail.

}

/**
 * Big endian uint32_t to native float, see floatToNetworkU32Bits
 */
inline float
networkU32BitsToFloat(const uint32_t ui)
{
    float res;
    unsigned char *f_bytes = reinterpret_cast<unsigned char *>(&res);

    f_bytes[0] = ui >> 24;
    f_bytes[1] = ui >> 16;
    f_bytes[2] = ui >> 8;
    f_bytes[3] = ui;

    return res;
}

// Double
/**
 * Native double to big endian uint64_t, see floatToNetworkU32Bits
 */
inline uint64_t
doubleToNetworkU64Bits(double d)
{
    const unsigned char *bytes = reinterpret_cast<const unsigned char *>(&d);

    return (uint64_t) bytes[0] << 56
        | (uint64_t) bytes[1] << 48
        | (uint64_t) bytes[2] << 40
        | (uint64_t) bytes[3] << 32
        | (uint64_t) bytes[4] << 24
        | (uint64_t) bytes[5] << 16
        | (uint64_t) bytes[6] << 8
        | (uint64_t) bytes[7];
}

/**
 * Big endian uint64_t to native double, see floatToNetworkU32Bits
 */
inline double
networkU64BitsToDouble(const uint64_t ui)
{
    double res;
    unsigned char *d_bytes = reinterpret_cast<unsigned char *>(&res);

    d_bytes[0] = ui >> 56;
    d_bytes[1] = ui >> 48;
    d_bytes[2] = ui >> 40;
    d_bytes[3] = ui >> 32;
    d_bytes[4] = ui >> 24;
    d_bytes[5] = ui >> 16;
    d_bytes[6] = ui >> 8;
    d_bytes[7] = ui;

    return res;
}


} // anonymous namespace


/**
 * Converts the input integer which is in in IEEE 754 floating-point
 * "single format" bit layout to the corresponding float. 
 * Bit 31 (the most significant) is the sign bit,
 * bits 30-23 (masked by 0x7f800000) represent the exponent,
 * and bits 22-0 (masked by 0x007fffff) are the mantissa.
 * This function leaves NaN alone.
 */
ResultSetDataType floatDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  uint32_t bigEndianInt;

  input.readBytes(4, reinterpret_cast<java_byte *>(&bigEndianInt));
  
  if (floats_inverted_endianness)
      throw NotImplementedException(L"Inverted endianness Not Implemented Yet");
  else // takes care of cast + endianness swap if needed
      res.as_float = networkU32BitsToFloat(bigEndianInt);

  return res;
}

/**
 * Converts the argument in IEEE 754 floating-point "double format" bit layout
 * to the corresponding float. Bit 63 (the most significant) is the sign bit,
 * bits 62-52 (masked by 0x7ff0000000000000L) represent the exponent,
 * and bits 51-0 (masked by 0x000fffffffffffffL) are the mantissa.
 * This function leaves NaN alone.
 */
ResultSetDataType doubleDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  uint64_t bigEndianInt;

  input.readBytes(8, reinterpret_cast<java_byte *>(&bigEndianInt));

  if (floats_inverted_endianness)
      throw NotImplementedException(L"Inverted endianness Not Implemented Yet");
  else // takes care of cast + endianness swap if needed
      res.as_double = networkU64BitsToDouble(bigEndianInt);

  return res;
}

// Byte Array
ResultSetDataType byteArrayDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  LargeData* ld = new LargeData;
  input>>ld->length;
  ld->data = new java_byte[ld->length];
  input.readBytes(ld->length, ld->data);
  res.as_other = static_cast<void*>(ld);
  return res;
}

// Date
/**
 * The long returned is the number of milliseconds since
 * January 1, 1970, 00:00:00 GMT
 */
ResultSetDataType SQLDateDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  input>>res.as_long;
  return res;
}

// Time
/**
 * The integer returned is Returns the number of milliseconds since
 * January 1, 1970, 00:00:00 GMT
 */
ResultSetDataType SQLTimeDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  ResultSetDataType res;
  input>>res.as_int;
  return res;
}

// Timestamp
/**
 */
ResultSetDataType SQLTimestampDeserializer(const DriverSocket& input)
    throw (SocketIOException, UnexpectedException)
{
  //Timestamp is a long *and* an integer. Make use of our struct to put
  //both values in it
  ResultSetDataType res;
  SQLTimeStamp* tsPtr = new SQLTimeStamp;
  input>>tsPtr->time;
  tsPtr->time = ((tsPtr->time/1000)*1000);
  input>>tsPtr->nanos;
  res.as_other = tsPtr;
  return res;
}

/** This serializer is only a placeholder for the undefined type tag,
 * it should never be called (nulls are not serialized, see
 * DriverResultSet#receiveRows()) */
ResultSetDataType undefinedDeserializer(const DriverSocket& input)
    throw (DriverException, UnexpectedException)
{
  throw DriverException(L"undefinedDeserializer should never be called!");
}

deserializerPtr SQLDataSerialization::getDeserializer(TypeTag ttPrm)
    throw (NotImplementedException, UnexpectedException)
{
  switch (ttPrm)
  {
    case TT_STRING:
      return stringDeserializer;
    break;
    case TT_BIGDECIMAL:
      return bigDecimalDeserializer;
    break;
    case TT_BOOLEAN:
      return booleanDeserializer;
    break;
    case TT_INTEGER:
      return integerDeserializer;
    break;
    case TT_LONG:
      return longDeserializer;
    break;
    case TT_FLOAT:
      return floatDeserializer;
    break;
    case TT_DOUBLE:
      return doubleDeserializer;
    break;
    case TT_BYTE_ARRAY:
      return byteArrayDeserializer;
    break;
    case TT_SQL_DATE:
      return SQLDateDeserializer;
    break;
    case TT_SQL_TIME:
      return SQLTimeDeserializer;
    break;
    case TT_SQL_TIMESTAMP:
      return SQLTimestampDeserializer;
    break;
    case TT_UNDEFINED:
      return undefinedDeserializer;
    break;
    case TT_BLOB:
      throw NotImplementedException(L"Blob deserialization unsupported yet");
    break;
    case TT_JAVA_SERIALIZABLE:
      throw NotImplementedException(L"Unsupported deserialization of java objects");
    break;
    default:
      throw NotImplementedException(L"Failed to deserialize unknown type");
  }
}

#ifdef CAROB_DEBUG_FLOAT
namespace CarobNS {
#else
namespace {
#endif

// See floatToNetworkU32Bits() above
uint32_t F2I(const float f)
{
    uint32_t result;
    const unsigned char *fbytes = reinterpret_cast<const unsigned char *>(&f);
    unsigned char *ibytes = reinterpret_cast<unsigned char *>(&result);

    for (int i=0; i<4; i++)
        ibytes[i] = fbytes[i];

    return result;
}

// See floatToNetworkU32Bits() above
uint64_t D2I(const double f)
{
    uint64_t result;
    const unsigned char *fbytes = reinterpret_cast<const unsigned char *>(&f);
    unsigned char *ibytes = reinterpret_cast<unsigned char *>(&result);

    for (int i=0; i<8; i++)
        ibytes[i] = fbytes[i];

    return result;
}

} // unnamed or CarobNS namespace


namespace {

bool
floats_little_endianness()
{ 

    // build some magical float constants
    
    // e125f = 2^-125, so we have: sign = 0, exponent = 2, mantissa =
    // 0 (first '1' in m is implicit). So that's just one bit set to
    // '1' in the hybrid (sign/exponent) octet, and all the other bits
    // are '0'
    float e125f_ = 1;
    for (int i=0; i < 5; i++)
        e125f_ /= (1 << 25);

    // In the mantissa, set the bit before the least significant to
    // '1'. That should set the least significant byte of the mantissa
    // to '2'
    float e125fbis_ = e125f_ + e125f_ / (1 << 22);

    // so now e125fbis_ should look in memory either like { 1, 0, 0,
    // 2} or like { 2, 0, 0, 1}, depending on endianness. Let's go and
    // check that.

    const uint32_t e125f = F2I(e125fbis_);
    const char *e125f_uc = reinterpret_cast<const char*>(&e125f);


    // build some magical double constants

    // e1007d = 2^-1007, and same conclusion as e125f above
    double e1007d_ = 1;
    for (int i=0; i < 53; i++)
        e1007d_ /= (1 << 19);
    // insert a '1' in the digit before the last
    double e1007dbis_ = e1007d_ + e1007d_ / ((uint64_t) 1 << 51);

    const uint64_t e1007d = D2I(e1007dbis_);
    const char *e1007d_uc = reinterpret_cast<const char*>(&e1007d);

    using std::string;
    // then check if we have an classical MSB-float arch
    // (sign, exponent, fraction)
    // tested on PowerPC G4
    const char e125f_msbf[] = { 1, 0, 0, 2 };
    const char e1007d_msbf[] = { 1, 0, 0, 0, 0, 0, 0, 2 };
    
    if (string(e125f_uc, 4) == string(e125f_msbf, 4)
        && string(e1007d_uc, 8) == string(e1007d_msbf, 8))
        return false;

    // else check if we have an classical LSB-float arch
    // tested on IA32
    const char e125f_lsbf[] = { 2, 0, 0, 1 };
    const char e1007d_lsbf[] = { 2, 0, 0, 0, 0, 0, 0, 1 };
    
    if (string(e125f_uc, 4) == string(e125f_lsbf, 4)
        && string(e1007d_uc, 8) == string(e1007d_lsbf, 8))
        return true;

    // else weird unsupported arch, abort
    throw NotImplementedException(L"Architecture using an unsupported floating point endianness");
}

bool
ints_little_endianness()
{
    const uint32_t I = 0x01000002U;
    const uint64_t L = 0x0100000000000002ULL;

    const char *i_uc = reinterpret_cast<const char*>(&I);
    const char *l_uc = reinterpret_cast<const char*>(&L);

    using std::string;

    // then check if we have an classical MSB arch
    // tested on PowerPC G4
    const char i_msbf[] = { 1, 0, 0, 2 };
    const char l_msbf[] = { 1, 0, 0, 0, 0, 0, 0, 2 };

    if (string(i_uc, 4) == string(i_msbf, 4)
        && string(l_uc, 8) == string(l_msbf, 8))
        return false;

    // else check if we have an classical LSB arch
    // tested on IA32
    const char i_lsbf[] = { 2, 0, 0, 1 };
    const char l_lsbf[] = { 2, 0, 0, 0, 0, 0, 0, 1 };

    if (string(i_uc, 4) == string(i_lsbf, 4)
        && string(l_uc, 8) == string(l_lsbf, 8))
        return true;

    // else weird unsupported arch, abort
    throw NotImplementedException(L"Architecture using an unsupported integer endianness");
}

const bool floats_inverted_endianness =
    (floats_little_endianness() != ints_little_endianness());

} // unnamed namespace


/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
