/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): Marc Herbert
 */

#include "SocketAddress.hpp"
#include "Common.hpp" // for logging
#include "StringCodecs.hpp"


#include <ostream>
#include <sstream>


#if defined(__APPLE__) && defined(__MACH__)
    // http://developer.apple.com/technotes/tn2002/tn2071.html#Section10w
#include <sys/types.h> // inet_ntop
#include <sys/socket.h>
#endif

#ifdef __MINGW32__
  #include <ws2tcpip.h>
#else
  #include <netinet/in.h> // in_port_t, ntohs
#endif

#include <arpa/inet.h> // inet_ntop

#include <string.h> // strerror_r
#include <errno.h>

using namespace CarobNS;

SocketAddress::SocketAddress() : address_length(static_cast<socklen_t>(-1))
{
}

SocketAddress::SocketAddress(const SocketAddress& sa)
{
  *this = sa;
}

SocketAddress::SocketAddress(sockaddr* sa, socklen_t l) : address_length(l)
{
  setAddress(sa, l);
}

SocketAddress::~SocketAddress()
{
}
void SocketAddress::setAddress(const sockaddr* sa, socklen_t l)
{
  if (sa != NULL && l > 0)
  {
    address_length = l;
    memcpy(&address, sa, l);
  }
}

SocketAddress& SocketAddress::operator = (const SocketAddress& sa)
{
  setAddress(reinterpret_cast<const sockaddr*>(&(sa.address)), sa.address_length);
  return *this;
}  

bool SocketAddress::operator < (const SocketAddress& sa) const
{
  if (address_length != sa.address_length)
    return (address_length < sa.address_length);
  if (getFamily() != sa.getFamily())
    return (getFamily() < sa.getFamily());
  if (getFamily() == AF_INET) // ipv4
  {
    // TODO: try instead a much faster:
    // const sockaddr_in& this_in = reinterpret_cast<sockaddr_in>(address);
    sockaddr_in this_in, sa_in;
    memcpy(&this_in, &address, address_length);
    memcpy(&sa_in, &(sa.address), sa.address_length);
    if (this_in.sin_port != sa_in.sin_port)
      return (this_in.sin_port < sa_in.sin_port);
    return (this_in.sin_addr.s_addr < sa_in.sin_addr.s_addr);
  }
  else if (getFamily() == AF_INET6)// ipv6
  {
    // TODO see above
    sockaddr_in6 this_in6, sa_in6;
    memcpy(&this_in6, &address, address_length);
    memcpy(&sa_in6, &(sa.address), sa.address_length);
    if (this_in6.sin6_port != sa_in6.sin6_port)
      return (this_in6.sin6_port < sa_in6.sin6_port);
    for (int i=0; i<16; i++)
      if (this_in6.sin6_addr.s6_addr[i] >= sa_in6.sin6_addr.s6_addr[i])
        return false;
    return true;
  }
  else // should never happen
  {
    if (isErrorEnabled())
      logError(L"SocketAddress::operator <", L"Unknown socket address family!");
    return false;
  }
}

namespace
{

std::wostream&
winet_ntop_or_error(std::wostream& oss, int af, const void * sin_addr)
{

    char printableAddress[INET6_ADDRSTRLEN+1];

    if (inet_ntop(af, sin_addr, printableAddress, INET6_ADDRSTRLEN)
        == NULL)
        strerror_r(errno, printableAddress, INET6_ADDRSTRLEN);

    oss << StaticCodecs::fromASCII(printableAddress);

    return oss;
}

}

std::wostream&
CarobNS::operator<< (std::wostream& oss, const sockaddr& sock)
{

    switch(sock.sa_family) {
    case AF_INET:
    {
        const sockaddr_in& inet4addr(reinterpret_cast<const sockaddr_in&>(sock));
        winet_ntop_or_error(oss, inet4addr.sin_family, & inet4addr.sin_addr);
        oss << L":" << ntohs(inet4addr.sin_port);
    }
    break;

    case AF_INET6:
    {
        const sockaddr_in6& inet6addr(reinterpret_cast<const sockaddr_in6&>(sock));
        oss << L"[";
        winet_ntop_or_error(oss, inet6addr.sin6_family, & inet6addr.sin6_addr);
        oss << L"]:" << ntohs(inet6addr.sin6_port);
    }
    break;

    default:
    {
        oss << L"Unknown sa_family number=" << sock.sa_family;
        break;
    }
    };

    return oss;
}
