/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Gilles Rayrat
 * Contributor(s): Zsolt Simon
 */

#include "Statement.hpp"

#include "Connection.hpp"
#include "ResultSetMetaData.hpp"
#include "RequestWithResultSetParameters.hpp"

#include "CarobException.hpp"
#include "Common.hpp"

#include <string>

using std::wstring;

using namespace CarobNS;

const ResultSetOrUpdateCount Statement::endOfResultList = {false, {-1}};

Statement::Statement(Connection* c) :
sql_request(L""),
connectionPtr(c),
lastResultPtr(NULL),
metadata_holder_ptr(NULL),
rs_metadata_ptr(NULL),
lastUpdateCount(-1),
warningsPtr(NULL),
timeout(0),
fetchSize(0),
cursorName(L""),
maxRows(0),
escapeProcessing(true)
{
}

Statement::~Statement()
{
  cleanUpResults();
}

void Statement::cleanUpResults()
{
  for (std::list<ResultSetOrUpdateCount>::iterator iter = resultList.begin();
      iter != resultList.end(); iter++)
  {
    if (iter->isResultSet)
    {
      //FIXME: do we close it ?
      //iter->resultSetPtr->close();
      delete iter->resultSetPtr;
    }
  }
  resultList.clear();
  lastResultPtr = NULL;
  lastUpdateCount = -1;
  delete (warningsPtr); warningsPtr = NULL;
  delete rs_metadata_ptr; rs_metadata_ptr = NULL;
  delete metadata_holder_ptr; metadata_holder_ptr = NULL;
}

void Statement::removeResultSetFromList(DriverResultSet* drsPtr)
{
  for (std::list<ResultSetOrUpdateCount>::iterator iter = resultList.begin();
      iter != resultList.end(); iter++)
  {
    if (iter->isResultSet && iter->resultSetPtr == drsPtr)
    {
      iter->resultSetPtr = NULL;
      break;
    }
  }
}

void Statement::setReadRequestParameters(RequestWithResultSetParameters& request)
{
  request.setMaxRows(maxRows);
  request.setFetchSize(fetchSize);
  request.setCursorName(cursorName);
}

bool Statement::execute(const std::wstring &skeleton, const std::wstring &parameters) 
    throw (DriverException, SocketIOException, BackendException, ControllerException, ProtocolException,
        NotImplementedException, UnexpectedException)
{
  const wchar_t fctName[] = L"Statement::execute";
  if (isClosed())
  {
    throw DriverException(std::wstring(fctName)
			  + L": unable to execute query on a closed statement");
  }
  // invalidate the last results
  cleanUpResults();
  sql_request = skeleton;
  RequestWithResultSetParameters request(sql_request);
  request.setEscapeProcessing(escapeProcessing).setTimeoutInSeconds(timeout);
  if (!parameters.empty())
    request.setParameters(parameters);
  setReadRequestParameters(request);
  ResultAndWarnings rw = connectionPtr->statementExecute(request);
  warningsPtr = rw.getWarnings();
  resultList =  rw.getResultList();
  resultListIterator = resultList.begin();
  return getMoreResults();
}

DriverResultSet* Statement::executeQuery(const std::wstring &skeleton, const std::wstring &parameters)
    throw (DriverException, SocketIOException, BackendException,
    ControllerException, ProtocolException, NotImplementedException,
    UnexpectedException)
{
  const wchar_t fctName[] = L"Statement::executeQuery";
  if (isClosed())
  {
    throw DriverException(std::wstring(fctName)
			  + L": unable to execute query on a closed statement");
  }
  // invalidate the last write result
  cleanUpResults();
  sql_request = skeleton;
  RequestWithResultSetParameters request(sql_request);
  request.setEscapeProcessing(escapeProcessing).setTimeoutInSeconds(timeout);
  if (!parameters.empty())
    request.setParameters(parameters);
  setReadRequestParameters(request);

  ResultSetOrUpdateCount uniqueRS;
  uniqueRS.isResultSet = true;
  uniqueRS.resultSetPtr = connectionPtr->statementExecuteQuery(request);
  resultList.push_back(uniqueRS);
  resultList.push_back(endOfResultList);
  resultListIterator = resultList.begin();
  if (!getMoreResults())
    throw DriverException(std::wstring(fctName)
			  + L": invalid executeQuery result !");
  return lastResultPtr;
}

//TODO: executeUpdateWithKeys
int Statement::executeUpdate(const std::wstring &skeleton, const std::wstring &parameters) 
    throw (DriverException, SocketIOException, BackendException, 
        ControllerException, ProtocolException, UnexpectedException)
{
  const wchar_t fctName[] = L"Statement::executeUpdate";
  if (isClosed())
  {
    throw DriverException(std::wstring(fctName)
			  + L": unable to execute query on a closed statement");
  }
  sql_request = skeleton;
  cleanUpResults();
  Request request(sql_request);
  request.setEscapeProcessing(escapeProcessing).setTimeoutInSeconds(timeout);
  if (!parameters.empty())
    request.setParameters(parameters);
  ResultSetOrUpdateCount uniqueUC;
  uniqueUC.isResultSet = false;
  ResultAndWarnings rw = connectionPtr->statementExecuteUpdate(request);
  warningsPtr = rw.getWarnings();
  uniqueUC.updateCount = rw.getUpdateCount();
  resultList.push_back(uniqueUC);
  resultList.push_back(endOfResultList);
  resultListIterator = resultList.begin();
  if (getMoreResults())
    throw DriverException(L"Invalid executeQuery result !");
  return lastUpdateCount;
}

bool Statement::moreResults() const
{
  return resultListIterator != resultList.end();
}

bool Statement::getMoreResults() throw (DriverException, SocketIOException,
BackendException, ControllerException, ProtocolException, UnexpectedException)
{
  return getMoreResults(CLOSE_CURRENT_RESULT);
}

bool Statement::getMoreResults(int current) throw (DriverException,
    SocketIOException, BackendException, ControllerException, ProtocolException,
    UnexpectedException)
{
  if (current != KEEP_CURRENT_RESULT)
  {
    if (lastResultPtr != NULL)
      try
      {
        lastResultPtr->close();
        lastResultPtr = NULL;
      }
      catch (...) {} //ignore
  }
  lastUpdateCount = -1;

  ResultSetOrUpdateCount nextResult;
  if (resultListIterator == resultList.end())
  {
    // End of list, return false (updateCount is already set to -1)
    return false;
  }
  nextResult = *resultListIterator;
  resultListIterator++;
  if (nextResult.isResultSet)
  {
    lastResultPtr = nextResult.resultSetPtr;
    lastResultPtr->setStatement(this);
    return true;
  }
  else
  {
    lastUpdateCount = nextResult.updateCount;
    return false;
  }
}

// GETTERS AND SETTERS
void Statement::setQueryTimeout(int seconds) throw (DriverException, UnexpectedException)
{
  if (seconds < 0)
  {
    throw DriverException(L"Invalid query timeout value: " + toUserString(seconds));
  }
  timeout = seconds;
}

void Statement::setFetchSize(int rows) throw (DriverException, UnexpectedException)
{
  if (rows < 0
  // The spec forgets the case maxRows = 0.
      || 0 < maxRows && maxRows < rows)
  {
    throw DriverException(L"Invalid fetch size value: " + toUserString(rows));
  }
  fetchSize = rows;
}

void Statement::setResultSetConcurrency(int value) throw (DriverException,
    UnexpectedException)
{
  switch (value)
  {
    case CONCUR_READ_ONLY :
    case CONCUR_UPDATABLE :
      resultSetConcurrency = value;
      break;
    default :
      throw DriverException(L"Invalid ResultSet concurrency mode: "
          + toUserString(value));
  }
}

void Statement::setMaxRows(int max) throw (DriverException, UnexpectedException)
{
  if (max < 0)
  {
    throw DriverException(L"Invalid max rows limit: " + toUserString(max));
  }
  // this may break fetchSize <= maxRows
  maxRows = max;
}

void Statement::close()
{
  // Force the current ResultSet to close
  // For multiple resultsets, streaming is forbidden, so we don't have to
  // remotely close() them
  if (lastResultPtr != NULL)
  {
    try
    {
      lastResultPtr->close();
    }
    catch (...) {} //ignore
  }
  // free mem
  cleanUpResults();
}

ResultSetMetaData* Statement::getMetaData() throw (SocketIOException,
    BackendException, ProtocolException, NotImplementedException,
    UnexpectedException)
{
  // clean-up old one (if none, won't do anything)
  delete metadata_holder_ptr; metadata_holder_ptr = NULL;
  delete rs_metadata_ptr; rs_metadata_ptr = NULL;

  if (lastResultPtr != NULL)
  {
    rs_metadata_ptr = new ResultSetMetaData(lastResultPtr);
  }
  else
  {
    metadata_holder_ptr = connectionPtr->preparedStatementGetMetaData(sql_request);
    rs_metadata_ptr = new ResultSetMetaData(metadata_holder_ptr);
  }
  return rs_metadata_ptr;
}

SQLWarning* Statement::getWarnings() throw (DriverException, UnexpectedException)
{
  if (isClosed())
  {
    throw DriverException(L"Unable to get warnings on a closed statement");
  }
  return warningsPtr;
}
