/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2005-2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Initial developer(s): Marc Herbert
 * Contributor(s): Csaba Simon
 */

#include "StringCodecs.hpp"

#include "Common.hpp"

#include <stdexcept>
#include <sstream>

using namespace CarobNS;

namespace {

#ifdef CAROB_USE_ICONV

const int CONVERTER_BUFFER_SIZE = 1024;

/**
 *  MAYBECONST is either "const char**" or "char**" (iconv prototype
 *  has unfortunately changed overtime)
 *
 * This brilliant solution is borrowed from gcj/convert/natIconv.cc
 *
 * This seems to be the only fully portable solution avoiding any
 * "configure"-like step (AM_ICONV, etc.)
 *
 * http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.17
 */
template<typename MAYBECONST>
static inline size_t
iconv_const_adapter (size_t (*iconv_function) (iconv_t, MAYBECONST, size_t *, char **, size_t *),
                     iconv_t handle, const char **inbuf, size_t *inavail,
                     char **outbuf, size_t *outavail)
{
    // in the CONST instantiation this is casting nothing away
    return (*iconv_function) (handle, const_cast<MAYBECONST>(inbuf),
                              inavail, outbuf, outavail);
}
#endif // CAROB_USE_ICONV

// Warning: this function needs the codec of fromString() to have been
// properly initialized.
std::wstring
exception_message(const char *what, const char *locname)
{
    std::wstring msg(fromString(what));
    msg += fromString(", locale name was: ") + fromString(locname);
    return msg;
}


} // anonymous namespace


/**
 * Catching std::runtime_error from std::codecvt_byname() segfaults
 * with gcc 4.0.2-5, see
 * <http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=343108>
 * So unfortunately do NOT do this. This code is a reminder.
 */
namespace {
template <typename facet_byname_t>
facet_byname_t * trynewfacet_byname(const char * const name)
{
    try {
        return new facet_byname_t(name);
    } catch (std::runtime_error& re) {
        // we need the default user_codec to be initialized here
        std::wstring msg = exception_message(re.what(), name);
        logFatal(__WFILE__ L":trynewfacet_byname()", msg);
        throw CodecException(msg);
    }
}
}


std::locale CarobNS::trylocale(const char * const name) throw (CodecException)
{
    try {
        return std::locale(name);
    } catch (std::runtime_error& re) {
        // we need the default user_codec to be initialized here
        std::wstring msg = exception_message(re.what(), name);
        logWarn(__WFILE__ L":trylocale()" , msg);
        throw CodecException(msg);
    }
}


std::string MBSCodec::encode(const std::wstring& w_arg) const throw (CodecException)
{
    mbstate_t mbst = { 0 };

    // prefered to cvt.length(in), assumed to be too costly cause of the scan
    // (it's a memory/CPU tradeoff)
    int res_maxlength = codecvt.max_length() * w_arg.length();
    char* res_buf = new char[res_maxlength];

    const wchar_t *w_argnext;
    char *res_next;
    const wchar_t *w_argdata = w_arg.data();

    std::codecvt_base::result success =
        codecvt.out(mbst, w_argdata, w_argdata+w_arg.length(), w_argnext,
                res_buf, res_buf+res_maxlength, res_next);

    if (success != std::codecvt_base::ok) // codecvt_base::noconv is not possible
    {
        std::wostringstream woss;
        woss << L"Error " << success << L" in "
                  << fromString(loc.name()) << L" encoding method";
        logError(__WFILE__ L"MBSCodec::encode", woss.str());
        delete[] res_buf;
        throw CodecException(woss.str());
    }
    std::string retStr(res_buf, res_next-res_buf);
    delete[] res_buf;
    return retStr;
}


std::wstring MBSCodec::decode(const std::string& arg) const throw (CodecException)
{
    mbstate_t mbst = { 0 };

    int arglength = arg.length();
    wchar_t* wres_buf = new wchar_t[arglength];
    
    const char *argnext;
    wchar_t *wres_next;
    const char *argdata = arg.data();

    std::codecvt_base::result success =
        codecvt.in(mbst, argdata, argdata+arglength, argnext,
                   wres_buf, wres_buf+arglength, wres_next);

    if (success != std::codecvt_base::ok)  // codecvt_base::noconv is not possible
    {
        std::wostringstream woss;
        woss << L"Error " << success << L" in "
                  << fromString(loc.name()) << L" decoding method";
        logError(__WFILE__ L"MBSCodec::decode()", woss.str());
        delete[] wres_buf;
        throw CodecException(woss.str());
    }
    std::wstring retWStr(wres_buf, wres_next-wres_buf);
    delete[] wres_buf;
    return retWStr;
}

#ifdef CAROB_USE_ICONV

#include <errno.h>

// need two steps to delay expansion of argument
#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)


namespace {

/**
 * Trying to figure out which codeset wchar_t is using.
 *
 * On some systems it may depend on the locale... => better not use
 * iconv at all in this case but locales instead.
 */
inline const char *wchar_codeset() throw (CodecException)
{
#ifdef CAROB_WCHAR_CODESET
    // let the command line decide for us
    return TOSTRING(CAROB_WCHAR_CODESET);
#elif defined(__GLIBC__)
    // the easiest way: not deciding.
    return "WCHAR_T";
#elif defined(__STDC_ISO_10646__) || defined(CAROB_FORCE_UNICODE_WCHAR)
    // __STDC_ISO_10646__ is from C99 (ISO 9899:1999, 6.10.8.2)

    // UCS-N-INTERNAL = native + no BOM

    if (sizeof(wchar_t) == 4) // compiler will optimize this away
        return "UCS-4-INTERNAL";

    if (sizeof(wchar_t) == 2)
        return "UCS-2-INTERNAL";

    throw CodecException(L"Unsupported sizeof(wchar_t) in wchar_codeset()");
#else
#error iconv problem: unknown wchar_t codeset. Define CAROB_WCHAR_CODESET?
#endif
}
}

IconvCodec::IconvCodec(const std::string &code) throw (CodecException)
{
    // From iconv_open() @ opengroup: "Settings of fromcode and tocode
    // and their permitted combinations are implementation-defined."
  const char * wide_codeset = wchar_codeset();

  if (isDebugEnabled())
      logDebug(L"IconvCodec::IconvCodec",
               std::wstring(L"creating new codec ")
               + fromString(wide_codeset) + L" / " + fromString(code));

  cd_from = iconv_open(code.c_str(), wide_codeset);
  if (cd_from == (iconv_t) -1)
      throw CodecException(L"Failed to open iconv encoder from "
                           + fromString(wide_codeset) + L" to " + fromString(code));

  cd_to = iconv_open(wide_codeset, code.c_str());
  if (cd_to == (iconv_t) -1)
      throw CodecException(L"Failed to open iconv decoder from " + fromString(code)
                           + L" to " + fromString(wide_codeset));

  encodingName = StaticCodecs::fromASCII(code);

}

IconvCodec::~IconvCodec()
{
  {
    LockScope ls(&iconvEncodeCS);
    iconv_close(cd_from);
  }
  {
    LockScope ls(&iconvDecodeCS);
    iconv_close(cd_to);
  }
}

namespace {
// reset the state
size_t
reset_iconv(iconv_t desc)
{
    return iconv_const_adapter(iconv, desc, 0, 0, 0, 0);
}
}

std::wstring
IconvCodec::decode(const std::string &s) const throw (CodecException)
{
  // lock the iconv decode descriptor
  LockScope ls(&iconvDecodeCS);

  return nomutex_decode(s);
}

std::wstring
IconvCodec::nomutex_decode(const std::string &s) const throw (CodecException)
{
  if (s.empty())
    return std::wstring();
   
  std::wstring result;

  const char *inbuf = s.data();
  size_t inbytesleft = s.length();


  wchar_t convertbuf[CONVERTER_BUFFER_SIZE];
  char *outbuf;
  size_t outbytesleft;
  
  while (inbytesleft > 0)
  {
    outbuf = (char *)convertbuf;
    outbytesleft = sizeof(convertbuf) - sizeof(convertbuf[0]);
    
    size_t ret = iconv_const_adapter(iconv, cd_to, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
    if (ret == size_t(-1))
    {
      switch(errno)
      {
        case E2BIG:
          break;
        case EILSEQ:
          reset_iconv(cd_to);
          throw CodecException(encodingName + L" IconvCodec: an invalid multibyte sequence"
                               L" has been encountered in the bytes to decode");
          break;
        case EINVAL:
        default:
          reset_iconv(cd_to);
          throw CodecException(encodingName + L" IconvCodec: an incomplete multibyte sequence"
                               L" has been encountered in the bytes to decode");
          break;
      }
    }
    result.append(convertbuf,(wchar_t *)outbuf - convertbuf);
  }
  
  return result;
}

std::string
IconvCodec::encode(const std::wstring &s) const throw (CodecException)
{
  // lock the iconv encode descriptor
  LockScope ls(&iconvEncodeCS);

  return nomutex_encode(s);
}


std::string
IconvCodec::nomutex_encode(const std::wstring &w) const throw (CodecException)
{
  if (w.empty())
    return std::string();
  
  std::string result;

  const char *inbuf = reinterpret_cast<const char *>(w.data());
  size_t inbytesleft = w.length() * sizeof(wchar_t);
  
  char convertbuf[CONVERTER_BUFFER_SIZE];
  char *outbuf;
  size_t outbytesleft;

  while (inbytesleft > 0)
  {
    outbuf = convertbuf;
    outbytesleft = sizeof(convertbuf) - sizeof(convertbuf[0]);
    
    size_t ret = iconv_const_adapter(iconv, cd_from, &inbuf, &inbytesleft, &outbuf, &outbytesleft);
    if (ret == size_t(-1))
    {
      switch(errno)
      {
        case E2BIG:
          break;
        case EILSEQ:
          reset_iconv(cd_from);
          throw CodecException(encodingName + L" IconvCodec: An invalid multibyte sequence"
                               L" has been encountered in the wstring to encode");
          break;
        case EINVAL:
        default:
          reset_iconv(cd_from);
          throw CodecException(encodingName + L" IconvCodec: An incomplete multibyte sequence"
                               L" has been encountered in the wstring to encode");
          break;
      }
    }
    result.append(convertbuf,outbuf - convertbuf);
  }
  
  return result;  
}
#endif // CAROB_USE_ICONV

/*
 * Local Variables:
 * c-file-style: "bsd"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
