/*
 * Sequoia: Database clustering technology.
 * Copyright (C) 2006 Continuent, Inc.
 * Contact: sequoia@continuent.org
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 *
 * Initial developer(s): Gilles Rayrat.
 * Contributor(s): ______________________.
 */

#include "WatchedControllers.hpp"
#include "ControllerStateChangedCallback.hpp"

#include "Common.hpp"

#include <string>

using namespace CarobNS;
using std::map;
using std::vector;
using std::wstring;

WatchedControllers::ControllerAndState::ControllerAndState() :
    responds_to_pings(true), last_time_seen(-1)
{
}

WatchedControllers::ControllerAndState::ControllerAndState(const ControllerInfo& ctrl) :
    controller(ctrl), responds_to_pings(true), last_time_seen(-1)
{
}

WatchedControllers::ControllerAndState::ControllerAndState(const ControllerInfo& ctrl, int64_t initTime) :
    controller(ctrl), responds_to_pings(true), last_time_seen(initTime)
{
}

WatchedControllers::ControllerAndState::ControllerAndState(const ControllerAndState& src)
{
  *this = src;
}

WatchedControllers::ControllerAndState& WatchedControllers::ControllerAndState::operator = (const ControllerAndState& src)
{
  controller = src.controller;
  responds_to_pings = src.responds_to_pings;
  last_time_seen = src.last_time_seen;
  return *this;
}

/* init static member critical section */
CriticalSection WatchedControllers::controllers_CS;
WatchedControllers::WatchedControllers(const vector<ControllerInfo>& controllersPrm,
    int64_t initTime, int controllerTimeout, ControllerStateChangedCallback* cbPtr) :
    controller_timeout(controllerTimeout),
    callback_ptr(cbPtr)
{
  // we are inside a contructor, so we can safely operate on the
  // list without worrying of concurrent modifications
  for (std::vector<ControllerInfo>::const_iterator iter = controllersPrm.begin(); iter != controllersPrm.end(); iter++)
  {
    ControllerAndState cs(*iter, initTime);
    this->controllers.insert(std::pair<SocketAddress, ControllerAndState>(iter->getSocketAddress(), cs));
  }
}

WatchedControllers::~WatchedControllers()
{
  controllers.clear();
  delete callback_ptr;
}

void WatchedControllers::addController(const ControllerInfo& ctrl, int64_t lastTimeSeen)
{
  LockScope ls(&controllers_CS);
  ControllerAndState cs(ctrl, lastTimeSeen);
  this->controllers.insert(std::pair<SocketAddress, ControllerAndState>(ctrl.getSocketAddress(), cs));
}

void WatchedControllers::removeController(const ControllerInfo& ctrl)
{
  LockScope ls(&controllers_CS);
  controllers.erase(ctrl.getSocketAddress());
}

void WatchedControllers::setControllerResponsed(const SocketAddress& ctrlAddr, int64_t newTime)
{
  const wchar_t fctName[] = L"WatchedControllers::setControllerResponsed";
  LockScope ls(&controllers_CS);
  ControllerAndState& state = controllers[ctrlAddr];
  // Set the new time
  state.updateTime(newTime);
  // if the controller was down, we launch the callback
  if (!state.isUp())
  {
    if (isInfoEnabled())
      logInfo(fctName, L"Controller " + static_cast<wstring>(state.getController())
           + L" responds to pings again. Adding it to the pool.");
    state.setUp();
    callback_ptr->onControllerUp(state.getController());
  }
}

void WatchedControllers::lookForDeadControllers(int64_t currentTime)
{
  const wchar_t fctName[] = L"WatchedControllers::lookForDeadControllers";
  LockScope ls(&controllers_CS);
  for (map<SocketAddress, ControllerAndState>::iterator iter = controllers.begin();
      iter != controllers.end(); iter++)
  {
    if (iter->second.isUp()
        && int(currentTime - iter->second.getLastTimeSeen()) > controller_timeout)
    {
      if (isWarnEnabled())
        logWarn(fctName, L"Controller " + static_cast<wstring>(iter->second.getController())
             + L" does not respond to pings. Considering it as down.");
      iter->second.setDown();
      callback_ptr->onControllerDown(iter->second.getController());
    }
  }
}

void WatchedControllers::setControllerDown(const SocketAddress& ctrlAddr)
{
  LockScope ls(&controllers_CS);
  ControllerAndState& state = controllers[ctrlAddr];
  if (state.isUp()) // don't do the job twice
  {
    if (isWarnEnabled())
      logWarn(L"WatchedControllers::setControllerDown", L"Controller " + static_cast<wstring>(state.getController())
           + L" detected as down by driver");
    state.setDown();
    callback_ptr->onControllerDown(state.getController());
  }
}

vector<SocketAddress> WatchedControllers::getControllerAddresses() const
{
  // get a copy of the map to work on.
  map<SocketAddress, ControllerAndState> copy;
  {
    LockScope ls(&controllers_CS);
    copy = controllers;
  }
  vector<SocketAddress> ctrls;
  for (map<SocketAddress, ControllerAndState>::const_iterator iter = copy.begin(); iter != copy.end(); iter++)
    ctrls.push_back(iter->first);
  return ctrls;
}

